/****************************************************************************
 *
 *		Copyright (C) 2002-2003	YAMAHA CORPORATION. All rights reserved.
 *
 *		Module		: marmdcnv.c
 *
 *		Description	: Realtime MIDI/GML Stream Converter Module.
 *
 *		Version		: 1.0.3 	2003.06.26
 *
 ****************************************************************************/


#include "marmdcnv.h"

/*--------------------------------------------------------------------------*/
/*   Defines                                                                */
/*--------------------------------------------------------------------------*/
#define	RMD_MAX_GAIN		76				/* - 9[dB]                  */
											/* - 6[dB] : 90             */
											/* - 9[dB] : 76             */
											/* -12[dB] : 64             */
											/* -18[dB] : 45             */

#define RMD_INTERVAL_CB		(17)

#define ENC_8BIT			(0)
#define ENC_7BIT			(1)

#define	MAX_NUM_OF_WAVES	(32)

#define MODE_DRUM			(0x00)			/* GMX 			*/

/*--------------------------------------------------------------------------*/
/*   Types                                                                  */
/*--------------------------------------------------------------------------*/
typedef struct _MASMW_MIDIMSG
{
	UINT8*	pbMsg;						/* pointer to MIDI message */
	UINT32	dSize;						/* size of MIDI message */
} MASMW_MIDIMSG, *PMASMW_MIDIMSG;


typedef	struct _MASMW_GETCTL
{
	UINT8	bControl;					/* contorl number */
	UINT8 	bCh;						/* channel number */
} MASMW_GETCTL, *PMASMW_GETCTL;


typedef	struct _tagRmdGlobals
{
 	SINT32		(*pfnCallbackFunc)(UINT8 bId);	/*                       */

	SINT32		sdSeqID;						/* Sequence ID           */
	SINT32		sdFileID;						/* File ID               */
	UINT32		dOpend;							/* 0:Closed, 1:Opend     */
	UINT32		dNumOfLoaded;					/*                       */
	UINT32		dIntervalTime;					/* 0 or 20[ms]           */

	UINT32		dMaxGain;						/* MaxGain (0..127)      */
	UINT32		dMasterVolume;					/* MasterVolume (0..127) */

	UINT32		dRamBase;						/*                       */
	UINT32		dRamOffset;						/*                       */
	UINT32		dRamSize;						/*                       */

	UINT32		dBankTable[16];					/* 0..0x7F7F             */
	UINT32		dRPNTable[16];					/* 0..0xFF7F             */
	UINT32		dChVolumeTable[16];				/* 0..127                */
	UINT32		dExpressionTable[16];			/* 0..127                */
	UINT32		dFineTune[16];					/* 0..0x3FFF			 */
	UINT32		dWtAdrTable[MAX_NUM_OF_WAVES];	/*                       */
} RMDGLOBAL, *PRMDGLOBAL;

/*--------------------------------------------------------------------------*/
/*   Consts                                                                 */
/*--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*   Globals                                                                 */
/*---------------------------------------------------------------------------*/
static RMDGLOBAL	gRmdInfo;
static PRMDGLOBAL	gpRmdInfo;

/*---------------------------------------------------------------------------*/
/*   Functions (internal use only)                                           */
/*---------------------------------------------------------------------------*/

/****************************************************************************
 *	SendMasterVolume(UINT32 dVol)
 *
 *	Description
 *		Send SetMasterVolume message
 *	Param
 *		dVol	Master volume (0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendMasterVolume(UINT32 dVol)
{
	if (gpRmdInfo->sdSeqID > 0)
	{
		MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_MASTER_VOLUME, dVol, 0, 0);
	}
	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	ErrorFunction(UINT32 dCh, UINT32 dPara1, UINT32 dPara2)
 *
 *	Description
 *		Invalid messages
 *
 *	Param
 *		dCh		MIDI channel (0..15)
 *		dPara1	Undefined
 *		dPara2	Undefined
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	ErrorFunction(UINT32 dCh, UINT32 dPara1, UINT32 dPara2)
{
	(void)dCh;
	(void)dPara1;
	(void)dPara2;

	MARMDCNV_DBGMSG(("ErrorFunction\n"));

	return (MASMW_ERROR_ARGUMENT);
}


/****************************************************************************
 *	NotSupported(UINT32 dCh, UINT32 dPara1, UINT32 dPara2)
 *
 *	Description
 *		Unsupported messages
 *	Param
 *		dCh		MIDI channel (0..15)
 *		dPara1	Undefined
 *		dPara2	Undefined
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	NotSupported(UINT32 dCh, UINT32 dPara1, UINT32 dPara2)
{
	(void)dCh;
	(void)dPara1;
	(void)dPara2;

	MARMDCNV_DBGMSG(("NotSupported\n"));

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	SendNoteOff(UINT32 dCh, UNIT32 dKey, UINT32 dVel)
 *
 *	Description
 *		Send NoteOff message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dKey		...	#Key (0..127)
 *		dVel		...	Velocity (0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendNoteOff(UINT32 dCh, UINT32 dKey, UINT32 dVel)
{
	(void)dVel;

	MARMDCNV_DBGMSG(("SendNoteOff[%ld, %ld] \n", dCh, dKey));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_NOTE_OFF, dCh, dKey, 0));
}


/****************************************************************************
 *	SendNoteOn(UINT32 dCh, UINT32 dKey, UINT32 dVel)
 *
 *	Description
 *		Send NoteOn message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dKey		...	#Key (0..127)
 *		dVel		...	Velocity (0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendNoteOn(UINT32 dCh, UINT32 dKey, UINT32 dVel)
{
	if (dVel == 0) return(SendNoteOff(dCh, dKey, dVel));

	MARMDCNV_DBGMSG(("SendNoteOn[%ld, %ld, %ld]\n", dCh, dKey, dVel));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_NOTE_ON, dCh, dKey, dVel));
}


/****************************************************************************
 *	SendProgram(UINT32 dCh, UNIT32 dProg, UINT32 dPara2)
 *
 *	Description
 *		Send Program message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dProg		...	#Program(0..127)
 *		dPara2
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendProgram(UINT32 dCh, UINT32 dProg, UINT32 dPara2)
{
	UINT8	bBk, bPg;
	UINT32	dBb;
	UINT8	bPp;

	static const UINT8	bMBank[128] = 
	{
		1, 2, 3, 4, 5, 6, 7, 8, 9, 10,0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};
	(void)dPara2;
	
	MARMDCNV_DBGMSG(("SendProgram[%ld] = %ld\n", dCh, dProg));

	bPp = (UINT8)(dProg & 0x7f);

	dBb = gpRmdInfo->dBankTable[dCh];

	switch (dBb)
	{
	case 0x7900:	/* Melody */
		bBk = 0;
		bPg = bPp;
		break;

	case 0x7800:	/* Drum */
		bBk = 128;
		bPg = 0;
		break;

	default:		/* Unknown : default GM */
		switch (dBb >> 8)
		{
		case 0x79:	/* Default Melody */
			bBk = 0;
			bPg = bPp;
			break;

		case 0x7C:	/* Ex Melody */
			bBk = (UINT8)(bMBank[dBb & 0x7F]);
			bPg = bPp;
			break;

		case 0x7D:	/* Ex Drum */
			bBk = (UINT8)(128 + bMBank[bPp]);
			bPg = 0;
			break;
		
		default:
			if (dCh == 9)
			{
				bBk = 128;
				bPg = 0;
			} else {
				bBk = 0;
				bPg = bPp;
			}
		}
	}
	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_PROGRAM_CHANGE, dCh, (UINT32)bBk, (UINT32)bPg));
}

/****************************************************************************
 *	SendModDepth(UINT32 dCh, UINT32 dMod)
 *
 *	Description
 *		Send SetModulation message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dMod		...	Modulation(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendModDepth(UINT32 dCh, UINT32 dMod)
{
	static const UINT8	bModTable[128] = 
	{
		0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
		2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
		3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
		3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
		4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
		4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4
	};

	MARMDCNV_DBGMSG(("SendModDepth[%ld] = %ld\n", dCh, dMod));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_MODULATION_DEPTH, dCh, (UINT32)bModTable[dMod], 0));
}


/****************************************************************************
 *	SendChVol(UINT32 dCh, UINT32 dVol)
 *
 *	Description
 *		Send ChannelVolume message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVol		...	Volume(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendChVol(UINT32 dCh, UINT32 dVol)
{
	MARMDCNV_DBGMSG(("SendChVol[%ld] =  %ld\n", dCh, dVol));

	gpRmdInfo->dChVolumeTable[dCh] = dVol;

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_CHANNEL_VOLUME, dCh, dVol, 0));
}


/****************************************************************************
 *	SendExpression(UINT32 dCh, UINT32 dVol)
 *
 *	Description
 *		Send Expression message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVol		...	Volume(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendExpression(UINT32 dCh, UINT32 dVol)
{
	UINT32	dMix;
	
	MARMDCNV_DBGMSG(("SendExpression[%ld] =  %ld\n", dCh, dVol));

	gpRmdInfo->dExpressionTable[dCh] = dVol;
	
	dMix = MaSndDrv_Mix40LogVol(dVol, gpRmdInfo->dMasterVolume);
	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_EXPRESSION, dCh, dMix, 0));
}


/****************************************************************************
 *	SendPanpot(UINT32 dCh, UINT32 dPan)
 *
 *	Description
 *		Send Panpot message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		bPan		...	Panpot(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendPanpot(UINT32 dCh, UINT32 dPan)
{
	MARMDCNV_DBGMSG(("SendPanpot[%ld] = %ld\n", dCh, dPan));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_PANPOT, dCh, dPan, 0));
}


/****************************************************************************
 *	SendHold1(UINT32 dCh, UINT32 dVal)
 *
 *	Description
 *		Send Hold1 message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVal		...	Hold1(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendHold1(UINT32 dCh, UINT32 dVal)
{
	MARMDCNV_DBGMSG(("SendHold1[%ld] = %ld\n", dCh, dVal));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_HOLD1, dCh, (UINT32)((dVal < 64) ? 0 : 1), 0));
}


/****************************************************************************
 *	DataEntryMSB(UINT32 dCh, UINT32 dVal)
 *
 *	Description
 *		Send RPN message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVal		...	data(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	DataEntryMSB(UINT32 dCh, UINT32 dVal)
{
	MARMDCNV_DBGMSG(("DataEntryMSB[%ld] = %ld\n", dCh, dVal));

	dVal &= 0x7f;
	switch (gpRmdInfo->dRPNTable[dCh])
	{
	case 0:	/* BendSens */
		return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_BEND_RANGE, dCh, dVal, 0));

	case 1:	/* FineTune */
		gpRmdInfo->dFineTune[dCh] = dVal << 7;
		return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_FINE_TUNE, dCh, gpRmdInfo->dFineTune[dCh], 0));
		break;

	case 2:	/* CoaseTune */
		return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_COASE_TUNE, dCh, dVal, 0));
		break;
	}
	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	DataEntryLSB(UINT32 dCh, UINT32 dVal)
 *
 *	Description
 *		Send RPN message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVal		...	data(0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	DataEntryLSB(UINT32 dCh, UINT32 dVal)
{
	MARMDCNV_DBGMSG(("DataEntryLSB[%ld] = %ld\n", dCh, dVal));

	switch (gpRmdInfo->dRPNTable[dCh])
	{
	case 1:	/* FineTune */
		gpRmdInfo->dFineTune[dCh] &= 0x3F80;
		gpRmdInfo->dFineTune[dCh] |= (dVal & 0x7f);
		return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_FINE_TUNE, dCh, gpRmdInfo->dFineTune[dCh], 0));
	}
	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	SendAllSoundOff(UINT32 dCh)
 *
 *	Description
 *		Send AllSoundOff message
 *	Param
 *		dCh			...	#Channel(0..15)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendAllSoundOff(UINT32 dCh)
{
	MARMDCNV_DBGMSG(("SendAllSoundOff[%ld] \n", dCh));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_ALL_SOUND_OFF, dCh, 0, 0));
}


/****************************************************************************
 *	SendAllNoteOff(UINT32 dCh)
 *
 *	Description
 *		Send AllNoteOff message
 *	Param
 *		dCh			...	#Channel(0..15)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendAllNoteOff(UINT32 dCh)
{
	MARMDCNV_DBGMSG(("SendAllNoteOff[%ld] \n", dCh));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_ALL_NOTE_OFF, dCh, 0, 0));
}


/****************************************************************************
 *	SendResetAllCtl(UINT32 dCh)
 *
 *	Description
 *		Send ResetAllControllers message
 *	Param
 *		dCh			...	#Channel(0..15)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendResetAllCtl(UINT32 dCh)
{
	MARMDCNV_DBGMSG(("SendResetAllCtl[%ld] \n", dCh));

	gpRmdInfo->dRPNTable[dCh] = 0x7F7F;
	gpRmdInfo->dExpressionTable[dCh] = 127;

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_RESET_ALL_CONTROLLERS, dCh, 0, 0));
}


/****************************************************************************
 *	SendMonoOn(UINT32 dCh, UINT32 dVal)
 *
 *	Description
 *		Send MonoOn message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVal		... 
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendMonoOn(UINT32 dCh, UINT32 dVal)
{
	MARMDCNV_DBGMSG(("SendMonoOn[%ld] \n", dCh));

	if (dVal != 1) return (MASMW_SUCCESS);

	MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_ALL_NOTE_OFF, dCh, 0, 0);
	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_MONO_MODE_ON, dCh, 0, 0));
}


/****************************************************************************
 *	SendPolyOn(UINT32 dCh, UINT32 dVal)
 *
 *	Description
 *		Send PolyOn message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dVal		... 
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendPolyOn(UINT32 dCh, UINT32 dVal)
{
	MARMDCNV_DBGMSG(("SendPolyOn[%ld] \n", dCh));

	if (dVal != 0) return (MASMW_SUCCESS);

	MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_ALL_NOTE_OFF, dCh, 0, 0);
	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_POLY_MODE_ON, dCh, 0, 0));
}


/****************************************************************************
 *	SendControl(UINT32 dCh, UNIT32 dMsg2, UINT32 dMsg3)
 *
 *	Description
 *		Control messages
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dMsg2		...	Param-1
 *		dMsg3		...	Param-2
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendControl(UINT32 dCh, UINT32 dMsg2, UINT32 dMsg3)
{
	MARMDCNV_DBGMSG(("SendControl[%ld, %ld, %ld] \n", dCh, dMsg2, dMsg3));

	switch (dMsg2)
	{
	case 0x00:	/* Bank select(MSB) */
		gpRmdInfo->dBankTable[dCh] = (gpRmdInfo->dBankTable[dCh] & 0x00FF) | ((UINT32)dMsg3 << 8);
		break;
    
    case 0x20:	/* Bank select (LSB) */
		gpRmdInfo->dBankTable[dCh] = (gpRmdInfo->dBankTable[dCh] & 0xFF00) | ((UINT32)dMsg3);
		break;

	case 0x01:
		return (SendModDepth(dCh, dMsg3));

	case 0x06:
		return (DataEntryMSB(dCh, dMsg3));

	case 0x07:
		return (SendChVol(dCh, dMsg3));

	case 0x0A:
		return (SendPanpot(dCh, dMsg3));

	case 0x0B:
		return (SendExpression(dCh, dMsg3));

	case 0x26:
		return (DataEntryLSB(dCh, dMsg3));

	case 0x40:
		return (SendHold1(dCh, dMsg3));


	case 0x62:	/* NRPN (LSB) */
	case 0x63:	/* NRPN (MSB) */
		gpRmdInfo->dRPNTable[dCh] |= 0x8000;
		break;

	case 0x64:	/* RPN (LSB) */
		gpRmdInfo->dRPNTable[dCh] = (gpRmdInfo->dRPNTable[dCh] & 0x7F00) | ((UINT32)dMsg3);
		break;

	case 0x65:	/* RPN (MSB) */
		gpRmdInfo->dRPNTable[dCh] = (gpRmdInfo->dRPNTable[dCh] & 0x007F) | ((UINT32)dMsg3 << 8);
		break;

	case 0x78:
		return (SendAllSoundOff(dCh));

	case 0x7B:
		return (SendAllNoteOff(dCh));

	case 0x79:
		return (SendResetAllCtl(dCh));

	case 0x7e:
		return (SendMonoOn(dCh,dMsg3));
		
	case 0x7f:
		return (SendPolyOn(dCh, dMsg3));

	default:
		return (MASMW_SUCCESS);
	}
	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	SendPitchBend(UINT32 dCh, UNIT32 dLl, UINT32 dHh)
 *
 *	Description
 *		Send PitchBend message
 *	Param
 *		dCh			...	#Channel(0..15)
 *		dLl			...	Lower 7bit of PitchBend (0..127)
 *		dHh			...	Upper 7bit of PitchBend (0..127)
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendPitchBend(UINT32 dCh, UINT32 dLl, UINT32 dHh)
{
	MARMDCNV_DBGMSG(("SendPitchBend[%ld] = %ld\n", dCh, dHh));

	return (MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_PITCH_BEND, dCh, ((dHh << 7) | (dLl)), 0));
}


/****************************************************************************
 *	SendGmOn(void)
 *
 *	Description
 *		Send GM_ON message
 *	Param
 *		none
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	SendGmOn(void)
{
	SINT32	sdRet;
	UINT32	dCh;
	
	MARMDCNV_DBGMSG(("SendGmOn\n"));

	gpRmdInfo->dMaxGain = RMD_MAX_GAIN;
	gpRmdInfo->dMasterVolume = 127;
	for (dCh = 0; dCh < 16; dCh++)
	{
		SendAllSoundOff(dCh);
		gpRmdInfo->dRPNTable[dCh] = 0x7f7f;
		gpRmdInfo->dBankTable[dCh] = (dCh == 9) ? 0x7800 : 0x7900;
		gpRmdInfo->dChVolumeTable[dCh] = 100;
		gpRmdInfo->dExpressionTable[dCh] = 127;
	}
	sdRet = MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_SYSTEM_ON, 0, 0, 0);
	if (sdRet < 0) return (sdRet);
	sdRet = SendMasterVolume(gpRmdInfo->dMaxGain);
	return (sdRet);
}


/****************************************************************************
 *	Decode7Enc(UINT8* pbSrc, UINT32 dSrcSize, UINT8* pbDst, UINT32 dDstSize)
 *
 *	Desc.
 *		7bit->8bit decoder
 *	Param
 *		pbSrc		Pointer to the source
 *		pbSrcSize	Size of the source[byte]
 *		pbDst		Pointer to the destination
 *		pbDstSize	Size of the destination[byte]
 *	Return
 *      size of decoded data
 ****************************************************************************/
static UINT32 Decode7Enc(UINT8* pbSrc, UINT32 dSrcSize, UINT8* pbDst, UINT32 dDstSize)
{
	UINT32	i, j;
	UINT8*	pbEnd;
	UINT8	bTag;
	
	i = 0;
	pbEnd = pbSrc + dSrcSize;

	while (i < dDstSize)
	{
		bTag = *(pbSrc++);
		if (pbSrc >= pbEnd) break;
		
		for (j = 0; j < 7; j++)
		{
			bTag <<= 1;
			pbDst[i++] = (UINT8)((bTag & 0x80) | *(pbSrc++));
			if (pbSrc >= pbEnd) break;
			if (i >= dDstSize) return (i);
		}
	}
	return (i);
}



/*---------------------------------------------------------------------------*/
/*   Functions                                                               */
/*---------------------------------------------------------------------------*/

/****************************************************************************
 *	MaRmdCnv_Initialize(void)
 *
 *	Description
 *		Initialize RealtimeMIDI converter
 *	Param
 *		none
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Initialize(void)
{
	gpRmdInfo = &gRmdInfo;
	gpRmdInfo->sdSeqID = -1;						/* Sequence ID      */
	gpRmdInfo->sdFileID = -1;						/* File ID          */
	gpRmdInfo->dOpend = 0;							/* 0:disabed        */
	gpRmdInfo->dNumOfLoaded = 0;					/*                  */

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_End(void)
 *
 *	Description
 *		Ending of RealtimeMIDI converter
 *	Param
 *		none
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_End(void)
{
	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_Convert(void)
 *
 *	Description
 *		No effect
 *	Param
 *		none
 *	Return
 *		1 : Continue
 ****************************************************************************/
static SINT32	MaRmdCnv_Convert(void)
{
	if ((gpRmdInfo->dIntervalTime > 0) && (gpRmdInfo->pfnCallbackFunc != NULL))
	{
		gpRmdInfo->pfnCallbackFunc(RMD_INTERVAL_CB);
	}
	return (1);
}


/****************************************************************************
 *	MaRmdCnv_Load(UINT8* pbFile, UINT32 dFileSize, 
 *					UINT32 dMode, SINT32 (*pfnFunc)(UINT8 bID), void* pvExtArgs)
 *
 *	Description
 *		Load MIDI settings
 *	Param
 *		pbFile		... pointer to the data
 *		dFileSize	... size fo the data
 *		dMode		... error check (0:No, 1:Yes, 2:Check)
 *		pfnFunc		... pointer of rhe callback function
 *		pvExtArgs	... NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Load(UINT8* pbFile, UINT32 dFileSize, UINT32 dMode, SINT32 (*pfnFunc)(UINT8 bID), void* pvExtArgs)
{
	(void)pbFile;
	(void)dFileSize;
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Load\n"));

	if (dMode > 1) return (MASMW_SUCCESS);
	if (gpRmdInfo->dNumOfLoaded > 0) return (MASMW_ERROR);
	gpRmdInfo->dNumOfLoaded++;
	gpRmdInfo->pfnCallbackFunc = pfnFunc;
	gpRmdInfo->sdFileID = gpRmdInfo->dNumOfLoaded;
	gpRmdInfo->dIntervalTime = 0;

	return ((SINT32)gpRmdInfo->sdFileID);
}


/****************************************************************************
 *	MaRmdCnv_Unload(SINT32 sdFileID, void* pvExtArgs)
 *
 *	Description
 *		Unload MIDI settings
 *	Param
 *		sdFileID		...	file ID
 *		pvExtArgs		...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Unload(SINT32 sdFileID, void* pvExtArgs)
{
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Unload[%ld] \n", sdFileID));

	if ((sdFileID < 0) || (sdFileID != (SINT32)gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	gpRmdInfo->dNumOfLoaded = 0;
	gpRmdInfo->sdFileID = -1;

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_Open(SINT32 sdFileID, void* pvExtArgs)
 *
 *	Description
 *		Enable MIDI
 *	Param
 *		sdFileID		...	file ID
 *		pvExtArgs		...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Open(SINT32 sdFileID, void* pvExtArgs)
{
	UINT32	i;
	UINT32	dRam;
	UINT32	dSize;

	(void)pvExtArgs;
	
	MARMDCNV_DBGMSG(("MaRmdCnv_Open[%ld] \n", sdFileID));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	gpRmdInfo->sdSeqID = MaSndDrv_Create(
			MASMW_SEQTYPE_DIRECT,
			gpRmdInfo->dIntervalTime,				/* Interval Timer       */
			4L,										/* 16+16 Mode           */
			7L,										/* 4-OP,GML,MA5VOL      */
			0L,										/* No Streams           */
			255L,									/* No AL                */
			MaRmdCnv_Convert,						/*                      */
			&dRam,									/*                      */
			&dSize);								/*                      */

	MARMDCNV_DBGMSG(("MaRmdCnv_Open/SeqID=%ld \n", gpRmdInfo->sdSeqID));
	if (gpRmdInfo->sdSeqID < 0) return (gpRmdInfo->sdSeqID);

	MaSndDrv_SetVolShift(gpRmdInfo->sdSeqID, 2, 2);	/* WT +12dB              */

	gpRmdInfo->dRamBase = dRam;
	gpRmdInfo->dRamSize = dSize;
	gpRmdInfo->dMasterVolume = 127L;
	gpRmdInfo->dOpend = 1;
	gpRmdInfo->dMaxGain = RMD_MAX_GAIN;				/* RMD_MAX_GAIN[dB]      */
	gpRmdInfo->dRamOffset = 0;
	for (i = 0; i < MAX_NUM_OF_WAVES ; i++) gpRmdInfo->dWtAdrTable[i] = 0;

	SendMasterVolume(gpRmdInfo->dMaxGain);

	return (SendGmOn());
}


/****************************************************************************
 *	MaRmdCnv_Close(SINT32 sdFileID, void* pvExtArgs)
 *
 *	Description
 *		Disable MIDI
 *	Param
 *		sdFileID		...	file ID
 *		ext_args		...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Close(SINT32 sdFileID, void* pvExtArgs)
{
	UINT32	dCh;
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Close[%ld] \n", sdFileID));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	for (dCh = 0; dCh < 16; dCh++)
	{
		MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_ALL_SOUND_OFF, dCh, 0L, 0L);
	}

	MaSndDrv_Free(gpRmdInfo->sdSeqID);

	gpRmdInfo->sdSeqID = -1;
	gpRmdInfo->dOpend = 0;

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_Standby(SINT32 sdFileID, void* pvExtArgs)
 *
 *	Description
 *		Initialize MIDI Engine
 *	Param
 *		sdFileID		...	file ID
 *		pvExtArgs		...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Standby(SINT32 sdFileID, void* pvExtArgs)
{
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Standby[%ld] \n", sdFileID));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_Seek(SINT32 sdFileID, UINT32 dPos, void* pvExtArgs)
 *
 *	Description
 *		No effect
 *	Praram
 *		sdFileID		...	file ID
 *		dPos			...	Start point(msec)
 *		pvExtArgs		...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Seek(SINT32 sdFileID, UINT32 dPos, void* pvExtArgs)
{
	(void)dPos;
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Seek[%ld %ld] \n", sdFileID, dPos));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_Start(SINT32 sdFileID, void* pvExtArgs)
 *
 *	Description
 *		No effect
 *	Param
 *		sdFileID	...	file ID
 *		pvExtArgs	...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Start(SINT32 sdFileID, void* pvExtArgs)
{
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Start[%p] \n", pvExtArgs));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}
	if (gpRmdInfo->dIntervalTime > 0) MaSndDrv_ControlSequencer(gpRmdInfo->sdSeqID, 1);

	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	MaRmdCnv_Stop(SINT32 sdFileID, void* pvExtArgs)
 *
 *	Description
 *		Stop MIDI Play
 *	Param
 *		sdFileID		...	file ID
 *		pvExtArgs		...	NULL
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Stop(SINT32 sdFileID, void* pvExtArgs)
{
	UINT32	dCh;
	(void)pvExtArgs;
	
	MARMDCNV_DBGMSG(("MaRmdCnv_Stop[%ld] \n", sdFileID));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	if (gpRmdInfo->dIntervalTime > 0) MaSndDrv_ControlSequencer(gpRmdInfo->sdSeqID, 0);

	for (dCh = 0; dCh < 16; dCh++)
	{
		MaSndDrv_SetCommand(gpRmdInfo->sdSeqID, -1, MASNDDRV_CMD_ALL_SOUND_OFF, dCh, 0, 0);
	}
	
	return (MASMW_SUCCESS);
}


/****************************************************************************
 *	SINT32	MaRmdCnv_SetShortMsg(SINT32 sdFileID, UINT32 dMsg)
 *
 *	Description
 *		Send MIDI short message
 *	Param
 *		sdFileID		...	file ID
 *		dMsg			...	MIDI message
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	MaRmdCnv_SetShortMsg(SINT32 sdFileID, UINT32 dMsg)
{
	UINT32	dMsg1, dMsg2, dMsg3;
	UINT32	dCh;

	static SINT32 (*fnMidiMsg[16])(UINT32 dChannel, UINT32 dPara1, UINT32 dPara2) =
	{
		ErrorFunction,	ErrorFunction,	ErrorFunction,	ErrorFunction, 
		ErrorFunction, 	ErrorFunction,	ErrorFunction,	ErrorFunction, 
		SendNoteOff,	SendNoteOn,		NotSupported,	SendControl,
		SendProgram,	NotSupported,	SendPitchBend,	NotSupported
	};

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	MARMDCNV_DBGMSG(("MaRmdCnv_SetShortMsg[%08lX]\n", dMsg));

	dMsg1 = (dMsg & 0xf0) >> 4;
	dCh   = dMsg & 0x0f;
	dMsg2 = (dMsg >> 8) & 0x7f;
	dMsg3 = (dMsg >> 16) & 0x7f;

	return (fnMidiMsg[dMsg1](dCh, dMsg2, dMsg3));
}


/****************************************************************************
 *	MaRmdCnv_SetLongMsg(SINT32 file_id, UINT8* pbMsg, UINT32 dSize)
 *
 *	Description
 *		Send MIDI Long message
 *	Param
 *		file_id		...	file ID
 *		pbMsg		...	pointer to the SysEx meassage
 *		dSize		...	Size of the message[byte]
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
static SINT32	MaRmdCnv_SetLongMsg(SINT32 sdFileID, UINT8* pbMsg, UINT32 dSize)
{
	SINT32	sdRet;
	UINT32	dBk, dPg, dOnKey;
	UINT32	dBl, dPc, dFlg;
	UINT32	dCh;
	UINT32	dKey, dAdr;
	SINT32	sdRom;
	UINT32	dWsize;
	UINT32	dWaveID;
	UINT8	bVoice[32];

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	MARMDCNV_DBGMSG(("MaRmdCnv_SetLongMsg[%p, %ld] \n", pbMsg, dSize));

	sdRet = MASMW_SUCCESS;
	if (dSize < 6) return (MASMW_SUCCESS);
	if (pbMsg[0] != 0xf0) return (MASMW_ERROR);
	if (pbMsg[dSize - 1] != 0xf7) return (MASMW_ERROR);
	
	if ((pbMsg[1] == 0x7e) &&
	    (pbMsg[2] == 0x7f) &&
	    (pbMsg[3] == 0x09) &&
	    ((pbMsg[4] == 0x01) || (pbMsg[4] == 0x02) || (pbMsg[4] == 0x03)) &&
	    (dSize == 6))
	{
		/* GM L1/L2 SystemOn/Off : 0xF0 0x7E 0x7F 0x09 (0x01/0x02/0x03) 0xF7 */
		return (SendGmOn());
	}

	if ((pbMsg[1] == 0x7f) &&
	    (pbMsg[2] == 0x7f) &&
	    (pbMsg[3] == 0x04) &&
	    (pbMsg[4] == 0x01) &&
	    (dSize == 8))
	{
		/* Universal MasterVolume : 0xF0 0x7F 0x7F 0x04 0x01 LL HH 0xF7 */
		gpRmdInfo->dMasterVolume = pbMsg[6] & 0x7f;
		SendExpression(0, gpRmdInfo->dExpressionTable[0]);
		for (dCh = 0; dCh < 16; dCh++) SendExpression(dCh, gpRmdInfo->dExpressionTable[dCh]);
		return (MASMW_SUCCESS);
	}
	
	if ((pbMsg[1] != 0x43) ||
	    (pbMsg[2] != 0x79) ||
	    (pbMsg[3] != 0x06) ||
	    (pbMsg[4] != 0x7f) ||
	    (dSize < 8)) return (MASMW_SUCCESS);

	switch (pbMsg[5])
	{
	case 0x00:		/* Max Gain : 0xF0 0x43 0x79 0x06 0x7F 0x00 VV 0xF7 */
		if (dSize != 8) return (MASMW_ERROR);
		gpRmdInfo->dMaxGain = (UINT8)(pbMsg[6] & 0x7f);
		SendMasterVolume(gpRmdInfo->dMaxGain);
		return (MASMW_SUCCESS);
	
	case 0x10:		/* UserEvent : 0xF0 0x43 0x79 0x06 0x7F 0x10 VV 0xF7 */
		if (dSize != 8) return (MASMW_ERROR);
		if (gpRmdInfo->pfnCallbackFunc == NULL) return (MASMW_ERROR);
		return (gpRmdInfo->pfnCallbackFunc((UINT8)(pbMsg[6] & 0x0f)));
	
	case 0x01:		/* Set Voice params : 0xF0 0x43 0x79 0x06 0x7F 0x01 BH BL PC KY FG ... 0xF7 */
		if (dSize < 31) return (MASMW_ERROR);
		dBl = (UINT32)(pbMsg[7] & 0x7f);
		dPc = (UINT32)(pbMsg[8] & 0x7f);
		dOnKey = (UINT32)(pbMsg[9] & 0x7f);
		dFlg = (UINT32)(pbMsg[10] + 1);

		switch (pbMsg[6])
		{
		case 0x7c:
			dBk = 1 + dBl;
			dPg = dPc;
			break;
		case 0x7d:
			dBk = (UINT32)(129 + dPc);
			dPg = dOnKey;
			break;
		default:
			return (MASMW_ERROR);
		}

		dAdr = gpRmdInfo->dRamBase + gpRmdInfo->dRamOffset;

		switch (dSize)
		{
		case 31:	/* WT */
			if (dFlg != 2) return (MASMW_ERROR);
			if ((gpRmdInfo->dRamOffset + 13) > gpRmdInfo->dRamSize) return (MASMW_ERROR);
			Decode7Enc(&pbMsg[11], 19, bVoice, 16);

			dKey = ((UINT32)bVoice[0] << 8) | (UINT32)bVoice[1];
			dWaveID = bVoice[15];

			if (dWaveID < 128)
			{
				if (dWaveID >= MAX_NUM_OF_WAVES)  return (MASMW_ERROR);
				if (gpRmdInfo->dWtAdrTable[dWaveID] == 0) return (MASMW_ERROR);
				bVoice[9] = (UINT8)((gpRmdInfo->dWtAdrTable[dWaveID] >> 8) & 0xff);
				bVoice[10] = (UINT8)(gpRmdInfo->dWtAdrTable[dWaveID] & 0xff);
			} else {
				sdRom = MaResMgr_GetDefWaveAddress((UINT8)(dWaveID & 0x7f));
				if (sdRom < 0) return (MASMW_ERROR);
				bVoice[9] = (UINT8)((sdRom >> 8) & 0xff);
				bVoice[10] = (UINT8)(sdRom & 0xff);
			}
			MaDevDrv_SendDirectRamData(dAdr, ENC_8BIT, &bVoice[2], 13);
			sdRet = MaSndDrv_SetVoice(gpRmdInfo->sdSeqID, dBk, dPg, dFlg, dKey, dAdr);
			gpRmdInfo->dRamOffset += 14;
			break;

		case 32:	/* 2-OP FM */
			if (dFlg != 1) return (MASMW_ERROR);
			if ((gpRmdInfo->dRamOffset + 16) > gpRmdInfo->dRamSize) return (MASMW_ERROR);
			Decode7Enc(&pbMsg[11], 20, bVoice, 17);
			dKey = bVoice[0];
			MaDevDrv_SendDirectRamData(dAdr, ENC_8BIT, &bVoice[1], 16);
			sdRet = MaSndDrv_SetVoice(gpRmdInfo->sdSeqID, dBk, dPg, dFlg, dKey, dAdr);
			gpRmdInfo->dRamOffset += 16;
			break;

		case 48:	/* 4-OP FM */
			if (dFlg != 1) return (MASMW_ERROR);
			if ((gpRmdInfo->dRamOffset + 31) > gpRmdInfo->dRamSize) return (MASMW_ERROR);
			Decode7Enc(&pbMsg[11], 36, bVoice, 31);
			dKey = bVoice[0];
			MaDevDrv_SendDirectRamData(dAdr, ENC_8BIT, &bVoice[1], 30);
			sdRet = MaSndDrv_SetVoice(gpRmdInfo->sdSeqID, dBk, dPg, dFlg, dKey, dAdr);
			gpRmdInfo->dRamOffset += 30;
			break;
		
		default:
			return (MASMW_ERROR);
		}
		break;

	case 0x03:		/* Set Wave */
		dAdr = gpRmdInfo->dRamBase + gpRmdInfo->dRamOffset;
		if (dSize < 10) return (MASMW_ERROR);
		dWsize = dSize - 9;
		dWsize = (dWsize >> 3) * 7 + (((dWsize & 0x07) == 0) ? 0 : ((dWsize & 0x07)-1));
		if ((gpRmdInfo->dRamOffset + dWsize) > gpRmdInfo->dRamSize) return (MASMW_ERROR);
		dPc = pbMsg[6];							/* wave ID */
		if (dPc > MAX_NUM_OF_WAVES) return (MASMW_ERROR);
		dFlg = (UINT8)(pbMsg[7] & 0x7f);
		sdRet = MaDevDrv_SendDirectRamData(dAdr, ENC_7BIT, &(pbMsg[8]), dSize - 9);
		gpRmdInfo->dRamOffset += ((dWsize + 1) >> 1) << 1;
		gpRmdInfo->dWtAdrTable[dPc] = dAdr;
		break;

	default:
		return (MASMW_SUCCESS);
	}

	return (sdRet);
}


/****************************************************************************
 *	MaRmdCnv_Control(SINT32 sdFileID, UINT32 dCtrlNum, void* pvPrm, void* pvExtArgs)
 *
 *	Description
 *		Special commands
 *	Param
 *		sdFileID		...	file ID
 *		dCtrlNum		...	Command ID
 *		pvPrm			...	param for the command
 *	Return
 *		0 : NoError, < 0 : Error
 ****************************************************************************/
SINT32	MaRmdCnv_Control(SINT32 sdFileID, UINT32 dCtrlNum, void* pvPrm, void* pvExtArgs)
{
	(void)pvExtArgs;

	MARMDCNV_DBGMSG(("MaRmdCnv_Control[%ld, %ld] \n", sdFileID, dCtrlNum));

	if ((sdFileID < 0) || (sdFileID != gpRmdInfo->sdFileID))
	{
		return (MASMW_ERROR);
	}

	switch(dCtrlNum)
	{
	case MASMW_SET_VOLUME :
		return (MaSndDrv_SetVolume(gpRmdInfo->sdSeqID, (UINT32)*((UINT8*)pvPrm), 7));

	case MASMW_SEND_MIDIMSG :
		if (gpRmdInfo->dOpend == 0) return (MASMW_ERROR);
		return (MaRmdCnv_SetShortMsg(sdFileID, *((UINT32*)pvPrm)));

	case MASMW_SEND_SYSEXMIDIMSG :
		if (gpRmdInfo->dOpend == 0) return (MASMW_ERROR);
		return (MaRmdCnv_SetLongMsg(sdFileID, ((PMASMW_MIDIMSG)pvPrm)->pbMsg, ((PMASMW_MIDIMSG)pvPrm)->dSize));

	case MASMW_GET_LENGTH :
		return (1000);					/* dummy length (1sec) */

	case MASMW_SET_REPEAT :
		return (MASMW_SUCCESS);			/* dummy func */

	case MASMW_GET_CONTROL_VAL:
		if (gpRmdInfo->dOpend == 0) return (MASMW_ERROR);
		switch (((PMASMW_GETCTL)pvPrm)->bControl)
		{
		case 0:	/* Master Volume */
			return (gpRmdInfo->dMasterVolume);
			
		case 1:	/* Channel Volume */
			if (((PMASMW_GETCTL)pvPrm)->bCh >= 16) return (MASMW_ERROR);
			return (gpRmdInfo->dChVolumeTable[((PMASMW_GETCTL)pvPrm)->bCh]);
		}
		return (MASMW_ERROR);
	
	case MASMW_SET_CB_INTERVAL: 
		if (gpRmdInfo->dOpend == 1) return (MASMW_ERROR);
		if ((*(UINT8*)pvPrm != 0) && (*(UINT8*)pvPrm != 20)) return (MASMW_ERROR);
		gpRmdInfo->dIntervalTime = (UINT32)*(UINT8*)pvPrm;
		return (MASMW_SUCCESS);
	}

	return (MASMW_ERROR);
}

