/**
 *	Author: 
 *		Robert.Chen
 *
 *	Module:
 *		Low level API wrapping Yamaha YMU765/762 chip sound middleware, interrupt driving
 *
 *  History:
 *		=== 2003/12/01   Robert.Chen ===	
 *		1)	Initiated.
 *
 *		=== 2004/05/08	 Robert.Chen ===
 *		1)	Adjusted volume setting sequence, replacing volume control after starting of
 *			playing a music (MI_PlayMusic_Start(), MI_PlayMusic_Resume()) to avoid pop noise
 *		2)  Removed EQ, HP and SP volume setting in MI_Melody_Initialize() to avoid pop noise
 *
 *		=== 2004/07/05	 Robert.Chen ===
 *		1) Add function to support play music data from FFS file, using static memory buffer
 *
 *		=== 2004/07/10	 Robert.Chen ===
 *		1) Add function to support play music data from FFS file, using dynamic memory buffer
 *		   (configuralbe in melody_cfg.h)	
 **/
 
#include "nucleus.h"
#include "typedefs.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "rvf_api.h"
#include "chipset.cfg"
#include "armio.h"
#include "mamachdep.h"
#include "madefs.h"

#include "rvm_use_id_list.h"
#include "melody_cfg.h"
#include "melody_struct_i.h"
#include "melody_api.h"
#include "melody_env.h"

#include "masound.h"					/* MA Sound Player API */
#include "madebug.h"

#include "ffs.h"
#include "mmiutilapi.h"


T_MELODY_CTRL_BLK	melodyCtrlBlk;

#if (MELODY_FFS_SUPPORT == 1)
  #if (MELODY_DATA_STATIC == 1)
	static UINT8 music_data_buffer[MELODY_DATA_BUF_SZ];
  #else
	static UINT8 *music_data_buffer = NULL;		
  #endif

  	static UINT8 music_info_buffer[MELODY_INFO_BUF_SZ];
	static T_FFS_FD melodyFd = -1;
#endif

extern SINT32 MaDevDrv_PowerManagement(UINT8 mode);
extern UINT8 spkLev2val[MI_MELODY_VOLUME_MAX+1];
extern UINT8 gbMelLP;

extern char *musicList[];

#if (MELODY_FFS_SUPPORT == 1)
/******************************************************************************
 *
 * Function:
 *		MI_LoadMusicByName
 *
 * Description:
 *		Loading a music file in FFS to the internal music data buffer.
 *
 * Input:
 *		musicName	string
 *		type		music type, if type equal to 0xFF means type not specified		
 *
 * Output:
 *		MI_MELODY_OK
 *		MI_MELODY_PARA_ERR
 *		MI_MELODY_FILE_NOT_EXIST
 *		MI_MELODY_FILE_SIZE_ZERO
 *		MI_MELODY_FILE_SIZE_LARGE
 *		MI_MELODY_FILE_READ_ERR
 *		   
 *****************************************************************************/

T_MELODY_RET MI_LoadMusicByName(const char *musicName, UINT8 type)
{
	T_FFS_RET ret;
	struct stat_s stat;
	UINT32 *pType, sz;
	
	MELODYTRACE_MSG(("MI_LoadMusicByName():name=\"%s\"",musicName));
	
	/* acquire file info */
	ret = ffs_stat(musicName, &stat);
	if(ret == EFFS_NOTFOUND || stat.type != OT_FILE)
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():acquire file info err!"));
		return MI_MELODY_FILE_NOT_EXIST;
	}

	if(stat.size == 0)
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():file size zero!"));
		return MI_MELODY_FILE_SIZE_ZERO;
	}

#if (MELODY_DATA_STATIC == 1)
	if(stat.size > (MELODY_DATA_BUF_SZ - 8))
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():file size large!"));
		return MI_MELODY_FILE_SIZE_LARGE;
	}
	else
#endif	
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():flen=%d", stat.size));
	}

	/* loading first 12 bytes file data to internal buffer */
	melodyFd = ffs_open(musicName, FFS_O_RDONLY);
	if(melodyFd < 0)
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():open file err,fd=%d", melodyFd));		
		return MI_MELODY_FILE_READ_ERR;
	}
	else
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():open file ok,fd=%d", melodyFd));		
	}	
	
	sz = ffs_read(melodyFd, (void*)&music_info_buffer[8], (MELODY_INFO_BUF_SZ - 8));
	if(sz != (MELODY_INFO_BUF_SZ - 8))
	{
		MELODYTRACE_MSG(("MI_LoadMusicByName():read file err,ret=%d", sz));		
		return MI_MELODY_FILE_READ_ERR;
	}
	
	if(type == 0xFF)	/* type not specified, we must detect music type */
	{
		ret = MI_DetectMusicType((const char*)&music_info_buffer[8], &type);
		if(ret != MI_MELODY_OK)
		{
			MELODYTRACE_MSG(("MI_LoadMusicByName():music type unknow,ret=%d", ret));	
			return MI_MELODY_FILE_TYPE_UNKNOW;		
		}	
	}
	
	pType 		= (UINT32*)music_info_buffer;
	*pType++	= type;
	*pType		= stat.size;
	return MI_MELODY_OK;
}

/******************************************************************************
 *
 * Function:
 *		MI_DetectMusicType
 *
 * Description:
 *		Detect the type of a music file pointer to by pData.
 *		At present, three types supported: WAV, MMF, MID
 *
 * Input:
 *		pData	pointer to music data buffer
 *		type	music type, available if return value is MI_MELODY_OK
 *
 * Output:
 *		MI_MELODY_OK
 *		MI_MELODY_PARA_ERR
 *		MI_MELODY_ERROR
 *		   
 *****************************************************************************/
T_MELODY_RET MI_DetectMusicType(const char *pData, UINT8 *type)
{	
	MELODYTRACE_MSG(("MI_DetectMusicType()"));
	
	if(pData == NULL || type == NULL)
	{
		MELODYTRACE_MSG(("MI_DetectMusicType():arg err!"));
		return MI_MELODY_PARA_ERR;
	}

	if(!memcmp(pData, "MMMD", 4))
		*type = MASMW_CNVID_MMF;
	else if(!memcmp(pData, "MThd", 4))
		*type = MASMW_CNVID_MID;
	else if(!memcmp(pData, "RIFF", 4))
		*type = MASMW_CNVID_WAV;
	else
	{
		MELODYTRACE_MSG(("MI_DetectMusicType():failed!,0x%x, 0x%x, 0x%x, 0x%x", *pData++, *pData++, *pData++, *pData++));
		return MI_MELODY_ERROR;
	}	

	return MI_MELODY_OK;	
}

/*** 
 * Close previously opened FFS file
 */
void MI_CloseMusicFile(void)
{
	MELODYTRACE_MSG(("MI_CloseMusicFile():fdi=%d", melodyFd));
	if(melodyFd >= 0)
	{
		ffs_close(melodyFd);
		melodyFd = -1;
	}
}

#endif //MELODY_FFS_SUPPORT
/******************************************************************************
 *
 * Function:
 *		MI_GetMusicInfoByID()
 *
 * Description:
 *		Get information about a music specific by its ID from music database, 
 *		the info returned such as music data start pointer, data lenght, music
 *      format.
 *
 * Input:
 *		musicID  within musicInfo struct
 *			
 * Output:
 *		musicInfo.
 *
 * History:
 *		2004/07/05	Robert.Chen add support to play music saved as FFS file.
 *		   
 *****************************************************************************/
T_MELODY_RET MI_GetMusicInfoByID(T_MUSIC_INFO * const musicInfo)
{
	T_MUSIC_INFO *pInf = musicInfo;		
	UINT32 *pTypeLen;
	
	MELODYTRACE_MSG(("MelFunc:MI_GetMusicInfoByID(),ID=%d.", musicInfo->mID));

#if (MELODY_FFS_SUPPORT == 1)	/* robert.chen, 2004-07-05 */
	if(pInf == NULL || (pInf->mID >= MI_MELODY_MUSICNUM_TOTAL && pInf->mID != MI_MELODY_MUSICID_FILE))
#else
	if(pInf == NULL || pInf->mID >= MI_MELODY_MUSICNUM_TOTAL)
#endif	
	{
		
		MELODYTRACE_MSG(("MelRet:error para: %d", pInf->mID));
		return MI_MELODY_PARA_ERR;
	}

#if (MELODY_FFS_SUPPORT == 1)	/* robert.chen, 2004-07-05 */
	if(pInf->mID == MI_MELODY_MUSICID_FILE)
		pTypeLen	= (UINT32*)music_info_buffer;
	else
		pTypeLen 	= (UINT32*)musicList[pInf->mID];	
#else
	pTypeLen 	= (UINT32*)musicList[pInf->mID];
#endif
	pInf->mType	= (UINT8)*pTypeLen++;
	pInf->mLen	= *pTypeLen++;
	pInf->mPtr	= (UINT8*)pTypeLen;
	return MI_MELODY_OK;	
}

/******************************************************************************
 *
 * Function:
 *		MI_Melody_RstCtrlBlk()
 *
 * Description:
 *		Reset music player control block data
 *
 * Input:
 *		none.
 *			
 * Output:
 *		none.
 *		   
 *****************************************************************************/
void MI_Melody_RstCtrlBlk(void)
{
#if (MELODY_FFS_SUPPORT == 1 && MELODY_DATA_STATIC == 0)
	int result;
#endif

	MELODYTRACE_MSG(("MelFunc:MI_Melody_RstCtrlBlk()"));
	melodyCtrlBlk.remainCnt = 0;
	melodyCtrlBlk.fileType  = 0;
	melodyCtrlBlk.curPos	= 0;
	melodyCtrlBlk.fileID	= -1;
	melodyCtrlBlk.musicID	= MI_MELODY_MUSICID_INVALID;

#if	(MELODY_FFS_SUPPORT == 1)
  #if (MELODY_DATA_STATIC == 0)
	if(music_data_buffer != NULL)
	{
		MI_FREE(music_data_buffer, result);
		music_data_buffer = NULL;
		
		if(result != EC_MMIMEM_OK)
		{
			MELODYTRACE_MSG(("MI_Melody_RstCtrlBlk():release mem err, ret=%d",result));		
		}
	}	
  #endif //MELODY_DATA_STATIC	
#endif  
	

  #ifdef MELODY_POWERSAVE
	NU_Control_Timer(&melodyCtrlBlk.pwrSaveTmr, NU_ENABLE_TIMER);
  #endif
}

/******************************************************************************
 *
 * Function:
 *		MI_Melody_CB()
 *
 * Description:
 *		Callback function invoked by MA sound middleware when event occurs, user
 *      concered process for these events can be executed here.
 *
 * Input:
 *		id    event ID.
 *			
 * Output:
 *		MASMW_SUCCESS.
 *		   
 *****************************************************************************/

signed long MI_Melody_CB(unsigned char id)
{
	SINT32 func_id;
	SINT32 file_id;	
	
	switch (id)
	{
	case MASMW_REPEAT:
		/* Repeat */
		/* remainCnt equal 0 means a loop play mode */
		if(melodyCtrlBlk.remainCnt != 0)
			melodyCtrlBlk.remainCnt--;
			
		MELODYTRACE_MSG(("MelCB:repeat once, remaincnt=%d", melodyCtrlBlk.remainCnt));
		break;

	case MASMW_END_OF_SEQUENCE:
		/* End of sequence */
		MELODYTRACE_MSG(("MelCB:play end!"));
		
		func_id = melodyCtrlBlk.funcID[melodyCtrlBlk.fileType];
		file_id = melodyCtrlBlk.fileID;
		
		MI_Melody_PostStop(func_id, file_id, MI_PostOnly);
		
		#if 0	/* Robert.Chen removed, 2004-04-08 */
		if(melodyCtrlBlk.vbrSta.ctrlSrc == MI_VBR_SRC_FORCED && melodyCtrlBlk.vbrSta.frcCtrl == MI_VBR_FORCED_OFF)
		{
			MI_Melody_HpAmpCtrl(0);
			gbMelLP = 1;	/* fix vibrator bug when deep sleep, 2004-04-08*/
		}
		#else
		MI_Melody_HpAmpCtrl(0, MI_PWRCTRL_MUSIC);
		#endif

		//gbMelLP = 1;	/* fix vibrator bug when deep sleep, 2004-04-08*/
		/* reset control block */
		MI_Melody_RstCtrlBlk();		
		break;

	default:
		/* User event */
		break;
	}

	return MASMW_SUCCESS;
}

/******************************************************************************
 *
 * Function:
 *		MI_Melody_ErrHdr()
 *
 * Description:
 *		Error recovery function to deal error occurs in each state, such as load
 *      state, open state, playing state.
 *
 * Input:
 *		funcID	function id
 *      fileID	file id 
 *			
 * Output:
 *		none.
 *		   
 *****************************************************************************/

void MI_Melody_ErrHdr(SINT32 funcID, SINT32	fileID)
{
	SINT32 state;
	
	state = MaSound_Control(funcID, fileID, MASMW_GET_STATE, NULL, NULL);

	MELODYTRACE_MSG(("MelErrHdr:err at state %d", state));
	
	switch (state)
	{
	case MASMW_STATE_PLAYING:
	case MASMW_STATE_PAUSE:
		MaSound_Stop(funcID, fileID, NULL);
		/* cotinue to below, don't break switch */

	case MASMW_STATE_READY:
	case MASMW_STATE_OPENED:
		MaSound_Close(funcID, fileID, NULL);
		/* cotinue to below, don't break switch */

	case MASMW_STATE_LOADED:
		MaSound_Unload(funcID, fileID, NULL);
		/* cotinue to below, don't break switch */
		

	case MASMW_STATE_IDLE:
		MaSound_Delete(funcID);
		/* cotinue to below, don't break switch */

	default:
		MI_Melody_RstCtrlBlk();
		break;
	}
}

#ifdef MELODY_POWERSAVE
/* power saving timer handler */
void MI_pwrSaveTmrProcess(UINT32 tmrID)
{
	SINT32 result;
	
	MELODYTRACE_MSG(("MelSta:pwr timeout!"));

	result = MaDevDrv_PowerManagement(1); /* power down */

	if(result != MASMW_SUCCESS)
	{
		MELODYTRACE_MSG(("MelRet:enter pwrdown failed,result=%d!", result));
	}
	else
	{
		MELODYTRACE_MSG(("MelSta:pwrdown entered!"));
		melodyCtrlBlk.pwrState = MI_pwrDown;
	}	
}


/******************************************************************************
 *
 * Function:
 *		MI_Melody_PwrRelase()
 *
 * Description:
 *		Release melody device from power down state
 *
 * Input:
 *		none.
 *			
 * Output:
 *		MI_MELODY_OK
 *      MI_MELODY_ERROR
 *		   
 *****************************************************************************/
T_MELODY_RET MI_Melody_PwrRelease(void)
{
	SINT32 result;

	MELODYTRACE_MSG(("MelFunc:MI_Melody_PwrRelease()"));
	
	/* check power state first */
  	if(melodyCtrlBlk.pwrState == MI_pwrRelease)  	
  	{
  		/* device in power release mode, power saving timer is enabled */
  		/* stop the timer */
  		NU_Control_Timer(&melodyCtrlBlk.pwrSaveTmr, NU_DISABLE_TIMER);
  		/* reset the timer expiretion duration for next run */
  		NU_Reset_Timer(&melodyCtrlBlk.pwrSaveTmr, 
  		              MI_pwrSaveTmrProcess,
  		              RVF_SECS_TO_TICKS(MI_POWERSAVE_SECS),
  		              0,
  		              NU_DISABLE_TIMER);  		
  	}
  	else
  	{
  		/* device in power down mode, must release this state before play music */
		MELODYTRACE_MSG(("MelSta:pwr down, release need."));
  		result = MaDevDrv_PowerManagement(2);	/* mode 2 release the device from power down mode */
		if(result != MASMW_SUCCESS)
		{
			MELODYTRACE_MSG(("MelRet:pwr release error,result=%d!", result));
			return MI_MELODY_ERROR;
		}
		else
		{
			MELODYTRACE_MSG(("MelRet:pwr released!"));
		}
  		melodyCtrlBlk.pwrState = MI_pwrRelease;
  	}

	/*
	 * check the power state to ensure that device in power release mode, avoid a timerout
	 * event occurs during the from the first time power state check to NU_Control_Timer()
	 * succuss, the timer handler, a HISR runs in HISR priority 2, can interrupt this function
	 * and set the device to power down, if we ignore that, a wrong power state will be assigned
	 * to the melody device
	 */
  	if(melodyCtrlBlk.pwrState == MI_pwrDown)
  	{
  		MELODYTRACE_MSG(("MelSta:pwr down by interrupt, release need."));
  		result = MaDevDrv_PowerManagement(2);	/* mode 2 release the device from power down mode */
		if(result != MASMW_SUCCESS)
		{
			MELODYTRACE_MSG(("MelRet:pwr release error,result=%d!", result));
			return MI_MELODY_ERROR;
		}
		else
		{
			MELODYTRACE_MSG(("MelRet:pwr released!"));
		}
  		melodyCtrlBlk.pwrState = MI_pwrRelease;
  	}

  	return MI_MELODY_OK;
}
#endif

/******************************************************************************
 *
 * Function:
 *		MI_PlayPrepare()
 *
 * Description:
 *		Get music data from music database, load data to the stream converter 
 *      corresponds to that music type, open data, and enter ready state.
 *
 * Input:
 *		musicID music index in music database
 *		count	playback count
 *
 * Output:
 *		none.
 *		   
 *****************************************************************************/

T_MELODY_RET MI_PlayPrepare(UINT8 musicID, UINT8 count)
{
	SINT32		 result;
	SINT32		 file_id;	/* file id */
	SINT32		 func_id;
	SINT32       volume;
	UINT16		 mode;
	T_MUSIC_INFO mInfo;
#if	(MELODY_FFS_SUPPORT == 1)
	T_FFS_SIZE	sz;
#endif
	

	MELODYTRACE_MSG(("MelFunc:MI_PlayPrepare(ID=%d, count=%d)", musicID, count));

	if(!melodyCtrlBlk.initialized)
	{
		MELODYTRACE_MSG(("MelRet:Not Init."));
		return MI_MELODY_NOT_INIT;
	}

	/* get the music detail information */
	mInfo.mID = musicID;
	result = MI_GetMusicInfoByID(&mInfo);
	if(result != MI_MELODY_OK)
	{
		MELODYTRACE_MSG(("MelRet:get music error,result=%d", result));
		return result;
	}
	else
	{
		MELODYTRACE_MSG(("MelRet:get music success, len = %d, type = %d", mInfo.mLen, mInfo.mType));
	}	

	/* check file and converter availability */
	if(mInfo.mLen == 0 || mInfo.mPtr == NULL)
	{		
		MELODYTRACE_MSG(("MelRet:len zero."));
		return MI_MELODY_FILE_SIZE_ZERO;
	}

		
	if(melodyCtrlBlk.funcID[mInfo.mType] < 0)
	{
		MELODYTRACE_MSG(("MelRet:type not support, funcID=%d!", func_id));
		return MI_MELODY_FILE_TYPE_ERR;
	}
	/* file and converter available */

  #ifdef MELODY_POWERSAVE
	result = MI_Melody_PwrRelease();
	if(result != MI_MELODY_OK)
		return MI_MELODY_ERROR;
  #endif
	
	/*
	 * a music is playing, we must stop it before initiate a new music if 
	 * there is some conflicts on the melody device, only SMAF/Audio can
	 * be played simultaneously with maximum 4 SMAF/Phrase tones, other 
	 * format music is player exclusively
	 */
	AI_MaskIT(0x02); /* 2004-04-01 added */
	
	if(melodyCtrlBlk.musicID != MI_MELODY_MUSICID_INVALID)
	{	
		func_id = melodyCtrlBlk.funcID[melodyCtrlBlk.fileType];
		file_id = melodyCtrlBlk.fileID;

		/* 2004-04-01 added, MUTE before stop playing */
	  #if (MIDI_STEREO == 1)
		MaSound_DeviceControl(MASMW_HP_VOLUME, 0, 0, 0);
	  #endif
		MaSound_DeviceControl(MASMW_EQ_VOLUME, 0, 0, 0);
		/* added end */
				
		result = MI_Melody_PostStop(func_id, file_id, MI_StopAndPost);
		if(result != MI_MELODY_OK)
		{
			MELODYTRACE_MSG(("MelRet:busy, stop err,result=%d!", result));			
		}
		/* reset control block */
		MI_Melody_RstCtrlBlk();		
	}
	else
	{
		MI_Melody_HpAmpCtrl(1, MI_PWRCTRL_MUSIC);
	}	
	AI_UnmaskIT(0x02);	/* 2004-04-01 added */

#if 0
	/* Set Volume for next time playing */
	MaSound_DeviceControl(MASMW_SP_VOLUME, spkLev2val[melodyCtrlBlk.curVolume], 0, 0);
	MaSound_DeviceControl(MASMW_HP_VOLUME, 0, spkLev2val[melodyCtrlBlk.curVolume], spkLev2val[melodyCtrlBlk.curVolume]);
	MaSound_DeviceControl(MASMW_EQ_VOLUME, 31, 0, 0);
#endif 

	/*
	 * prepare to play the new music returned by 'mInfo' struct
	 */
	 
	/* get created function ID corresponds to the requested music type */
	func_id = melodyCtrlBlk.funcID[mInfo.mType];

	/* decide function open mode */
	switch(mInfo.mType)
	{
	case MASMW_CNVID_RMD:
		mode = 0;
		break;

	case MASMW_CNVID_WAV:
		mode = 0;
		break;
		
	case MASMW_CNVID_AUD:
		mode = 1;
		break;
		
	case MASMW_CNVID_MMF:
	case MASMW_CNVID_MID:
	default:
		mode = 0;
		break;	
	}
	
	/* Load the file */
#if	(MELODY_FFS_SUPPORT == 1)

  #if (MELODY_DATA_STATIC == 0)
	if(music_data_buffer != NULL)
	{
		MI_FREE(music_data_buffer, result);
		music_data_buffer = NULL;
		
		if(result != EC_MMIMEM_OK)
		{
			MELODYTRACE_MSG(("MI_Melody_RstCtrlBlk():release mem err, ret=%d",result));		
		}
	}	
  #endif //MELODY_DATA_STATIC	

	/* allocate memory for FFS file */
	if(musicID == MI_MELODY_MUSICID_FILE)
	{
	  #if (MELODY_DATA_STATIC == 0) /* use dynamic memory */	
	  	MI_ALLOC(music_data_buffer, mInfo.mLen, sz);
	  	if(sz != EC_MMIMEM_OK)
	  	{
	  		MELODYTRACE_MSG(("MI_PlayPrepare():alloc dynmem failed, ret=%d", sz));
			music_data_buffer = NULL;
			return MI_MELODY_FILE_READ_ERR;
	  	}
	  #endif
		
		ffs_seek(melodyFd, 0, FFS_SEEK_SET);
		sz = ffs_read(melodyFd, (void*)music_data_buffer, mInfo.mLen);
	  
		if(sz != mInfo.mLen)
		{
			MELODYTRACE_MSG(("MI_PlayPrepare():read FFS file err,len=%d", sz));
			MI_Melody_RstCtrlBlk();
			return MI_MELODY_FILE_READ_ERR;
		}
		else
		{
			mInfo.mPtr = music_data_buffer;
		}		
	}
#endif

	file_id = MaSound_Load(func_id, mInfo.mPtr, mInfo.mLen, 1, MI_Melody_CB, NULL);
	if (file_id < MASMW_SUCCESS)
	{
		MELODYTRACE_MSG(("MelRet:Load file err%d", file_id));
		MI_Melody_ErrHdr(func_id, file_id);
		return MI_MELODY_ERROR;
	}

	/* Open the file */
	result = MaSound_Open(func_id, file_id, mode, NULL);
	if(result != MASMW_SUCCESS)
	{
		MELODYTRACE_MSG(("MelRet:Open file err%d", result));
		MI_Melody_ErrHdr(func_id, file_id);
		return MI_MELODY_ERROR;
	}
	
	/* Set volume */
	volume = 100; /* 0~127, default is 100 */
	result = MaSound_Control(func_id, file_id, MASMW_SET_VOLUME, &volume, NULL);
	if(result != MASMW_SUCCESS)
	{
		MI_Melody_ErrHdr(func_id, file_id);
		return MI_MELODY_ERROR;
	}

	/* Standby for playing */
	result = MaSound_Standby(func_id, file_id, NULL);
	if(result != MASMW_SUCCESS)
	{
		MI_Melody_ErrHdr(func_id, file_id);
		return MI_MELODY_ERROR;
	}	

	melodyCtrlBlk.musicID	= musicID;
	melodyCtrlBlk.remainCnt = count;
	melodyCtrlBlk.curPos	= 0;
	melodyCtrlBlk.fileID	= file_id;
	melodyCtrlBlk.fileType	= mInfo.mType;

	return MI_MELODY_OK;
}

/******************************************************************************
 *
 * Function:
 *		MI_Melody_PostStop()
 *
 * Description:
 *		Process after stop music playing.
 *
 * Input:
 *		func_id function ID
 *      file_id file ID
 *      mode  MI_PostOnly
 *            MI_StopAndPost
 *
 * Output:
 *		MI_MELODY_OK
 *      MI_MELODY_ERROR
 *		   
 *****************************************************************************/
T_MELODY_RET MI_Melody_PostStop(SINT32 funcID, SINT32 fileID, T_POSTSTOP mode)
{
	int result;

	MELODYTRACE_MSG(("MelFunc:MI_Melody_PostStop(), mode = %d", mode));
	
	switch(mode)
	{
	case MI_StopAndPost:
		result = MaSound_Stop(funcID, fileID, NULL);
		if(result != MASMW_SUCCESS)
		{
			MI_Melody_ErrHdr(funcID, fileID);
			return(MI_MELODY_ERROR);
		}
		/* following process need, don't break */
	case MI_PostOnly:
		result = MaSound_Close(funcID, fileID, NULL);
		if(result != MASMW_SUCCESS)
		{
			MI_Melody_ErrHdr(funcID, fileID);
			return(MI_MELODY_ERROR);
		}

		result = MaSound_Unload(funcID, fileID, NULL);
		if(result != MASMW_SUCCESS)
		{
			MI_Melody_ErrHdr(funcID, fileID);
			return(MI_MELODY_ERROR);
		}
	default:
		break;
	}

	return MI_MELODY_OK;
}

void MI_Melody_Reset(void)
{
	volatile int loop = 0, dummy = 0;

#if (MIDI_STEREO == 1)
	AI_EnableBit(BIT_EXTAMP_IO_SEL);
	AI_ConfigBitAsOutput(PIN_EXTAMP_CTRL);
#elif (MIDI_STEREO == 0)
	AI_ConfigBitAsInput(PIN_EXTAMP_CTRL);
#endif
	
	/* Set SIM_PWCTRL as GPIO5 for YMU765 RST */
	AI_EnableBit(1);
	AI_ConfigBitAsOutput(5);

	/* Reset YMU765 */
	AI_SetBit(5);
	for(loop = 0; loop < 1000; loop++)
		dummy ++;

	/* reset low pulse, duration at least 100us */
	AI_ResetBit(5);
	for(loop = 0; loop < 65535; loop++)
		dummy++;
	
	AI_SetBit(5);
	for(loop = 0; loop < 1000; loop++)
		dummy++;
}

/******************************************************************************
 *
 * Function:
 *		MI_Melody_HpAmpCtrl()
 *
 * Description:
 *		Perform melody chip power management at each states.
 *
 * Input:
 *		AmpOn  --- 1 for switch on, 0 for switch off
 *		reason --- MI_PWRCTRL_VBR for vibrator
 *				   MI_PWRCTRL_MUSIC for music
 *				   MI_PWRCTRL_ALL for all 
 *
 * Output:
 *		none.
 *		   
 *****************************************************************************/

void MI_Melody_HpAmpCtrl(UINT8 AmpOn, UINT8 reason)
{
	SINT32 result;
	static UINT8 pwrMode = 1; /* power mode: 2 on, 1 down */
	UINT8 prePwrMode; 
  
	MELODYTRACE_MSG(("MelFunc:MI_Melody_HpAmpCtrl(%d, %d)", AmpOn, reason));
	
	#if 0	/* Robert.Chen removed, 2004-04-08 */
	if(AmpOn)	/* enable amplifier */
	{		
		result = MaDevDrv_PowerManagement(2);	/* release from power down mode */
		if(result != MASMW_SUCCESS)
		{
			MELODYTRACE_MSG(("MelRet:Release power down error = %d", result));
		}
		
		MaSound_DeviceControl(MASMW_SP_VOLUME, spkLev2val[melodyCtrlBlk.curVolume], 0, 0);
		MaSound_DeviceControl(MASMW_HP_VOLUME, 0, spkLev2val[melodyCtrlBlk.curVolume], spkLev2val[melodyCtrlBlk.curVolume]);
		MaSound_DeviceControl(MASMW_EQ_VOLUME, 31, 0, 0);
		
		AI_SetBit(6);	/* enable external headphone amplifier */
	}
	else	/* disable amplifier */
	{
		AI_ResetBit(6);
		
		result = MaDevDrv_PowerManagement(1);	/* set to power down mode */
		if(result != MASMW_SUCCESS)
		{
			MELODYTRACE_MSG(("MelRet:Set to power down error = %d", result));
		}

		MaSound_DeviceControl(MASMW_HP_VOLUME, 0, 0, 0);
		MaSound_DeviceControl(MASMW_SP_VOLUME, 0, 0, 0);
		MaSound_DeviceControl(MASMW_EQ_VOLUME, 0, 0, 0);
	}
	#else
	prePwrMode = pwrMode;
	if(AmpOn)	/* power release part or all of midi chip components */
	{	
		if(pwrMode == 1)
		{
			#if (MIDI_DEVICE == 1)
			result = MaDevDrv_PowerManagement(3);	/* release from power down mode */
			#endif
			#if (MIDI_DEVICE == 2)
			result = MaDevDrv_PowerManagement(2);	/* release from power down mode */
			#endif
			
			if(result == MASMW_SUCCESS)
			{
				gbMelLP	= 0;		/* deep sleep disabled */ 				
				pwrMode	= 2;		/* power released mode */
			}
			else
			{
				MELODYTRACE_MSG(("MelRet:Release power down error = %d", result));
			}
		}		

		/* if power release for playing music, open external audio PA, set volume */
		if(reason == MI_PWRCTRL_MUSIC || reason == MI_PWRCTRL_ALL)
		{
			if(prePwrMode == 2 && melodyCtrlBlk.musicID == MI_MELODY_MUSICID_INVALID)	/* previously power on already, only need to set analog part */
			{
				#if (MIDI_STEREO == 1)
				MaSound_DeviceControl(MASMW_PWM_ANALOG, 0x0, 0, 0);	/* power release all analog */
				#else
				MaSound_DeviceControl(MASMW_PWM_ANALOG, 0x30, 0, 0);	/* power release all analog except HP */
				#endif
				
				MaSound_DeviceControl(MASMW_PWM_DIGITAL, 0x0, 0, 0);	/* power release internal ROM */
			}			
		#if 0
			MaSound_DeviceControl(MASMW_SP_VOLUME, spkLev2val[melodyCtrlBlk.curVolume], 0, 0);
			MaSound_DeviceControl(MASMW_HP_VOLUME, 0, spkLev2val[melodyCtrlBlk.curVolume], spkLev2val[melodyCtrlBlk.curVolume]);
			MaSound_DeviceControl(MASMW_EQ_VOLUME, 31, 0, 0);
		#endif
		
			#if (MIDI_STEREO == 1)
			AI_SetBit(PIN_EXTAMP_CTRL);	/* enable external headphone amplifier */			
			#endif
		}

		/* if power release for vibrator control, only need digital part powered */
		if(reason == MI_PWRCTRL_VBR && melodyCtrlBlk.musicID == MI_MELODY_MUSICID_INVALID)
		{
			MaSound_DeviceControl(MASMW_PWM_ANALOG, 0x3F, 0, 0);	/* power down all analog */
			MaSound_DeviceControl(MASMW_PWM_DIGITAL, 0x08, 0, 0);	/* power down internal ROM */
		}
	}
	else	/* power down part or all the midi chip components */
	{
		switch(reason)
		{
		case MI_PWRCTRL_MUSIC:
		  #if (MIDI_STEREO == 1)
			AI_ResetBit(PIN_EXTAMP_CTRL);
		  #endif

		  #if (MIDI_STEREO == 1)
			MaSound_DeviceControl(MASMW_HP_VOLUME, 0, 0, 0);
		  #endif
		  
			MaSound_DeviceControl(MASMW_SP_VOLUME, 0, 0, 0);		  	
			MaSound_DeviceControl(MASMW_EQ_VOLUME, 0, 0, 0);

			MaSound_DeviceControl(MASMW_PWM_ANALOG, 0x3F, 0, 0);	/* power down all analog */
			MaSound_DeviceControl(MASMW_PWM_DIGITAL, 0x08, 0, 0);	/* power down internal ROM */			

			if(melodyCtrlBlk.vbrSta.ctrlSrc == MI_VBR_SRC_FORCED && melodyCtrlBlk.vbrSta.frcCtrl == MI_VBR_FORCED_ON)
			{
			/* vibrator is working, only power down analog part */
				return;
			}
			break;
			
		case MI_PWRCTRL_VBR:		
			if(melodyCtrlBlk.musicID != MI_MELODY_MUSICID_INVALID)
			{
			/* music is playing, only power down vibrator part, do nothing */
				return;
			}
			break;
			
		case MI_PWRCTRL_ALL:
		default:
			/* all part isn't working, power down the chip */
			if(melodyCtrlBlk.musicID != MI_MELODY_MUSICID_INVALID)
			{
			  #if (MIDI_STEREO == 1)
				AI_ResetBit(PIN_EXTAMP_CTRL);		
			  #endif	

			  #if (MIDI_STEREO == 1) 	
				MaSound_DeviceControl(MASMW_HP_VOLUME, 0, 0, 0);
			  #endif
				MaSound_DeviceControl(MASMW_SP_VOLUME, 0, 0, 0);			  	
				MaSound_DeviceControl(MASMW_EQ_VOLUME, 0, 0, 0);
			}
			break;
		}

		/* execute to here, power down the chip */
	  #if (MIDI_DEVICE == 1)
	  	result = MaDevDrv_PowerManagement(2);	/* enter power down mode */ 
	  #endif
	  #if (MIDI_DEVICE == 2)
		result = MaDevDrv_PowerManagement(1);	/* enter power down mode */
	  #endif	
		if(result == MASMW_SUCCESS)
		{
			gbMelLP	= 1;		/* deep sleep enabled */ 				
			pwrMode	= 1;		/* power down mode */
		}
		else
		{
			MELODYTRACE_MSG(("MelRet:Release power down error = %d", result));
		}
	}	
	#endif
}

#if 0
void Melody_Test(void)
{
	SINT32 result;
	int i, tmrcnt = 0;
	
	if((result = MI_Melody_Initialize(MelInitNormal)) != MI_MELODY_OK)
	{
		MELODYTRACE_MSG(("MelRet:Test init err, result=%d!", result));
		return;
	}
	else
	{
		MELODYTRACE_MSG(("MelRet:Test init ok."));
	}	
	
	for(i = 0; i < 12; i++)
	{		
		MELODYTRACE_MSG(("Melody Play:###musicID%d###", i));
		
		result = MI_PlayMusic_Start(i, 1);			
		if(result != MI_MELODY_OK)
			MELODYTRACE_MSG(("MelRet:Test start err, m%d,r%d,", i, result));
			
		while(melodyCtrlBlk.musicID != MI_MELODY_MUSICID_INVALID)
		{
			NU_Sleep(RVF_SECS_TO_TICKS(1));
			if(i == 10)
			{
				if(++tmrcnt > 10)
				{
					result = MI_PlayMusic_Stop();
					if(result != MI_MELODY_OK)
						MELODYTRACE_MSG(("MelRet:stop music in Melody_Test() error,result=%d", result));
					break;
				}	
			}
		}		
		NU_Sleep(RVF_SECS_TO_TICKS(2));
	}
	
	/*
	for(i = 0; i < 5; i++)
	{
		result = MI_PlayMusic_Start(4, 1);
		if(result != MI_MELODY_OK)
			MELODYTRACE_MSG(("MelRet:Test start err, m%d,r%d,", i, result));
		NU_Sleep(RVF_SECS_TO_TICKS(3));
	}
	*/
	/*
	MELODYTRACE_MSG(("Melody Play:m1"));
	result = MI_PlayMusic_Start(3, 1);
	if(result != MI_MELODY_OK)
		MELODYTRACE_MSG(("MelRet:Test start err, m%d,r%d,", 3, result));
			
	while(melodyCtrlBlk.musicID != MI_MELODY_MUSICID_INVALID)
		{NU_Sleep(RVF_SECS_TO_TICKS(1));};
	NU_Sleep(RVF_SECS_TO_TICKS(12));

	MELODYTRACE_MSG(("Melody Play:m2"));
	result = MI_PlayMusic_Start(2, 3);
	if(result != MI_MELODY_OK)
		MELODYTRACE_MSG(("MelRet:Test start err, m%d,r%d,", 2, result));
			
	while(melodyCtrlBlk.musicID != MI_MELODY_MUSICID_INVALID)
		{NU_Sleep(RVF_SECS_TO_TICKS(1));};
	NU_Sleep(RVF_SECS_TO_TICKS(6));

	MELODYTRACE_MSG(("Melody Play:m4"));
	result = MI_PlayMusic_Start(4, 3);
	if(result != MI_MELODY_OK)
		MELODYTRACE_MSG(("MelRet:Test start err, m%d,r%d,", 4, result));
	*/
	
}
#endif

#if 0
/****************************************************************************
 *	phrase_cbfunc
 *
 *	Description:
 *			callback function
 *	Argument:
 *			id			id (0..127)
 *	Return:
 *			0			success
 *
 ****************************************************************************/
static void phrase_cbfunc( struct event * eve )
{
	switch ( eve->mode )
	{
	case -1:
		/* End of play */
		_flag = 1;
		break;

	case -2:
		/* End of loop play */
		_flag = 2;
		break;

	default:
		/* User event */
		break;
	}
}
/****************************************************************************
 *	phrase_error_handler
 *
 *	Description:
 *			error handler
 *	Argument:
 *			ch			sequence number
 *	Return:
 *			None
 *
 ****************************************************************************/
static void phrase_error_handler
(
	signed long ch					/* channel number */
)
{
	int state;

	state = Phrase_GetStatus( ch );

	switch ( state & 0x0F )
	{
	case ID_STATUS_PAUSE:			/* pause */
	case ID_STATUS_PLAY:			/* play */
		Phrase_Stop( ch );
		/* cotinue to below */

	case ID_STATUS_READY:			/* ready */
		Phrase_RemoveData( ch );
		/* cotinue to below */

	case ID_STATUS_NODATA:			/* no data */
		Phrase_Terminate();
		/* cotinue to below */

	default:
		break;
	}
}
/****************************************************************************
 *	phrase_main
 *
 *	Description:
 *			Sample of playing the SMAF/Phrase files using Phrase Wrapper API.
 *	Argument:
 *			file_ptr	pointer to the SMAF/Phrase file
 *			file_size	size of the SMAF/Phrase file
 *	Return:
 *			0		success
 *			-1		error
 *
 ****************************************************************************/
signed long phrase_main
(
	unsigned char * file_ptr,			/* pointer to the SMAF/Phrase file */
	unsigned long 	file_size			/* size of the SMAF/Phrase file */
)
{
	int	ch;								/* channel number */
	signed long	result;					/* result of function */

	_flag = 0;

	/* Initialize the MA Sound Middleware */
	result = MaSound_Initialize();
	if ( result < MASMW_SUCCESS )
	{
		return MASMW_ERROR;
	}

	/* Device Control */
	result = MaSound_DeviceControl( MASMW_EQ_VOLUME, 31, 0, 0 );
	if ( result < MASMW_SUCCESS )
	{
		return MASMW_ERROR;
	}

	result = MaSound_DeviceControl( MASMW_SP_VOLUME, 31, 0, 0 );
	if ( result < MASMW_SUCCESS )
	{
		return MASMW_ERROR;
	}

	result = MaSound_DeviceControl( MASMW_HP_VOLUME, 0, 31, 31 );
	if ( result < MASMW_SUCCESS )
	{
		return MASMW_ERROR;
	}

	/* Initialize the SMAF/Phrase I/F */
	result = Phrase_Initialize();
	if ( result < MASMW_SUCCESS )
	{
		return MASMW_ERROR;
	}

	/* Set the callback */
	result = Phrase_SetEvHandler( phrase_cbfunc );
	if ( result < MASMW_SUCCESS )
	{
		phrase_error_handler( 0 );
		return MASMW_ERROR;
	}

	/* Use sequence number is 0 */
	ch = 0;

	/* Specify the SMAF/Phrase file for play */
	result = Phrase_SetData( ch, file_ptr, file_size, 1 );
	if ( result < MASMW_SUCCESS )
	{
		phrase_error_handler( ch );
		return MASMW_ERROR;
	}

	/* Play the SMAF/Phrase file */
	result = Phrase_Play( ch, 1 );
	if ( result < MASMW_SUCCESS )
	{
		phrase_error_handler( ch );
		return MASMW_ERROR;
	}

	while ( _flag == 0 )
		;

	/* Remove the SMAF/Phrase file */
	result = Phrase_RemoveData( ch );
	if ( result < MASMW_SUCCESS )
	{
		phrase_error_handler( ch );
		return MASMW_ERROR;
	}

	/* Terminate the SMAF/Phrase I/F */
	result = Phrase_Terminate();
	if ( result < MASMW_SUCCESS )
	{
		return MASMW_ERROR;
	}

	return MASMW_SUCCESS;
}
#endif

