//=============================================================================
//	HAL.C
//-----------------------------------------------------------------------------
//	The routines in this file comprise the core of the LCD controller miniHAL.
//-----------------------------------------------------------------------------
//
//  Copyright (c) 2003 Epson Research and Development, Inc.
//  All Rights Reserved.
//
//=============================================================================



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Robert.Chen add, 2004-06-07 */
#include "rvf_api.h"
#include "sys_types.h"
#include "camera_cfg.h"
/* end */

#include "hal_private.h"
#include "hal_i2c.h"
#include "hal_indirect.h"
//#include "hal_platform.h"


extern void wait_ARM_cycles(SYS_UWORD32 cpt_loop);
extern SYS_UWORD32 convert_nanosec_to_cycles(SYS_UWORD32 time);

void halDelayUS_forInit( UInt32 microseconds );
//===========================================================================
// PRIVATE GLOBAL VARIABLES (local to HAL only)
//===========================================================================

static DATA_STRUCT gHD;							
//{
	//members are default initialized to zero.
//};// Global HAL private data storage

//===========================================================================
// PUBLIC GLOBAL VARIABLES
//===========================================================================

const DATA_STRUCT* const gpHalData = &gHD;

//===========================================================================
// PRIVATE (STATIC) FUNCTION PROTOTYPES (local to HAL only)
//===========================================================================

static void		InitRegisters( void );
static void		ProgramParallelPanels( void );
static void		ClearVideoMemory( void );
static UInt16		CalculateHalInfoCRC( const PHAL_STRUCT pHalData );




//===========================================================================
// PUBLIC FUNCTIONS EXPORTED IN HAL.H
//===========================================================================



//---------------------------------------------------------------------------
// PUBLIC FUNCTION: halAcquireController()
//---------------------------------------------------------------------------
Boolean halAcquireController( void )
{
	static Boolean fFirstTime = FALSE;

	memset(&gHD,0,sizeof(gHD));
	
	gHD.nErrorCode = ERR_NONE;

	// Init, one time, default state of debug register writes flag.

	if ( !fFirstTime )
	{
		fFirstTime = TRUE;
	}
	
	// Use the configured base address to retrieve the virtual address, if any.  On
	// non-Windows platforms, the base address will be copied to the virtual address.

	if ( gHD.BaseAddress == 0 )	// Only do this once, so we are re-entrant.
	{
		gHD.BaseAddress=	HalInfo.dwBaseAddress;
		gHD.RegisterAddress = gHD.BaseAddress + HalInfo.dwRegisterOffset;
		gHD.MemoryAddress   = gHD.BaseAddress + HalInfo.dwMemoryOffset;

		//if ( HalInfo.dwMemoryOffset == HalInfo.dwRegisterOffset )
		//{
		//	gHD.nErrorCode=ERR_SAME_REGMEM;;
		//	return FALSE;
		//}
	}

	halpIndirectInit( gHD.BaseAddress );

	return TRUE;
}



//---------------------------------------------------------------------------
// PUBLIC FUNCTION: halInitController()
//---------------------------------------------------------------------------
Boolean halInitController( UInt32 Flags )
{
	UINT8 i,j;
	// If the program has not been CFGed then we CANNOT init the controller.
	// Set and error code and return FALSE.

	gHD.nErrorCode = ERR_NONE;

	// If the caller specified ANY of the fJUST_xxx flags, then replace ALL the 
	// fDONT_xxx flags with the inverted logic of the fJUST_xxx flags.
	if ( (Flags & (fJUST_RESET|fJUST_INIT_I2C|fJUST_INIT_REGS|fJUST_PROG_LCDS|fJUST_INIT_LUT|fJUST_CLEAR_MEM|fJUST_CHECK_CRC)) )
		Flags = (~Flags & (fJUST_RESET|fJUST_INIT_I2C|fJUST_INIT_REGS|fJUST_PROG_LCDS|fJUST_INIT_LUT|fJUST_CLEAR_MEM|fJUST_CHECK_CRC)) >> 16;

	// Fail if halAcquireController() was not called, or if it failed.
	if ( gHD.BaseAddress == 0 )
	{
		gHD.nErrorCode = ERR_NOT_ACQUIRED;
		return FALSE;
	}

	// Should we reset the controller?
	if ( !(Flags & fDONT_RESET))
	{
		HAL_WRITE_REG(REG0008_SOFTRESET, 0);
		halDelayUS( HAL_DELAY_RESET );
	}

	
//	MI_dbgPrintf("reset is ok");
	
	// Init the registers - unless the caller has requested us not to
	if ( !(Flags & fDONT_INIT_REGS) )
	{
		// Make sure that the LCD Interface Auto Transfer Enable is turned off, or else this function will hang.
		if ( HAL_READ_REG(0x38) & 0x1 )
		{
			HAL_WRITE_REG( REG003C_LCDXFER, 0 );
			while ( HAL_READ_REG(0x38) & 0x1 )
				continue;
		}

		InitRegisters();
	}
//	MI_dbgPrintf("init is ok");

	// Program the parallel LCD panels - unless:
	//	a) the caller has requested us not to
	//	b) CFG has requested us not to
	if ( !(Flags & fDONT_PROG_LCDS) && ((HalInfo.LCD1Regs[0] != HAL_LCD_ENDOFTABLE) || (HalInfo.LCD2Regs[0] != HAL_LCD_ENDOFTABLE) ))
	{
		ProgramParallelPanels();
	}
	
//	MI_dbgPrintf("lcd_init is ok");
//		MI_dbgPrintf("initlcd ok");
	// Init the LUT - unless the caller has requested us not to
	//if ( !(Flags & fDONT_INIT_LUT) )
	//{
	//	halInitLUT( FALSE );
	//}

	// Init the I2C bus - unless the caller or config has requested us not to.
	// Programming Epson Camera may corrupt VRAM, so clear it AFTER this.
	if ( !(Flags & fDONT_INIT_I2C) )
	{
		//halWriteReg8(REG0104_CMMODE0, 0x01);
		if ( !halpInitI2C(-1) )
			gHD.nErrorCode = ERR_BAD_I2C_INIT;
	}
//	MI_dbgPrintf("initi2c ok");
	// Clear video memory - unless the caller has requested us not to.
	if ( !(Flags & fDONT_CLEAR_MEM) )
	{
		ClearVideoMemory();
	}
//	MI_dbgPrintf("ok");
	return ( gHD.nErrorCode == ERR_NONE );
}

//-----------------------------------------------------------------------------
// PUBLIC FUNCTION: halGetLastError()
//-----------------------------------------------------------------------------
int halGetLastError()
{
	return gHD.nErrorCode;
}

//---------------------------------------------------------------------------
// PUBLIC FUNCTION: halInitLUT()
//---------------------------------------------------------------------------
void halInitLUT( Boolean fUseGradientPalette )
{
#if 0
	UIntReg Reg4C;
	int bpp, LutBypass;
	int i;
	RGB_UNION CR;
	UInt8 *pLut;

	Reg4C = HAL_READ_REG(REG004C_DISPSETTING);
	bpp       = (0 == (Reg4C & 0x20))? 16 : 8;
	LutBypass = (0 == (Reg4C & 0x10)) ? TRUE : FALSE;

	if (8 == bpp)
	{
		// At 8 BPP there are three configurations we may be setting the LUT
		//	- If the LUT is bypassed configure for 3-3-2 bypass
		//	- If a gradient LUT is specified, configure for a gradient LUT
		//	- Else use the static LUT
		if (LutBypass)
		{
			// 8 bpp LUT bypass - fill LUT with 3-3-2 data
			for (i = 0; i < 8; i++)
			{
				// Only the first 4 Blue elements are used. the
				// next 4 values are discarded by the controller.
				CR.rgb.Blue  = (UInt8)((i * 0xFF) / 3);
				CR.rgb.Red   = (UInt8)((i * 0xFF) / 7);
				CR.rgb.Green = (UInt8)((i * 0xFF) / 7);
				HAL_WRITE_LUT(i, CR.Value32);
			}
		}
		else if (fUseGradientPalette)
		{
			// 8 bpp  - gradient filled LUT
			for (i = 0; i < 256; i++)
			{
				UInt8 c = (UInt8)(i << 2);

				CR.Value32 = 0x000000;
				switch (i/64)
				{
				  case 0:	CR.rgb.Red   = c;								break;
				  case 1:	CR.rgb.Green = c;								break;
				  case 2:	CR.rgb.Blue  = c;								break;
				  case 3:	CR.rgb.Red = CR.rgb.Green = CR.rgb.Blue = c;	break;
				}

				HAL_WRITE_LUT(i, CR.Value32);
			}
		}
		else
		{
			pLut = LutInfo.lut1;

			// 8 bpp - static LUT table
			for (i = 0; i < 256; i++)
			{
				CR.rgb.Red   = *pLut++;
				CR.rgb.Green = *pLut++;
				CR.rgb.Blue  = *pLut++;
				HAL_WRITE_LUT(i, CR.Value32);
			}
		}
	}
	else	// we are at 16 bpp
	{
		// 16 BPP is selected - configure for 5-6-5
		for (i = 0; i < 64; i++)
		{
			if (i < 32)
				CR.rgb.Red = CR.rgb.Blue = (UInt8)((i * 0xFF) / 31);
			else
				CR.rgb.Red = CR.rgb.Blue = 0xFF;

			CR.rgb.Green = (UInt8)((i * 0xFF) / 63);
			HAL_WRITE_LUT(i, CR.Value32);
		}
	}
#endif
}

//===========================================================================
// COMMON FUNCTIONS exported in HALAPI.H
//===========================================================================




//-----------------------------------------------------------------------------
// COMMON FUNCTION: halReadDisplay8()
// COMMON FUNCTION: halReadDisplay16()
// COMMON FUNCTION: halReadDisplay32()
//-----------------------------------------------------------------------------
UInt8 halReadDisplay8( UInt32 Offset )
{
	UInt8 Data;
	Data=halpParReadDisplay8( Offset );
	return Data;
}

UInt16 halReadDisplay16( UInt32 Offset )
{
	UInt8 Data;
	Data=halpParReadDisplay16( Offset );
	return Data;
}

UInt32 halReadDisplay32( UInt32 Offset )
{
	UInt8 Data;
	Data=halpParReadDisplay32( Offset );
	return Data;
}



//-----------------------------------------------------------------------------
// COMMON FUNCTION: halWriteDisplay8()
// COMMON FUNCTION: halWriteDisplay16()
// COMMON FUNCTION: halWriteDisplay32()
//-----------------------------------------------------------------------------
void halWriteDisplay8( UInt32 Offset, UInt8 Value, UInt32 Count )
{
	halpParWriteDisplay8( Offset, Value, Count );
}

void halWriteDisplay16( UInt32 Offset, UInt16 Value, UInt32 Count )
{
	halpParWriteDisplay16( Offset, Value, Count );
}

void halWriteDisplay32( UInt32 Offset, UInt32 Value, UInt32 Count )
{
	halpParWriteDisplay32( Offset, Value, Count );
}



//-----------------------------------------------------------------------------
// COMMON FUNCTION: halReadReg8()
// COMMON FUNCTION: halReadReg16()
// COMMON FUNCTION: halReadReg32()
//
//	*** NOTE ***
//	The register access routines do not perform any error checking. This means:
//	 1) NEVER call these routines if halAcquireController() did not succeed.
//	 2) No checks are performed on the Offset to ensure it is a valid register
//		address. It is possible to request a read or write to a non-register
//		offset.
//-----------------------------------------------------------------------------
UInt8 halReadReg8( UInt32 Index )
{
	UInt8 Data;
	Data = halpParReadReg8( Index );
	return Data;
}


UInt16 halReadReg16( UInt32 Index )
{
	UInt8 Data;
	Data=halpParReadReg16( Index );
	return Data;
}

UInt32 halReadReg32( UInt32 Index )
{
	UInt8 Data;
	Data=halpParReadReg32( Index );
	return Data; 
}



//-----------------------------------------------------------------------------
// COMMON FUNCTION: halWriteReg8()
// COMMON FUNCTION: halWriteReg16()
// COMMON FUNCTION: halWriteReg32()
//-----------------------------------------------------------------------------
void halWriteReg8( UInt32 Index, UInt8 Value )
{
	halpParWriteReg8( Index, Value );
}


void halWriteReg16( UInt32 Index, UInt16 Value )
{
	halpParWriteReg16( Index, Value );
}


void halWriteReg32( UInt32 Index, UInt32 Value )
{
	halpParWriteReg32( Index, Value );
}


//-----------------------------------------------------------------------------
// COMMON FUNCTION: halModifyReg8()
// COMMON FUNCTION: halModifyReg16()
// COMMON FUNCTION: halModifyReg32()
//-----------------------------------------------------------------------------
UInt8 halModifyReg8( UInt32 Index, UInt32 AndMask, UInt32 OrMask )
{
	UInt8 Value = (UInt8)((halReadReg8(Index) & ~AndMask) | OrMask);
	halWriteReg8( Index, Value );
	return Value;
}

#if 0
UInt16 halModifyReg16( UInt32 Index, UInt32 AndMask, UInt32 OrMask )
{
	UInt16 Value = (UInt16)((halReadReg16(Index) & ~AndMask) | OrMask);
	halWriteReg16( Index, Value );
	return Value;
}

UInt32 halModifyReg32( UInt32 Index, UInt32 AndMask, UInt32 OrMask )
{
	UInt32 Value = ((halReadReg32(Index) & ~AndMask) | OrMask);
	halWriteReg32( Index, Value );
	return Value;
}
#endif


//-----------------------------------------------------------------------------
// COMMON FUNCTION: halReadData()
// COMMON FUNCTION: halWriteData()
//-----------------------------------------------------------------------------
void halReadData( int AccessSize, UInt32 Offset, void* pData, UInt32 nBytes )
{
	if ( !AccessSize )
		AccessSize = HAL_REGBYTES;

	halpParReadData( AccessSize, Offset, pData, nBytes/AccessSize );
}

void halWriteData( int AccessSize, UInt32 Offset, const void* pData, UInt32 nBytes )
{
	if ( !AccessSize )
		AccessSize = HAL_REGBYTES;

	halpParWriteData( AccessSize, Offset, pData, nBytes/AccessSize );
}




//----------------------------------------------------------------------------
// COMMON FUNCTION: halDelayUS - Delay for the specified number of microseconds.
//
// This function can be used to delay for short times.
//----------------------------------------------------------------------------
Boolean halDelayUS( UInt32 microseconds )
{
#if 0
	CAMERA_TR(CAM_TR_MASK_HAL, ("halDelayUS(%d)", microseconds));
	
	if(microseconds > 5000)
		rvf_delay(RVF_MS_TO_TICKS(microseconds/1000));
	else
		wait_ARM_cycles(convert_nanosec_to_cycles(microseconds * 1000));
	
	return TRUE;
#else
	UINT32 i;
	for( i = 0; i < microseconds; i++ )		// 1s at 40MHz main clk and 20MHz bus clk wait 1 cycle
	{
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
	}
	return TRUE;

//	volatile unsigned short a;

//    	while (microseconds-- > 0) for (a=0;a<5;a++) asm(" nop");

//	return TRUE;
#endif
}

void halDelayUS_forInit( UInt32 microseconds )
{
	UINT32 i;

	for( i = 0; i < microseconds; i++ )		// 1s at 40MHz main clk and 20MHz bus clk wait 1 cycle
	{
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
		asm(" nop ");
	}

}



//===========================================================================
// PRIVATE (STATIC) FUNCTIONS (local to HAL only)
//===========================================================================



//---------------------------------------------------------------------------
// PRIVATE FUNCTION: InitRegisters()
//---------------------------------------------------------------------------
static void InitRegisters( void )
{
	int idx;
	UInt32 reg;
	UInt8 reg1,reg2;

	for (idx = 0; HalInfo.Regs[idx].Index != REGFLAG_ENDOFTABLE; idx++)
	{
		reg = HalInfo.Regs[idx].Index;
		
		if (reg < REGFLAG_BASE)
		{
			//reg1=HAL_READ_REG(reg);
			//MI_dbgPrintf("1reg%x=%x",reg,reg1);
			//halDelayUS(1000);
			HAL_WRITE_REG(reg, (UInt8)HalInfo.Regs[idx].Value);
			//reg2=HAL_READ_REG(reg);
			//MI_dbgPrintf("2reg%x=%x",reg,reg2);
			//halDelayUS(1000);
		}
		else if (reg == REGFLAG_DELAY || reg == REGFLAG_OFFDELAY || reg == REGFLAG_ONDELAY)
		{
			halDelayUS(HalInfo.Regs[idx].Value * 1000UL);
		}
	}
}


//---------------------------------------------------------------------------
//	PRIVATE FUNCTION: ProgramParallelPanels()
//
//	DESCRIPTION:
//---------------------------------------------------------------------------
static void ProgramParallelPanels( void )
{
	int Panel;
	UInt32 *pLCDRegs;
	UIntReg Reg003C = HAL_READ_REG(REG003C_LCDXFER);
	UIntReg Reg004C = HAL_READ_REG(REG004C_DISPSETTING);

	// Auto frame transfer must be disabled before accessing the LCD interface registers.
	HAL_WRITE_REG(REG003C_LCDXFER, 0);

	Panel = 1;				// LCD panel 1 or 2
	do
	{
		pLCDRegs = (Panel==1) ? HalInfo.LCD1Regs : (Panel==2) ? HalInfo.LCD2Regs : HalInfo.LCD3Regs;

		if ( *pLCDRegs == HAL_LCD_ENDOFTABLE )
			continue;

		// Select the appropriate LCD Output Port
		HAL_WRITE_REG(REG004C_DISPSETTING, (UIntReg)((HAL_READ_REG(REG004C_DISPSETTING) & ~3) | Panel));

		// *** Don't try to write to the command/data registers when the LCD interface is busy
		while ( *pLCDRegs != HAL_LCD_ENDOFTABLE )
		{
			UInt32 Value = (*pLCDRegs & HAL_LCD_DATAMASK);

			switch ( (*pLCDRegs & HAL_LCD_FLAGMASK) )
			{
			case HAL_LCD_COMMAND:
				while ((HAL_READ_REG(REG0038_LCDSTATUS) & 0x01))
					continue;
				HAL_WRITE_REG(REG0030_LCDCMD0, LOBYTE(Value));
				//HAL_WRITE_REG(REG0031_LCDCMD1, HIBYTE(Value));
				break;

			case HAL_LCD_PARAM:
				while ((HAL_READ_REG(REG0038_LCDSTATUS) & 0x01))
					continue;
				HAL_WRITE_REG(REG0032_LCDWRITE0, LOBYTE(Value));
				//HAL_WRITE_REG(REG0033_LCDWRITE1, HIBYTE(Value));
				break;

			  case HAL_LCD_DELAY:
				halDelayUS( Value * 1000UL );
				break;

			  default:
				break;
			}

			pLCDRegs++;
		};

	} while ( ++Panel <= 3 );

	// Restore the selected display panel and transfer mode.
	HAL_WRITE_REG(REG004C_DISPSETTING, Reg004C);
	HAL_WRITE_REG(REG003C_LCDXFER, Reg003C);
}


//---------------------------------------------------------------------------
// PRIVATE FUNCTION: ClearVideoMemory()
//---------------------------------------------------------------------------
static void ClearVideoMemory( void )
{
	UInt32 MemSize = HAL_READ_REG(REG0001_PRODINFO1) * 4096UL;

	halWriteDisplay32( 0, 0, MemSize/sizeof(UInt32) );
	halWriteReg8( REG003A_LCDFRAMEXFER, 1 );

}


void halSetProcessExclusivity( Boolean fExclusive )
{

}
