//===========================================================================
//	HAL_INDIRECT.C
//
//	The routines in this file comprise the register and access methods for
//	the S1D13716.
//
//	The 13716 supports two indirect access modes (and no direct methods).
//	The modes have been labeled "Serial Indirect" and "Parallel Indirect".
//	The following text describes the access modes, from a programming
//	point of view as they pertain to accessing the 13716 mounted on a
//	S5U13716B00? evaluation board.
//
//	Regardless of the access mode the 13716 interprets the current access
//	cycle according to the state of the RS control line. On 13716 eval
//	boards the RS line is connected to Address 1. Writing to odd addresses
//	result in the RS line being pulled ???? and writing to even addresses
//	result in the RS line being pulled ????.
//
//	Parallel Indirect
//	This mode consists of two distinct cycles, address and data. When the
//	RS line is low an address cycle is performed. When the RS line is high
//	the 13716 interprets this to be a data cycle.
//
//	A typical Parallel Indirect sequence to access a control register
//	will resemble the following:
//		With RS low
//		 - Write the register LSB
//		 - Write the register MSB	// this step can be skipped if the
//									// MSB was previousley set
//		With RS high
//		 - Read/Write the register data
//
//	Serial Indirect
//	This mode uses three types of cycles. A command cycle, when RS is low
//	(valid commands consist of SER_SET_ADDR, WRITE and READ) and address or
//	data cycles when the RS line is high.
//
//	A typical Serial Indirect sequence to access a control register
//	will resemble the following:
//		With RS low
//		 - SER_SET_ADDR
//		With RS high
//		 - Write the Addr LSB 		// Write the address
//		 - Write the Addr MSB		// this step can be skipped if the
//									// MSB was previously set
//		With RS low
//		 - WRITE					// Inform the system we want to write data
//		With RS high
//		 -  Read/Write register data
//---------------------------------------------------------------------------
//  Copyright (c) 2003 Epson Research and Development, Inc.
//  All Rights Reserved.
//===========================================================================
// (Tabs set to every 4)

#include "hal.h"
#include "hal_platform.h"
#include "hal_private.h"
#include "hal_indirect.h"

HALIND_STRUCT HalIndInfo;			// HAL indirect pointers structure.

UInt16 tmp;

UInt16 swap16addr( in UInt16 Value )
{
	return (UInt16)( ((Value>>8)&0x00FF) | ((Value&0x00FF)<<8) );
}




//---------------------------------------------------------------------------
// INDIRECT FUNCTION: halpIndirectInit() - Initialize HAL to communicate using indirect interface
//---------------------------------------------------------------------------
void halpIndirectInit( UInt32 RegAddr )
{
	HalIndInfo.pIndex8	= (pvUInt8)(RegAddr + 0);
	HalIndInfo.pData8	= (pvUInt8)(RegAddr + 2);

	HalIndInfo.pIndex16	= (pvUInt16)(RegAddr + 0);
	HalIndInfo.pData16	= (pvUInt16)(RegAddr + 2);
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParWriteReg8()
//
//	DESCRIPTION:
//		Writes one eight bit control register with the specified data
//		using indirect parallel I/O.
//
//	PARAMETERS:
//		index	- Register offset to write the value to
//		value	- The eight bit value to write to the register
//---------------------------------------------------------------------------
void halpParWriteReg8( UInt32 index, UInt8 value )
{
	halSetProcessExclusivity( TRUE );

	// Setup the address
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	
	// Write the value
	*HalIndInfo.pData8 = value;

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParWriteReg16()
//
//	DESCRIPTION:
//		Writes two consective eight bit control registers with
//		one sixteen bit value using parallel indirect I/O.
//
//	PARAMETERS:
//		index	- Register offset to the LSB register
//				  The MSB of value will be written to index + 1
//		value	- The sixteen bit value to write to the registers
//---------------------------------------------------------------------------
void halpParWriteReg16( UInt32 index, UInt16 value )
{
	halSetProcessExclusivity( TRUE );

	// Read first 8-bit value.
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	*HalIndInfo.pData8 = LOBYTE(value);

	// Read second 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	*HalIndInfo.pData8 = HIBYTE(value);

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParWriteReg32()
//
//	DESCRIPTION:
//		Writes four consective eight bit control registers with
//		one thirty-two bit value using parallel indirect I/O.
//
//	PARAMETERS:
//		index	- Register offset to the LSB register
//				  The MSB of value will be written to index + 1 ...
//		value	- The thirty-two bit value to write to the registers
//---------------------------------------------------------------------------
void halpParWriteReg32( UInt32 index, UInt32 value )
{
	halSetProcessExclusivity( TRUE );

	// Read first 8-bit value.
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	*HalIndInfo.pData8 = LOBYTE(LOWORD(value));

	// Read second 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	*HalIndInfo.pData8 = HIBYTE(LOWORD(value));

	// Read third 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	*HalIndInfo.pData8 = LOBYTE(HIWORD(value));

	// Read fourth 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	*HalIndInfo.pData8 = HIBYTE(HIWORD(value));

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParReadReg8()
//
//	DESCRIPTION:
//		Reads one eight bit register and returns the value to the caller
//		using parallel indirect I/O.
//
//	PARAMETERS:
//		index	- The register offset to read data from
//
//	RETURNS:
//		The eight bit value read from the register.
//---------------------------------------------------------------------------
UInt8 halpParReadReg8( UInt32 index )
{
	UInt8 val = 0;

	halSetProcessExclusivity( TRUE );

	// Setup the register address
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}

	val = *HalIndInfo.pData8;

	halSetProcessExclusivity( FALSE );

	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParReadReg16()
//
//	DESCRIPTION:
//		Reads two consective eight bit control registers and returns the
//		resulting sixteen bit value using parallel indirect I/O.
//
//	PARAMETERS:
//		index	- Register offset to the LSB register
//				  The MSB value will be read from index + 1
//
//	RETURNS:
//		The sixteen bit value read from the registers.
//---------------------------------------------------------------------------
UInt16 halpParReadReg16( UInt32 index )
{
	UInt16 val;

	halSetProcessExclusivity( TRUE );

	// Read first 8-bit value.
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	val = *HalIndInfo.pData8;

	// Read second 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	val |= (UInt16)(*HalIndInfo.pData8<<8);

	halSetProcessExclusivity( FALSE );

	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParReadReg32()
//
//	DESCRIPTION:
//		Reads four consective eight bit control registers and returns the
//		resulting thirty-two bit value using parallel indirect I/O.
//
//	PARAMETERS:
//		index	- Register offset to the LSB register
//
//	RETURNS:
//		The thirty-two bit value read from the registers.
//---------------------------------------------------------------------------
UInt32 halpParReadReg32( UInt32 index )
{
	UInt16 loword, hiword;

	halSetProcessExclusivity( TRUE );

	// Read first 8-bit value.
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	loword = *HalIndInfo.pData8;

	// Read second 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	loword |= (UInt16)(*HalIndInfo.pData8<<8);

	// Read third 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	hiword = *HalIndInfo.pData8;

	// Read fourth 8-bit value.
	index++;
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
		*HalIndInfo.pIndex16 = LOWORD(index);
	else
	{
		*HalIndInfo.pIndex8 = LOBYTE(index);
		*HalIndInfo.pIndex8 = HIBYTE(index);
	}
	hiword |= (UInt16)(*HalIndInfo.pData8<<8);

	halSetProcessExclusivity( FALSE );

	return MAKELONG(loword, hiword);
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParWriteDisplay8()
//
//	DESCRIPTION:
//		This routine writes one eight bit value to the specified display
//		memory offset using parallel indirect I/O.
//
//	PARAMETERS:
//		addr	- The offset in display memory to write the value to.
//		data	- The eight bit value to be written to display memory.
//---------------------------------------------------------------------------
void halpParWriteDisplay8( UInt32 addr, UInt8 data, UInt32 count )
{
	UInt32 loop;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrParallel(addr);
	
	for ( loop=0; loop<count; loop++ )
		*HalIndInfo.pData8 = data;

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParWriteDisplay16()
//
//	DESCRIPTION:
//		This routine writes one sixteen bit value to two consective
//		display memory addresses using parallel indirect I/O.
//
//	PARAMETERS:
//		addr	- The offset in display memory to write the LSB of value to
//				  The MSB will be wrtten to addr + 1
//		data	- The 16-bit value to be written to display memory.
//---------------------------------------------------------------------------
void halpParWriteDisplay16( UInt32 addr, UInt16 data, UInt32 count )
{
	UInt32 loop;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrParallel(addr);

	// Write the data
	for ( loop=0; loop<count; loop++ )
	{
		if (HalInfo.dwFlags & fINDIRECT_DATA16)
			*HalIndInfo.pData16 = data;
		else
		{
			*HalIndInfo.pData8 = LOBYTE(data);
			*HalIndInfo.pData8 = HIBYTE(data);
		}
	}

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParWriteDisplay32()
//
//	DESCRIPTION:
//		This routine writes four consective eight bit display memory addresses
//		with the thirty-two bits of data using parallel indirect I/O.
//
//	PARAMETERS:
//		addr	- Specifies the offset to the LSB of the data to be written
//		data	- The 32-bit data to be written to the specified address
//---------------------------------------------------------------------------
void halpParWriteDisplay32( UInt32 addr, UInt32 data, UInt32 count )
{
	UInt32 loop;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrParallel(addr);

	// Write the data
	for ( loop=0; loop<count; loop++ )
	{
		if (HalInfo.dwFlags & fINDIRECT_DATA16)
		{
			*HalIndInfo.pData16 = LOWORD(data);
			*HalIndInfo.pData16 = HIWORD(data);
		}
		else
		{
			*HalIndInfo.pData8 = LOBYTE(LOWORD(data));
			*HalIndInfo.pData8 = HIBYTE(LOWORD(data));
			*HalIndInfo.pData8 = LOBYTE(HIWORD(data));
			*HalIndInfo.pData8 = HIBYTE(HIWORD(data));
		}
	}

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParReadDisplay8()
//
//	DESCRIPTION:
//		This routine reads one eight bit value from the specified display
//		memory address using parallel indirect I/O.
//
//	PARAMETERS:
//		addr	- The address to read from
//
//	RETURNS:
//		The value read from the offset specified by addr.
//---------------------------------------------------------------------------
UInt8 halpParReadDisplay8( UInt32 addr )
{
	UInt8 val;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrParallel(addr);
	val = *HalIndInfo.pData8;

	halSetProcessExclusivity( FALSE );

	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParReadDisplay16()
//
//	DESCRIPTION:
//		This routine reads sixteen bits of data (word) from the specified
//		display memory offset using parallel indirect I/O.
//
//	PARAMETERS:
//		addr	- The address to read the LSB of data from
//
//	RETURNS:
//		The 16-bit value read from display memory.
//---------------------------------------------------------------------------
UInt16 halpParReadDisplay16( UInt32 addr )
{
	UInt16 val;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrParallel(addr);
	val = *HalIndInfo.pData16;

	halSetProcessExclusivity( FALSE );

	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpParReadDisplay32()
//
//	DESCRIPTION:
//		This routine reads thirty-two bits of data (dword) from the specified
//		display memory offset using parallel indirect I/O.
//
//	PARAMETERS:
//		addr	- The address to read the LSB of data from
//
//	RETURNS:
//		The 32-bit value read from display memory.
//---------------------------------------------------------------------------
UInt32 halpParReadDisplay32( UInt32 addr )
{
	UInt16 loword, hiword;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrParallel(addr);

	loword = *HalIndInfo.pData16;
	hiword = *HalIndInfo.pData16;

	halSetProcessExclusivity( FALSE );

	return MAKELONG(loword, hiword);
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerWriteReg8()
//
//	DESCRIPTION:
//		Writes one eight bit register with the specified data using indirect
//		serial I/O
//
//	PARAMETERS:
//		index	- the register offset to write the value to
//		value	- an 8-bit value to be written to the specified register
//---------------------------------------------------------------------------
void halpSerWriteReg8( UInt32 index, UInt8 value )
{
	halSetProcessExclusivity( TRUE );

	*HalIndInfo.pIndex8 = SER_SET_ADDR;
	*HalIndInfo.pData8  = LOBYTE(index);
	*HalIndInfo.pData8  = HIBYTE(index);
	*HalIndInfo.pIndex8 = SER_WRITE;
	*HalIndInfo.pData8  = value;

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerWriteReg16()
//
//	DESCRIPTION:
//		Writes two eight bit registers with the specified data using indirect
//		serial I/O
//
//	PARAMETERS:
//		index	- the register offset to write the LSB of value to
//		value	- a 16-bit value to be written to the specified register
//---------------------------------------------------------------------------
void halpSerWriteReg16( UInt32 index, UInt16 value )
{
	halSetProcessExclusivity( TRUE );

	// Write first 8-bit value.
	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index);
	*HalIndInfo.pData8   = HIBYTE(index);
	*HalIndInfo.pIndex8  = SER_WRITE;
	*HalIndInfo.pData8   = LOBYTE(value);

	// Write second 8-bit value.
	index++;
	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index);
	*HalIndInfo.pData8   = HIBYTE(index);
	*HalIndInfo.pIndex8  = SER_WRITE;
	*HalIndInfo.pData8   = HIBYTE(value);

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerWriteReg32()
//
//	DESCRIPTION:
//		Writes four eight bit registers with the specified data using indirect
//		serial I/O
//
//	PARAMETERS:
//		index	- the register offset to write the LSB of value
//		value	- a 32-bit value to be written to the specified register
//---------------------------------------------------------------------------
void halpSerWriteReg32( UInt32 index, UInt32 value )
{
	UInt8  hibyte;

	halSetProcessExclusivity( TRUE );

	// Setup the register address
	hibyte = HIBYTE(index);
	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index);
	*HalIndInfo.pData8   = HIBYTE(index);
	*HalIndInfo.pIndex8  = SER_WRITE;
	*HalIndInfo.pData8   = LOBYTE(LOWORD(value));
	*HalIndInfo.pData8   = HIBYTE(LOWORD(value));

	// Set the next register address
	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index+2);
	if( hibyte!=HIBYTE(index+2) )
		*HalIndInfo.pData8 = HIBYTE(index+2);
	*HalIndInfo.pIndex8  = SER_WRITE;
	*HalIndInfo.pData8   = LOBYTE(HIWORD(value));
	*HalIndInfo.pData8   = HIBYTE(HIWORD(value));

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerReadReg8()
//
//	DESCRIPTION:
//		This routine reads one eight bit value from the register specified by
//		index using serial indirect I/O.
//
//	PARAMETERS:
//		index	- The register offset to read data from
//
//	RETURNS:
//		The 8-bit value read from the register.
//---------------------------------------------------------------------------
UInt8 halpSerReadReg8( UInt32 index )
{
	UInt8 val;

	halSetProcessExclusivity( TRUE );

	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index);
	*HalIndInfo.pData8   = HIBYTE(index);
	*HalIndInfo.pIndex8 = SER_READ;
	val = *HalIndInfo.pData8;

	halSetProcessExclusivity( FALSE );
	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerReadReg16()
//
//	DESCRIPTION:
//		This routine reads two eight bit registers using serial indirect I/O
//		and returns the result as a sixteen bit value. Must always read 8-bit values.
//
//	PARAMETERS:
//		index	- the register offset to read the LSB of the return value.
//
//	RETURNS:
//		The sixteen bit combined value of the two registers.
//---------------------------------------------------------------------------
UInt16 halpSerReadReg16( UInt32 index )
{
	UInt16 val;

	halSetProcessExclusivity( TRUE );

	// Read first 8-bit value.
	*HalIndInfo.pIndex8 = SER_SET_ADDR;
	*HalIndInfo.pData8  = LOBYTE(index);
	*HalIndInfo.pData8  = HIBYTE(index);
	*HalIndInfo.pIndex8 = SER_READ;
	val = *HalIndInfo.pData8;

	// Read second 8-bit value.
	index++;
	*HalIndInfo.pIndex8 = SER_SET_ADDR;
	*HalIndInfo.pData8  = LOBYTE(index);
	*HalIndInfo.pData8  = HIBYTE(index);
	*HalIndInfo.pIndex8 = SER_READ;
	val |= (UInt16)(*HalIndInfo.pData8)<<8;

	halSetProcessExclusivity( FALSE );
	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpIndirectReadReg32()
//
//	DESCRIPTION:
//		This routine reads four eight bit registers using serial indirect I/O
//		and returns the result as a thirty-two bit value.
//
//	PARAMETERS:
//		index	- the register offset to read the LSB of the return value.
//
//	RETURNS:
//		The thirty-two bit combined value of the four registers.
//---------------------------------------------------------------------------
UInt32 halpSerReadReg32( UInt32 index )
{
	UInt8	hibyte;
	UInt16	loword, hiword;

	halSetProcessExclusivity( TRUE );

	hibyte = HIBYTE(index);
	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index);
	*HalIndInfo.pData8   = HIBYTE(index);
	*HalIndInfo.pIndex8 = SER_READ;
	loword  = *HalIndInfo.pData8;
	loword |= (UInt16)(*HalIndInfo.pData8<<8);

	*HalIndInfo.pIndex8  = SER_SET_ADDR;
	*HalIndInfo.pData8   = LOBYTE(index+2);
	if( hibyte!=HIBYTE(index+2) )
		*HalIndInfo.pData8   = HIBYTE(index+2);
	*HalIndInfo.pIndex8 = SER_READ;
	hiword  = *HalIndInfo.pData8;
	hiword |= (UInt16)(*HalIndInfo.pData8<<8);

	halSetProcessExclusivity( FALSE );
	return MAKELONG(loword, hiword);
}


//---------------------------------------------------------------------------
//	FUNCTION: halpIndirectWriteDisplay8()
//
//	DESCRIPTION:
//		This routine writes one eight bit value to the specified display
//		memory offset using serial indirect I/O.
//
//	PARAMETERS:
//		addr	- The offset in display memory to write the value to.
//		data	- The eight bit value to be written to display memory.
//---------------------------------------------------------------------------
void halpSerWriteDisplay8( UInt32 addr, UInt8 data, UInt32 count )
{
	UInt32 loop;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrSerial(addr);
	
	*HalIndInfo.pIndex8  = SER_WRITE;
	for (loop = 0; loop < count; loop++)
		*HalIndInfo.pData8 = data;

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpIndirectWriteDisplay16()
//
//	DESCRIPTION:
//		This routine writes one sixteen bit value to two consective
//		display memory addresses using serial indirect I/O.
//
//	PARAMETERS:
//		addr	- The offset in display memory to write the LSB of value to
//				  The MSB will be wrtten 
//		data	- The 16-bit value to be written to display memory.
//---------------------------------------------------------------------------
void halpSerWriteDisplay16( UInt32 addr, UInt16 data, UInt32 count )
{
	UInt32 loop;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrSerial(addr);
	
	*HalIndInfo.pIndex8  = SER_WRITE;
	if (HalInfo.dwFlags & fINDIRECT_DATA16)
    {
	    for (loop = 0; loop < count; loop++)
		    *HalIndInfo.pData16 = data;
    } else
    {
	    for (loop = 0; loop < count; loop++)
        {
		    *HalIndInfo.pData8 = LOBYTE(data);
		    *HalIndInfo.pData8 = HIBYTE(data);
        }   
    }
	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpIndirectWriteDisplay32()
//
//	DESCRIPTION:
//		This routine writes four consective eight bit display memory addresses
//		with the thirty-two bits of data using serial indirect I/O.
//
//	PARAMETERS:
//		addr	- Specifies the offset to the LSB of the data to be written
//		data	- The 32-bit data to be written to the specified address
//---------------------------------------------------------------------------
void halpSerWriteDisplay32( UInt32 addr, UInt32 data, UInt32 count )
{
	UInt32 loop;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrSerial(addr);
	
	*HalIndInfo.pIndex8  = SER_WRITE;
	if (HalInfo.dwFlags & fINDIRECT_DATA16)
	{
		for (loop = 0; loop < count; loop++)
		{
			*HalIndInfo.pData16 = LOWORD(data);
			*HalIndInfo.pData16 = HIWORD(data);
		}
	} else
	{
		for (loop = 0; loop < count; loop++)
		{
			*HalIndInfo.pData8 = LOBYTE(LOWORD(data));
			*HalIndInfo.pData8 = HIBYTE(LOWORD(data));
			*HalIndInfo.pData8 = LOBYTE(HIWORD(data));
			*HalIndInfo.pData8 = HIBYTE(HIWORD(data));
		}
	}

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerReadDisplay8()
//
//	DESCRIPTION:
//		This routine reads one eight bit value from the specified display
//		memory address using serial indirect I/O.
//
//	PARAMETERS:
//		addr	- The address to read from
//
//	RETURNS:
//		The value read from the offset specified by addr.
//---------------------------------------------------------------------------
UInt8 halpSerReadDisplay8( UInt32 addr )
{
	UInt8 val;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrSerial(addr);
	
	*HalIndInfo.pIndex8  = SER_READ;
	val = *HalIndInfo.pData8;

	halSetProcessExclusivity( FALSE );

	return val;
}



//---------------------------------------------------------------------------
//	FUNCTION: halpSerReadDisplay16()
//
//	DESCRIPTION:
//		This routine reads sixteen bits of data (word) from the specified
//		display memory offset using serial indirect I/O.
//
//	PARAMETERS:
//		addr	- The address to read the LSB of data from
//
//	RETURNS:
//		The 16-bit value read from display memory.
//---------------------------------------------------------------------------
UInt16 halpSerReadDisplay16( UInt32 addr )
{
	UInt16 val;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrSerial(addr);
	
	*HalIndInfo.pIndex8  = SER_READ;
	if (HalInfo.dwFlags & fINDIRECT_DATA16)
	    val = *HalIndInfo.pData16;
    else
    {
	    val  = *HalIndInfo.pData8;
	    val |= (UInt16)((*HalIndInfo.pData8)<<8);
    }

	halSetProcessExclusivity( FALSE );

	return val;
}


//---------------------------------------------------------------------------
//	FUNCTION: halpSerReadDisplay32()
//
//	DESCRIPTION:
//		This routine reads thirty-two bits of data (dword) from the specified
//		display memory offset using serial indirect I/O.
//
//	PARAMETERS:
//		addr	- The address to read the LSB of data from
//
//	RETURNS:
//		The 32-bit value read from display memory.
//---------------------------------------------------------------------------
UInt32 halpSerReadDisplay32( UInt32 addr )
{
	UInt16 loword, hiword;

	halSetProcessExclusivity( TRUE );

	SetDisplayAddrSerial(addr);
	
	*HalIndInfo.pIndex8  = SER_READ;
	if (HalInfo.dwFlags & fINDIRECT_DATA16)
    {
	    loword = *HalIndInfo.pData16;
	    hiword = *HalIndInfo.pData16;
    }
	else
    {
	    loword  = *HalIndInfo.pData8;
	    loword |= (UInt16)((*HalIndInfo.pData8)<<8);
	    hiword  = *HalIndInfo.pData8;
	    hiword |= (UInt16)((*HalIndInfo.pData8)<<8);
    }
	halSetProcessExclusivity( FALSE );

	return MAKELONG(loword, hiword);
}




//---------------------------------------------------------------------------
// INDIRECT FUNCTION: halpSerReadData()
// WARNING: DO NOT USE ODD ADDRESSES FOR 16-bit or 32bit ACCESSES!
//---------------------------------------------------------------------------
void halpSerReadData( int Size, UInt32 Offset, void* pData, UInt32 Count )
{
	if ( Count && pData )
	{
		halSetProcessExclusivity( TRUE );

		if ( Size == sizeof(UInt8) )
		{
			pUInt8 pWrite = pData;
			SetDisplayAddrSerial( Offset );

			*HalIndInfo.pIndex8 = SER_READ;
			do
			{
				*pWrite++ = *HalIndInfo.pData8;
			} while ( --Count );
		}
		else if ( Size == sizeof(UInt16) )
		{
			pUInt16 pWrite  = pData;
            UInt16  wVal;
			SetDisplayAddrSerial( Offset );

			*HalIndInfo.pIndex8 = SER_READ;
			if (HalInfo.dwFlags & fINDIRECT_DATA16)
            {
			    do
			    {
				    *pWrite++ = *HalIndInfo.pData16;
			    } while ( --Count );
            } else
            {
			    do
			    {
				    wVal  = *HalIndInfo.pData8;
				    wVal |= (UInt16)((*HalIndInfo.pData8)<<8);
                    *pWrite++ = wVal;
			    } while ( --Count );
            }
		}
		else if ( Size == sizeof(UInt32) )
		{
			UInt16 loword, hiword;
			pUInt32 pWrite = pData;
			SetDisplayAddrSerial( Offset );

			*HalIndInfo.pIndex8 = SER_READ;
			if (HalInfo.dwFlags & fINDIRECT_DATA16)
            {
			    do
			    {
				    loword = *HalIndInfo.pData16;
				    hiword = *HalIndInfo.pData16;
				    *pWrite++ = MAKELONG(loword,hiword);
			    } while ( --Count );
            } else
            {
			    do
			    {
				    loword  = *HalIndInfo.pData8;
				    loword |= (UInt16)((*HalIndInfo.pData8)<<8);
				    hiword  = *HalIndInfo.pData8;
				    hiword |= (UInt16)((*HalIndInfo.pData8)<<8);
				    *pWrite++ = MAKELONG(loword,hiword);
			    } while ( --Count );
            }
		}

		halSetProcessExclusivity( FALSE );
	}
}




//---------------------------------------------------------------------------
// INDIRECT FUNCTION: halpParReadData()
// WARNING: DO NOT USE ODD ADDRESSES FOR 16-bit or 32bit ACCESSES!
//---------------------------------------------------------------------------
void halpParReadData( int Size, UInt32 Offset, void* pData, UInt32 Count )
{
	if ( Count && pData )
	{
		halSetProcessExclusivity( TRUE );

		if ( Size == sizeof(UInt8) )
		{
			pUInt8 pWrite = pData;
			SetDisplayAddrParallel( Offset );
			do
			{
				*pWrite++ = *HalIndInfo.pData8;
			} while ( --Count );
		}
		else if ( Size == sizeof(UInt16) )
		{
			pUInt16 pWrite = pData;
			SetDisplayAddrParallel( Offset );
			do
			{
				tmp = *HalIndInfo.pData16;
				//*pWrite++ = *HalIndInfo.pData16;
				*pWrite++ = tmp;
			} while ( --Count );
		}
		else if ( Size == sizeof(UInt32) )
		{
			UInt16 loword, hiword;
			pUInt32 pWrite = pData;
			SetDisplayAddrParallel( Offset );
			do
			{
				loword = *HalIndInfo.pData16;
				hiword = *HalIndInfo.pData16;

				*pWrite++ = MAKELONG(loword,hiword);
			} while ( --Count );
		}

		halSetProcessExclusivity( FALSE );
	}
}




//---------------------------------------------------------------------------
// INDIRECT FUNCTION: halpSerWriteData()
// WARNING: DO NOT USE ODD ADDRESSES FOR 16-bit or 32bit ACCESSES!
//---------------------------------------------------------------------------
void halpSerWriteData( int Size, UInt32 Offset, const void* pData, UInt32 Count )
{
	if ( Count && pData )
	{
		halSetProcessExclusivity( TRUE );

		if ( Size == sizeof(UInt8) )
		{
			const UInt8* pRead = pData;
			SetDisplayAddrSerial( Offset );

			*HalIndInfo.pIndex8  = SER_WRITE;
			do
			{
				*HalIndInfo.pData8 = *pRead++;
			} while ( --Count );
		}
		else if ( Size == sizeof(UInt16) )
		{
			const UInt16* pRead = pData;
			SetDisplayAddrSerial( Offset );

			*HalIndInfo.pIndex8  = SER_WRITE;
			if (HalInfo.dwFlags & fINDIRECT_DATA16)
            {
			    do
			    {
				    *HalIndInfo.pData16 = LOWORD(*pRead);
				    pRead++;
			    } while ( --Count );
            } else
            {
                UInt16 wVal;
			    do
			    {
                    wVal = LOWORD(*pRead);
				    *HalIndInfo.pData8 = LOBYTE(wVal);
				    *HalIndInfo.pData8 = HIBYTE(wVal);
				    pRead++;
			    } while ( --Count );
            }
		}
		else if ( Size == sizeof(UInt32) )
		{
			const UInt32* pRead = pData;
			SetDisplayAddrSerial( Offset );

			*HalIndInfo.pIndex8  = SER_WRITE;
			if (HalInfo.dwFlags & fINDIRECT_DATA16)
            {
			    do
			    {
				    *HalIndInfo.pData16 = LOWORD(*pRead);
				    *HalIndInfo.pData16 = HIWORD(*pRead);
					pRead++;
			    } while ( --Count );
            } else
            {
                UInt32 dwVal;
                UInt16 wVal;
			    do
			    {
                    dwVal = *pRead++;
                    wVal = LOWORD(dwVal);
				    *HalIndInfo.pData8 = LOBYTE(wVal);
				    *HalIndInfo.pData8 = HIBYTE(wVal);
                    wVal = HIWORD(dwVal);
				    *HalIndInfo.pData8 = LOBYTE(wVal);
				    *HalIndInfo.pData8 = HIBYTE(wVal);
			    } while ( --Count );
            }
		}

		halSetProcessExclusivity( FALSE );
	}
}




//---------------------------------------------------------------------------
// INDIRECT FUNCTION: halpParWriteData()
// WARNING: DO NOT USE ODD ADDRESSES FOR 16-bit or 32bit ACCESSES!
//---------------------------------------------------------------------------
void halpParWriteData( int Size, UInt32 Offset, const void* pData, UInt32 Count )
{
	if ( Count && pData )
	{
		halSetProcessExclusivity( TRUE );

		if ( Size == sizeof(UInt8) )
		{
			const UInt8* pRead = pData;
			SetDisplayAddrParallel( Offset );
			do
			{
				*HalIndInfo.pData8 = *pRead++;
			} while ( --Count );
		}
		else if ( Size == sizeof(UInt16) )
		{
			const UInt16* pRead = pData;
			SetDisplayAddrParallel( Offset );
			do
			{
				//*HalIndInfo.pData16 = *pRead;
				
				//*HalIndInfo.pData16  = *( volatile unsigned short *)0x200c000;
				
				tmp =*pRead;
				*HalIndInfo.pData16 = tmp;
				pRead++;
			} while ( --Count );
			
			/*
			SetDisplayAddrParallel( Offset );
			do
			{	
				if (HalInfo.dwFlags & fINDIRECT_DATA16)
					*HalIndInfo.pData16 = *pRead;
				else
				{
					*HalIndInfo.pData8 = HIBYTE(*pRead);
					*HalIndInfo.pData8 = LOBYTE(*pRead);
					//*HalIndInfo.pData8 = HIBYTE(0xFFFF);
					//*HalIndInfo.pData8 = LOBYTE(0xFFFF);					
					 //pRead++;
				}
				
			} while ( --Count );
			*/
		}
		else if ( Size == sizeof(UInt32) )
		{
			const UInt32* pRead = pData;
			SetDisplayAddrParallel( Offset );
			do
			{
				*HalIndInfo.pData16 = LOWORD(*pRead);
				*HalIndInfo.pData16 = HIWORD(*pRead);
				pRead++;
			} while ( --Count );
		}

		halSetProcessExclusivity( FALSE );
	}
}




//---------------------------------------------------------------------------
//	FUNCTION: IndirectSetDisplayAddr()
//
//	DESCRIPTION:
//		This routine sets the Host Interface Address registers to the address
//		specified by addr. For the Serial Interface, this routine also sets
//		the interface to read or write.
//
//	PARAMETERS:
//		addr	- a thirty-two bit value containing the offset to the desired
//				  memory addrsss.
//		SerialDir - for serial indirect I/O, set SerialDir to
//					either SER_WRITE or SER_READ.
//					For parallel indirect I/O, this parameter is ignored.
//---------------------------------------------------------------------------
void IndirectSetDisplayAddr( UInt32 addr, UInt8 SerialDir )
{
	halSetProcessExclusivity( TRUE );

	if (HalInfo.dwFlags & fINDIRECT_SERIAL)
	{
		SetDisplayAddrSerial(addr);

		if (HalInfo.dwFlags & fINDIRECT_ADDR16)
			*HalIndInfo.pIndex16 = SerialDir;
		else
			*HalIndInfo.pIndex8  = SerialDir;
	}
	else
		SetDisplayAddrParallel(addr);

	halSetProcessExclusivity( FALSE );
}


//---------------------------------------------------------------------------
//	FUNCTION: SetSerialAddrParallel()
//
//	DESCRIPTION:
//		This routine sets the Host Interface Address registers to the address
//		specified by addr.
//
//		Because we don't know the last register to be accessed we must write
//		the full address every time.
//
//		This function MUST be called within a process exclusive block!
//
//	PARAMETERS:
//		addr	- a thirty-two bit value containing the offset to the desired
//				  memory addrsss.
//---------------------------------------------------------------------------
static void SetDisplayAddrParallel( UInt32 addr )
{
	// Setup the address
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
	{
		*HalIndInfo.pIndex16 = 0x0020;
		*HalIndInfo.pData8   = LOBYTE(LOWORD(addr));
		*HalIndInfo.pIndex16 = 0x0021;
		*HalIndInfo.pData8   = HIBYTE(LOWORD(addr));
		*HalIndInfo.pIndex16 = 0x0022;
		*HalIndInfo.pData8   = LOBYTE(HIWORD(addr));
		
		*HalIndInfo.pIndex16 = 0x28;
	}
	else
	{
		*HalIndInfo.pIndex8 = 0x20;
		*HalIndInfo.pIndex8 = 0x00;
		*HalIndInfo.pData8  = LOBYTE(LOWORD(addr));
		*HalIndInfo.pIndex8 = 0x21;
		*HalIndInfo.pData8  = HIBYTE(LOWORD(addr));
		*HalIndInfo.pIndex8 = 0x22;
		*HalIndInfo.pData8  = LOBYTE(HIWORD(addr));
		
		*HalIndInfo.pIndex8 = 0x28;
	}
}

//---------------------------------------------------------------------------
//	FUNCTION: SetDisplayAddrSerial()
//
//	DESCRIPTION:
//		This routine sets the Host Interface Address registers to the address
//		specified by addr. After setting the address the routine points the
//		'register pointer' to the Host Interface Memory Access Port.
//
//		Because we don't know the last register to be accessed we must write
//		the full address every time.
//
//		This function MUST be called within a process exclusive block!
//
//	PARAMETERS:
//		addrs	- a thirty-two bit value containing the offset to the desired
//				  memory addrsss.
//---------------------------------------------------------------------------


static void SetDisplayAddrSerial( UInt32 addr )
{
	// Setup the register address of the Host Interface Memory Address register
	if (HalInfo.dwFlags & fINDIRECT_ADDR16)
	{
		*HalIndInfo.pIndex16 = SER_SET_ADDR;
		*HalIndInfo.pData16   = 0x2000;
		
		// Set the LSB of 'addr'
		*HalIndInfo.pIndex16 = SER_WRITE;
		*HalIndInfo.pData8   = LOBYTE(LOWORD(addr));
		
		// Point to the next Host Interface Memory Address register
		*HalIndInfo.pIndex16 = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x21;
		
		// Set the next byte of 'addr'
		*HalIndInfo.pIndex16 = SER_WRITE;
		*HalIndInfo.pData8   = HIBYTE(LOWORD(addr));
		
		// Point to the last Host Interface Memory Address register
		*HalIndInfo.pIndex16 = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x22;
		
		// Set the MSB byte of 'addr'
		*HalIndInfo.pIndex16 = SER_WRITE;
		*HalIndInfo.pData8   = LOBYTE(HIWORD(addr));
		
		// Point to the Host Interface Memory Access Port
		*HalIndInfo.pIndex16 = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x28;
	}
	else
	{
		*HalIndInfo.pIndex8  = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x20;		// REG0020_HOSTMEMADDR0
		*HalIndInfo.pData8   = 0x00;		// MSB of 0x0020 = 0x00
		
		// Set the LSB of 'addr'
		*HalIndInfo.pIndex8  = SER_WRITE;
		*HalIndInfo.pData8   = LOBYTE(LOWORD(addr));
		
		// Point to the next Host Interface Memory Address register
		*HalIndInfo.pIndex8  = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x21;
		
		// Set the next byte of 'addr'
		*HalIndInfo.pIndex8  = SER_WRITE;
		*HalIndInfo.pData8   = HIBYTE(LOWORD(addr));
		
		// Point to the last Host Interface Memory Address register
		*HalIndInfo.pIndex8  = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x22;
		
		// Set the MSB byte of 'addr'
		*HalIndInfo.pIndex8  = SER_WRITE;
		*HalIndInfo.pData8   = LOBYTE(HIWORD(addr));
		
		// Point to the Host Interface Memory Access Port
		*HalIndInfo.pIndex8  = SER_SET_ADDR;
		*HalIndInfo.pData8   = 0x28;
	}
}

