//===========================================================================
//	I2C.C
//
//	This file implements the I2C support for the S1D13713.
//---------------------------------------------------------------------------
//  Copyright (c) 2002 Epson Research and Development, Inc.
//  All Rights Reserved.
//===========================================================================
// (Tabs set to every 4)

//#ifdef _WIN32
//#pragma warning( disable : 4706 )	// Disable "assignment within conditional expression" warning.
//#endif

#include "halapi.h"
#include "hal_regs.h"
#include "i2c.h"

static UInt8	gSlaveDevice;					// I2C Slave Device ID in use

static int		gLastError = I2C_ERR_NONE;		// I2C last error code
static Boolean	gfRepeatedStart = TRUE;			// I2C Repeated Start condition flag for reads
static Boolean	gfAutoIncrement = TRUE;			// I2C Auto Increment mode flag for writes

static Boolean	PrvStartI2C( UInt8 Data );
static Boolean	PrvStopI2C( void );
static Boolean	PrvWriteI2CByte( UInt8 Data, Boolean LastByte );
static Boolean	PrvReadI2CByte( UInt8 *pData, Boolean ACK );


//---------------------------------------------------------------------------
//  FUNCTION:	i2cInitialize()
//
//  DESCRIPTION:
//		Initializes the I2C operating environment.
//
//  PARAMETERS:
//		For backwards and forward compatibility this function still uses
//		the argument list from the GPIO based I2C initialization routine.
//		For use in the 13713 project the caller should set each of these
//		parameter to 0.
//
//  RETURNS:
//		TRUE	- If the initialization is successuful
//		FALSE	- if there is an error during the intialization.
//				- Call 12cGetLastError() or 12cGetLastErrorText() for
//				  more information on the error.
//
//	NOTES:
//		For backwards and forward compatibility this function still uses
//		the argument list from the GPIO based I2C initialization routine
//		and returns a value.
//---------------------------------------------------------------------------
Boolean i2cInitialize( int RegSize, UInt32 RegDir, UInt32 RegData, UInt32 SCLDir, UInt32 SCLData, UInt32 SDADir, UInt32 SDAData )
{
	if (TRUE == i2cInitializeLib())
		if (TRUE == i2cInitializeHW())
			return TRUE;

	return FALSE;

	// Quiet the compiler
	RegSize;	RegDir;		RegData;	SCLDir;		SCLData;	SDADir;	SDAData;
}

//---------------------------------------------------------------------------
//  FUNCTION:	i2cInitializeLib()
//
//  DESCRIPTION:
//		Initialize I2C library only
//
//	RETURNS:
//		TRUE	- For backwards and forward compatibility this function
//				  still returns a value.
//---------------------------------------------------------------------------
Boolean i2cInitializeLib( void )
{
	gLastError = I2C_ERR_NONE;

	gfRepeatedStart = TRUE;
	gfAutoIncrement = TRUE;

	return TRUE;
}

//---------------------------------------------------------------------------
//  FUNCTION:	i2cInitializeHW()
//
//  DESCRIPTION:
//		Performs any HW intialization that needs to be done for I2C.
//
//	RETURNS:
//		TRUE
//
//	NOTES:
//		Enabling the I2C module will enable clocking to the I2C section
//		which will inrease the power used by the system.
//
//		For backwards and forward compatibility this function still uses
//		the argument list from the GPIO based I2C initialization routine.
//		For use in the 13713 project the caller should set each of these
//		parameter to 0.
//---------------------------------------------------------------------------
Boolean i2cInitializeHW( void )
{
	// Clear any error status that may be set and enable the I2C section.
	halWriteReg8(REG0080_I2CSTATUS, 0xC0);

	// Delay the data read time 
//	halWriteReg8(REG008A_I2CREADACK, 4);

	return TRUE;
}


//---------------------------------------------------------------------------
//  FUNCTION: 	i2cSetSlaveDevice()
//
//  DESCRIPTION:
//		Set Slave device ID which will be used for all subsequent
//		I2C operations.
//
//  PARAMETERS:
//		SlaveDevice		- Device ID for the new slave device to communicate
//						  with
//
//  RETURNS: 
//		The device ID of the previous slave device
//---------------------------------------------------------------------------
UInt8 i2cSetSlaveDevice( UInt8 SlaveDevice )
{
	UInt8 LastID = gSlaveDevice;
	gSlaveDevice = SlaveDevice;

	gLastError = I2C_ERR_NONE;

	return LastID;
}


//---------------------------------------------------------------------------
//  FUNCTION: 	i2cGetLastError()
//
//  DESCRIPTION:
//		Returns the numerice value of the last I2C error.
//
//  RETURNS: 
//		Integer value of the last I2C error.
//---------------------------------------------------------------------------
int i2cGetLastError( void )
{
	return gLastError;
}


//---------------------------------------------------------------------------
//  FUNCTION:	i2cGetLastErrorText()
//
//  DESCRIPTION:
//		Returns a pointer to the text representation of the last I2C error.
//
//  RETURNS: 
//		Pointer to the first character of the ASCIIZ string describing
//		the last I2C error.
//
//	NOTE:
//		It is imperative to maintain this list in conjuction with the list
//		of error values in i2c.h.
//---------------------------------------------------------------------------
const char * i2cGetLastErrorText( void )
{
	static const char * const apszErrors[] =
	{
		"No error",					// I2C_ERR_NONE
		"SCL/SDA line stuck low",	// I2C_ERR_STUCK_BIT
		"NAK device ID",			// I2C_ERR_SLAVE_NACK_ID
		"NAK device data",			// I2C_ERR_SLAVE_NACK_DATA
		"Invalid argument",			// I2C_ERR_INVALID_ARGUMENT
		"Unknown error"				// I2C_ERR_UNKNOWN_ERROR - MUST BE LAST!!!
	};

	if ((gLastError < 0) || (gLastError > I2C_ERR_UNKNOWN_ERROR))
		return apszErrors[I2C_ERR_UNKNOWN_ERROR];

	return apszErrors[gLastError];
}


//---------------------------------------------------------------------------
//  FUNCTION:	i2cSetAutoIncrement()
//
//  DESCRIPTION:
//		Set I2C Auto Increment mode for writes.
//
//  PARAMETERS:
//		fAutoIncrement	- flag containing the new auto-increment value
//						  fAutoIncrement == TRUE,  enables auto-incrment
//						  fAutoIncrement == FALSE, disables auto-incrment
//  RETURNS: 
//		Previous auto-increment state.
//---------------------------------------------------------------------------
Boolean i2cSetAutoIncrement( Boolean fAutoIncrement )
{
	Boolean fLastAI = gfAutoIncrement;
	gfAutoIncrement = fAutoIncrement;

	gLastError = I2C_ERR_NONE;

	return fLastAI;
}


//---------------------------------------------------------------------------
//  FUNCTION:	i2cSetRepeatedStart()
//
//  DESCRIPTION:
//		Set I2C Repeated Start condition flag for reads.
//
//  PARAMETERS:
//		fRepeatedStart	- This flag contols the 
//
//  RETURNS: 
//		Previous state of the repeated start flag
//---------------------------------------------------------------------------
Boolean i2cSetRepeatedStart( Boolean fRepeatedStart )
{
	Boolean fLastRS = gfRepeatedStart;
	gfRepeatedStart = fRepeatedStart;

	gLastError = I2C_ERR_NONE;

	return fLastRS;
}


//---------------------------------------------------------------------------
//  FUNCTION:	i2cWriteBytes()
//
//  DESCRIPTION:
//		Write byte data to the I2C bus.
//
//  PARAMETERS:
//		pBytes	- Pointer to an array of 8-bit values to sent to the I2C bus
//		nBytes	- Number of bytes to send to the I2C bus
//
//  RETURNS: 
//		TRUE	- if the function is successful in writing the byte strings
//				  to the I2C bus.
//		FALSE	- if there is an error during the transfer.
//				- Call 12cGetLastError() or 12cGetLastErrorText() for
//				  more information on the error.
//---------------------------------------------------------------------------
Boolean i2cWriteBytes( const UInt8 *pBytes, int nBytes )
{
	int i;
	Boolean ret = TRUE;

	gLastError = I2C_ERR_NONE;

	if (!gfAutoIncrement && (nBytes > 1))		// Auto Increment clear (and at least 2 data bytes)
	{
		UInt8 SubAddress = *pBytes++;
		for (i = 1; ret && i < nBytes; i++)
		{
			ret = PrvStartI2C(gSlaveDevice);
			if (ret)
			{
				ret = PrvWriteI2CByte(SubAddress++, FALSE);
				if (ret)
					ret = PrvWriteI2CByte(*pBytes++, i == (nBytes - 1));
			}
			else if (gLastError == I2C_ERR_SLAVE_NACK_DATA)
				gLastError = I2C_ERR_SLAVE_NACK_ID;

			PrvStopI2C();						// Must always be able to STOP the I2C bus!
		}
	}
	else										// Auto Increment set
	{
		ret = PrvStartI2C(gSlaveDevice);
		if (ret)
		{
			for (i = 0; i < nBytes; i++)
			{
				ret = PrvWriteI2CByte(*pBytes, i == (nBytes - 1));
				if (FALSE == ret)
					break;
				pBytes++;
			}
		}
		else if (gLastError == I2C_ERR_SLAVE_NACK_DATA)
			gLastError = I2C_ERR_SLAVE_NACK_ID;

		PrvStopI2C();					// Must always be able to STOP the I2C bus!
	}

	return ret;
}


//---------------------------------------------------------------------------
//  FUNCTION:	i2cWrite2Bytes()
//
//  DESCRIPTION:
//		Write 2 bytes of data to the I2C bus.
//
//  PARAMETERS:
//		Byte1	- first 8-bit value to write to the I2C bus
//		Byte2	- second 8-bit value to write tot the I2C bus
//
//  RETURNS: 
//		TRUE	- if the function is successful in writing the bytes
//				  to the I2C bus.
//		FALSE	- if there is an error during the transfer.
//				- Call 12cGetLastError() or 12cGetLastErrorText() for
//				  more information on the error.
//---------------------------------------------------------------------------
Boolean i2cWrite2Bytes( UInt8 Byte1, UInt8 Byte2 )
{
	UInt8 aBytes[2];

	aBytes[0] = Byte1;
	aBytes[1] = Byte2;

	return i2cWriteBytes(aBytes, 2);
}


//---------------------------------------------------------------------------
//  FUNCTION:	i2cReadBytes()
//
//  DESCRIPTION:
//		Reads a string of bytes from the I2C bus and places them in a buffer.
//
//  PARAMETERS:
//		pSubAddress	-
//		pBytes		- 
//		nBytes		- Maximum number of bytes to read from the device
//
//  RETURNS: 
//		TRUE	- if the function is successful in read the bytes stream
//				  from the I2C bus.
//		FALSE	- if there is an error during the transfer.
//				- Call 12cGetLastError() or 12cGetLastErrorText() for
//				  more information on the error.
//---------------------------------------------------------------------------
Boolean i2cReadBytes( const UInt8 *pSubAddress, UInt8 *pBytes, int nBytes )
{
	int i;
	Boolean ret = TRUE;

	gLastError = I2C_ERR_NONE;

	if (pSubAddress)
	{
		ret = PrvStartI2C(gSlaveDevice);
		if (ret)
		{
			ret = PrvWriteI2CByte(*pSubAddress, FALSE);
			if (ret && !gfRepeatedStart)
				PrvStopI2C();								// Repeated Start condition not supported
		}
		else if (gLastError == I2C_ERR_SLAVE_NACK_DATA)
			gLastError = I2C_ERR_SLAVE_NACK_ID;
	}

	if (ret)
	{
		ret =  PrvStartI2C((UInt8)(gSlaveDevice | 0x01));	// This may be a repeated-START condition.
		if (ret)
		{
			for (i = 1; i <= nBytes; i++)
			{
				ret = PrvReadI2CByte(pBytes, (i != nBytes));
				if (!ret)
					break;
				pBytes++;
			}
		}
		else if (gLastError == I2C_ERR_SLAVE_NACK_DATA)
			gLastError = I2C_ERR_SLAVE_NACK_ID;

	}
	PrvStopI2C();											// Must always be able to STOP the I2C bus!

	return ret;
}

//===========================================================================

//---------------------------------------------------------------------------
//  FUNCTION:	PrvStartI2C()
//
//  DESCRIPTION:
//		This routine sends an I2C Start sequence with data.
//
//  RETURNS: 
//		TRUE
//
//	NOTES:
//		So far, in every case I've seen, the data transmitted with START has
//		always been the slave ID.
//		- Is this always the case?
//---------------------------------------------------------------------------
static Boolean PrvStartI2C( UInt8 Data )
{
	halWriteReg8(REG0088_I2CWRITE,     Data);	// Set the data value to write.
	halWriteReg8(REG0084_I2CSTARTSTOP, 0x03);	// Set for Start with Data
	halWriteReg8(REG0082_I2CCOMMAND,   0x01);	// Go

	halWriteReg8(REG0084_I2CSTARTSTOP, 0x00);	// Set the REG[0084] value for data writes

	// Check for ACK/NAK (but first wait for the TIP bit to clear)
	while (halReadReg8(REG0080_I2CSTATUS) & 0x08)
		NULL;
	if (0x02 == (halReadReg8(REG0080_I2CSTATUS) & 0x02))
	{
		gLastError = I2C_ERR_SLAVE_NACK_DATA;
		return FALSE;
	}

	return TRUE;
}


//---------------------------------------------------------------------------
//  FUNCTION:	PrvStopI2C()
//
//  DESCRIPTION:
//		This routine sends an I2C Stop command to the I2C bus.
//
//		Issuing a STOP when the I2C bus is already stopped will generate
//		an error (REG[0080] bit 6). To prevent this condition from
//		occuring this routine looks at I2C Busy REG[0080] bit 2. If the
//		I2C is not busy then the STOP is not set.
//
//  RETURNS: 
//		TRUE
//---------------------------------------------------------------------------
static Boolean PrvStopI2C( void )
{
	if (0x04 == (halReadReg8(REG0080_I2CSTATUS) & 0x04))
	{
		// Set STOP without data and execute the command
		halWriteReg8(REG0084_I2CSTARTSTOP, 0x05);
		halWriteReg8(REG0082_I2CCOMMAND,   0x01);
	}

	return TRUE;
}


//---------------------------------------------------------------------------
//  FUNCTION:	PrvWriteI2CByte()
//
//  DESCRIPTION:
//		This routine sends one byte over the I2C bus.
//
//  PARAMETERS:
//		Data	- the eight bit value to send over I2C
//
//  RETURNS: 
//		TRUE	- if the receiving device returns an ACK in response to
//				  receiving the data.
//		FALSE	- if the receiving decice returns a NAK.
//---------------------------------------------------------------------------
static Boolean PrvWriteI2CByte( UInt8 Data, Boolean LastByte )
{
	Boolean RetVal = TRUE;

	// Write the data to the data register and start the transfer.
	halWriteReg8(REG0088_I2CWRITE, Data);		// Write the data to the data register
	if (LastByte)
		halWriteReg8(REG0084_I2CSTARTSTOP, 0x01);
	else
		halWriteReg8(REG0084_I2CSTARTSTOP, 0x00);
	halWriteReg8(REG0082_I2CCOMMAND, 0x01);		// Go

	// Get ACK/NAK
	// Wait for I2C TIP bit to go low then read the I2CRxAck bit
	while (halReadReg8(REG0080_I2CSTATUS) & 0x08)
		NULL;
	if (0x02 == (halReadReg8(REG0080_I2CSTATUS) & 0x02))
	{
		gLastError = I2C_ERR_SLAVE_NACK_DATA;
		RetVal = FALSE;
	}

	return RetVal;
}


//---------------------------------------------------------------------------
//  FUNCTION:	PrvReadI2CByte()
//
//  DESCRIPTION:
//		Reads one byte from the I2C bus.
//
//  PARAMETERS:
//
//  RETURNS: 
//---------------------------------------------------------------------------
static Boolean PrvReadI2CByte( UInt8 *pData, Boolean ACK )
{
	if (ACK)
		halWriteReg8(REG0082_I2CCOMMAND, 0x03);
	else
	{
		halWriteReg8(REG0084_I2CSTARTSTOP, 0x01);
		halWriteReg8(REG0082_I2CCOMMAND, 0x07);
	}

	// Wait for I2C TIP bit to go low then read the data register.
	while (halReadReg8(REG0080_I2CSTATUS) & 0x08)
		NULL;
	*pData = halReadReg8(REG0086_I2CREAD);

	// ****
	// RELE - We should be able to read some sort of transfer status here.
	//        Find out from Shon what the bits to read are.
	// ****
	
	return TRUE;
}
