//===========================================================================
//	JCAMENCODE.C
//	This file contains JPEG functions implemented for the JPEG library to 
//	implement JPEG encoding via Agilent camera.
//---------------------------------------------------------------------------
//  Copyright (c) 2003 Epson Research and Development, Inc.
//  All Rights Reserved.
//===========================================================================

//---------------------------------------------------------------------------
// Include Files
//---------------------------------------------------------------------------
#include <stdio.h>

#include "multiframe.h"
#include "hal.h"
#include "chiplib.h"
#include "jpegcomm.h"
#include "JpegTable.h"

//---------------------------------------------------------------------------
// Local Data
//---------------------------------------------------------------------------
static UInt16 gCamEncodeRegIndex[] =
{
	REG0102_CMSIGNAL	,
	REG0108_CMCONTROL	,
	REG003C_LCDXFER	,
	REG0224_YUVWRSTART0,
};

//---------------------------------------------------------------------------
// Local Type Definitions
//---------------------------------------------------------------------------
#define CAMENC_REG_SIZE	(sizeof(gCamEncodeRegIndex) / sizeof( UInt16))

typedef struct tagStartAddr
{
	UInt32 m_Cur;
	UInt32 m_First;
	UInt32 m_Prev;
} prvStartAddr;

typedef struct tagCamEncodeState
{
	UInt8			m_AutoFrameXfer;
	UInt32			m_BufSize;
	Boolean			m_fCamClk;
	UInt8			m_CamControl;
	UInt8			m_CamSignal;
	UInt32			m_Count;
	Boolean			m_fYUV;
	JPEGMULTIFRAME*	m_pMulFrm;
	Boolean			m_fNoTimeOut;
	UInt8*			m_pStart;
	prvStartAddr	m_StartAddr;
	UInt32			m_TotalSize;
	UInt8			m_aSaveReg[ CAMENC_REG_SIZE ];
	Boolean			m_fUnWind;
} prvCamEncodeState;

//---------------------------------------------------------------------------
// Local Function Prototypes
//---------------------------------------------------------------------------
static Boolean	prvAccessAgilent( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState );
static Boolean	prvInitCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState );
static Boolean	prvAppendHeader( JPEGINFO* pJpegInfo );
static Boolean	prvIsCaptureComplete( void );
static Boolean	prvStartCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState, Boolean* pComplete );
static UInt8	prvStopCapture( void );
static Boolean	prvWaitForCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState );
static Boolean	prvCopy2Jpeg( UInt8** ppDest, const void* pSource, UInt32 size, UInt8* pEnd );
static Boolean	prvAcquireRawJpeg( JPEGINFO* pJpegInfo, UInt32 startAddr );
static Boolean	prvHasDataCapInt( void );
static Boolean	prvSetJpegMode( JPEGINFO *pJpegInfo, prvCamEncodeState* pCamEncState );
static Boolean	prvInitNextCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState );
static Boolean	prvInitAnyCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState );
static void		prvRestoreReg( UInt8* pSaveReg );
static void		prvRestoreState( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState, Boolean fSuccess );
static void		prvSaveReg( UInt8* pSaveReg );
static Boolean	prvSelectQTable( JPEGINFO* pJpegInfo );
static void		prvAdvanceStartAddr( JPEGINFO* pJpegInfo, prvStartAddr* pStartAddr );
static void		prvSetStartAddrImg( UInt32 address );
static void		prvInitStartAddr( JPEGINFO* pJpegInfo, prvStartAddr* pStartAddr );
static Boolean	prvValidateParam( JPEGINFO* pJpegInfo );
static void		prvUnWind( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState );

//------------------------------------------------------------------------------
//	jpegCamEncode - fSuccesss camera image into valid JPEG image
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//
//	Return:	TRUE successful
//			FALSE if error.
//
//------------------------------------------------------------------------------
Boolean jpegCamEncode( JPEGINFO* pJpegInfo )
{
	Boolean fSuccess = jpegValidateBuf( pJpegInfo );
	if ( fSuccess )
	{
		prvCamEncodeState camEncodeState;
		fSuccess = prvInitCapture( pJpegInfo, &camEncodeState );

		// Capture for each image specified
		for 
		( 
			camEncodeState.m_Count = 1; 
			fSuccess && (camEncodeState.m_Count <= pJpegInfo->m_FrameCount); 
			++camEncodeState.m_Count 
		)
		{
#ifndef MV850E
			jpegFindTimeOut( pJpegInfo );
#endif
			camEncodeState.m_pMulFrm = (JPEGMULTIFRAME*)pJpegInfo->m_pBuf;
			if ( pJpegInfo->m_FrameCount > 1 )
			{
				pJpegInfo->m_pBuf		+= sizeof( JPEGMULTIFRAME );
				pJpegInfo->m_BufSize	-= sizeof( JPEGMULTIFRAME );
			}

			if ( pJpegInfo->m_fVerbose )
			{	
				UInt32 resizeXEnd;
				UInt32 resizeYEnd;
				UInt32 resizeXStart;
				UInt32 resizeYStart;
				UInt32 XStart;
				UInt32 YStart;
				UInt32 width;
				UInt32 height;

				GetResizerSize( &resizeXStart, &resizeYStart, &resizeXEnd, &resizeYEnd );
				AgilentDimensionGet( &XStart, &YStart, &width, &height );

				//printf( "       AGILENT   RESIZE\n" );
				//printf( "XStart  %03u      %03u\n", XStart, resizeXStart );
				//printf( "YStart  %03u      %03u\n", YStart, resizeXStart );
				//printf( "XEnd    %03u      %03u\n", width - 1, resizeXEnd );
				//printf( "YEnd    %03u      %03u\n", height - 1, resizeYEnd );
			}


			// Allow capture until notified that it is complete
			fSuccess = prvWaitForCapture( pJpegInfo, &camEncodeState );
			if ( fSuccess )
			{
				// Retrieve JPEG data from memory
				fSuccess = prvAcquireRawJpeg( pJpegInfo, camEncodeState.m_StartAddr.m_Cur );
				if ( fSuccess )
				{
					// Append headers from list of tables
					fSuccess = prvAppendHeader( pJpegInfo );
					if ( fSuccess )
					{
						// Prepare for next image to capture
						fSuccess = prvInitNextCapture( pJpegInfo, &camEncodeState );
					}
				}
			}
		}

		prvRestoreState( pJpegInfo, &camEncodeState, fSuccess );
	}
	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvAcquireRawJpeg - retrieves JPEG data from vide memory
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			startAddr		start address of data
//
//	Return:	TRUE means valid image
//			FALSE means NOT valid image
//
//------------------------------------------------------------------------------
Boolean prvAcquireRawJpeg( JPEGINFO* pJpegInfo, UInt32 startAddr )
{
	Boolean fSuccess	= TRUE;
	Boolean fEndOfImage = FALSE;
	UInt8	prevData	= 0;

	if ( pJpegInfo->m_fVerbose )
	{
		//printf( "ImageAddress=%XL\n", startAddr );
	}

	for 
	( 
		pJpegInfo->m_EncSize = 0; 
		(pJpegInfo->m_EncSize < pJpegInfo->m_BufSize) && (!fEndOfImage) && fSuccess; 
		++pJpegInfo->m_EncSize 
	)
	{
		UInt8 data									= halReadDisplay8( startAddr + pJpegInfo->m_EncSize );
		pJpegInfo->m_pBuf[ pJpegInfo->m_EncSize ]	= data;
		if ( prevData == 0xFF )
		{
			if ( data == 0xD9 )
			{
				++pJpegInfo->m_EncSize;
				fSuccess = (pJpegInfo->m_EncSize > 4 + 2) ? TRUE : FALSE;
				if ( !fSuccess )
				{
					jpegSetLastError( JPEG_ERR_INVALID_JPEG_DATA );
				}
				else
				{
					fEndOfImage = TRUE;
				}
			}
		}

		prevData = data;
	}

	if ( fSuccess && (!fEndOfImage) )
	{
		static aCode[] = {0xFF, 0xDA, 0x00, 63};
		fSuccess = (memcmp( aCode, pJpegInfo->m_pBuf, sizeof( aCode ) ) == 0) ? TRUE : FALSE;
		if ( !fSuccess )
		{
			jpegSetLastError( JPEG_ERR_ENCSIZELIMIT );
		}
		else
		{
			jpegSetLastError( JPEG_ERR_INVALID_JPEG_DATA );
			fSuccess = FALSE;
		}
	}

	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvInitCapture - Saves and prepares hardware and software for encoding
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to camEncodeState object
//
//	Return:	TRUE means successfully appended JPEG tables to image
//			FALSE means NOT successfully appended JPEG tables to image
//
//------------------------------------------------------------------------------
Boolean prvInitCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
	Boolean fSuccess;

	prvSaveReg( pCamEncState->m_aSaveReg );

#ifndef MV850E
	jpegFindTimeOut( pJpegInfo );
#endif

	// Prepare to fSuccess JPEG image
	pCamEncState->m_fUnWind = fSuccess = prvInitAnyCapture( pJpegInfo, pCamEncState );
	if ( fSuccess )
	{
		// Set Agilent Camera in JPEG mode
		fSuccess = prvSetJpegMode( pJpegInfo, pCamEncState );
		if ( fSuccess )
		{
			halDelayUS( pJpegInfo->m_Delay * 1000 );

			// CMHREF & CMVREF Active High
			halWriteReg8( REG0102_CMSIGNAL, 0x06 );
			//76543210b
			//|||||||+--Valid Input Clock Edge
			//||||||+---CMVREF Active Select 1=Active Hi 0=Active Lo
			//|||||+----CMHREF Active Select 1=Active Hi 0=Active Lo
			//|||++-----YUV Data Format Select
			//||+-------Clock Mode Select
			//|+--------Reserved
			//+---------n/a

			// Camera Data Type=YUV Camera Stop after next CMVREF Repeat Capture enabled
			halWriteReg8( REG0108_CMCONTROL, 0x1C );
			//76543210b
			//|||||||+--CMCLK Output Hold Reset
			//||||||+---Capture Frame
			//|||||+----Repeat Capture Frame
			//||||+-----Camera Capture Stop
			//|||+------Camera Data Type 1=YUV 0=JPEG
			//+++-------n/a

			// Camera Data Type=JPEG Camera Stop after next CMVREF Repeat Capture enabled
			halWriteReg8( REG0108_CMCONTROL, 0x0C );

			// Camera Data Type=JPEG Repeat Capture enabled
			halWriteReg8( REG0108_CMCONTROL, 0x04 );

			fSuccess = prvSelectQTable( pJpegInfo );
			if ( fSuccess )
			{
				// Reverse 13711 HREF/VREF polarities
				pCamEncState->m_CamSignal = halReadReg8( REG0102_CMSIGNAL );
				//76543210b
				//|||||||+--Valid Input Clock Edge
				//||||||+---CMVREF Active Select 1=Active Hi 0=Active Lo
				//|||||+----CMHREF Active Select 1=Active Hi 0=Active Lo
				//|||++-----YUV Data Format Select
				//||+-------Clock Mode Select
				//|+--------Reserved
				//+---------n/a

				// CMHREF & CMVREF Active High
				halWriteReg8( REG0102_CMSIGNAL, (UInt8)(pCamEncState->m_CamSignal | 0x06) );

				prvInitStartAddr( pJpegInfo, &pCamEncState->m_StartAddr );
			}
		}
	}

	pCamEncState->m_pStart		= pJpegInfo->m_pBuf;
	pCamEncState->m_BufSize		= pJpegInfo->m_BufSize;
	pCamEncState->m_TotalSize	= 0;
	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvAppendHeader - reorganizes JPEG image to include appropriate JPEG headers
//
//	Input:	pJpegInfo	pointer to JPEGINFO object
//
//	Return:	TRUE means successfully appended JPEG tables to image
//			FALSE means NOT successfully appended JPEG tables to image
//
//------------------------------------------------------------------------------
Boolean prvAppendHeader( JPEGINFO* pJpegInfo )
{
#define QUANT_SIZE (3 * (sizeof( QUANTIZATIONHEADER ) + 64))
#define HEADER_SIZE (sizeof( g_JpegHeader) + QUANT_SIZE + sizeof( REMAINDERINFO ))
	UInt32	size		= HEADER_SIZE + pJpegInfo->m_EncSize - 4;
	Boolean fSuccess	= (size <= pJpegInfo->m_BufSize) ? TRUE : FALSE;
	if ( !fSuccess )
	{
		jpegSetLastError( JPEG_ERR_ENCSIZELIMIT );
	}
	else
	{
		UInt8*	pStartJpeg	= (UInt8*)malloc( size + 4 );
		fSuccess			= (pStartJpeg != NULL) ? TRUE : FALSE;
		if ( !fSuccess )
		{
			jpegSetLastError( JPEG_ERR_ENCSIZELIMIT );
		}
		else
		{
			UInt8			iQTable = *(pJpegInfo->m_pBuf + 3) - 1;
			const UInt8*	pQTable = &g_aQTable[ iQTable ][ 0 ];
			UInt8*			pDest	= pStartJpeg;
			UInt8*			pEnd	= pDest + (size + (4 - 1));

			if ( pJpegInfo->m_fVerbose )
			{
				//printf( "QTable:%u\n", iQTable + 1 );
			}

			// Get static portion of header
			fSuccess = prvCopy2Jpeg( &pDest, &g_JpegHeader[ 0 ], sizeof( g_JpegHeader ), pEnd );
			if ( fSuccess )
			{
				// Now select each QTable
				for 
				( 
					g_QuantizationHeader.m_Identifier = 0; 
					(g_QuantizationHeader.m_Identifier < 3) && fSuccess; 
					++g_QuantizationHeader.m_Identifier 
				)
				{
					fSuccess = prvCopy2Jpeg( &pDest, &g_QuantizationHeader, sizeof( g_QuantizationHeader ), pEnd );
					if ( fSuccess )
					{
						fSuccess = prvCopy2Jpeg( &pDest, pQTable, 64, pEnd );
					}
				}

				if ( fSuccess )
				{
					UInt32 width;
					UInt32 height;

					fSuccess = AgilentResGet( &width, &height );
					if ( !fSuccess )
					{
						jpegSetLastError( JPEG_ERR_AGILENT_ERROR );
					}
					else
					{
						g_RemainderInfo.m_FrameHeader.m_Rows[0]    = (UInt8)((height >> 8) & 0xff);
						g_RemainderInfo.m_FrameHeader.m_Rows[1]    = (UInt8) (height & 0xff);
						g_RemainderInfo.m_FrameHeader.m_Columns[0] = (UInt8)((width >> 8) & 0xff);
						g_RemainderInfo.m_FrameHeader.m_Columns[1] = (UInt8) (width & 0xff);

						fSuccess = prvCopy2Jpeg( &pDest, &g_RemainderInfo, sizeof( g_RemainderInfo ), pEnd );
						if ( fSuccess )
						{
							// Skip over first 4 bytes
							fSuccess = prvCopy2Jpeg( &pDest, (pJpegInfo->m_pBuf + 4), pJpegInfo->m_EncSize - 4, pEnd );
							if ( fSuccess )
							{
								memcpy( pJpegInfo->m_pBuf, pStartJpeg, size );
								free( pStartJpeg );
								pJpegInfo->m_EncSize	= size;
								fSuccess				= jpegHasImageMarkers( pJpegInfo );
							}
						}
					}
				}
			}
		}
	}

	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvStartCapture - start JPEG capture in 13711 hardware
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//			pComplete		pointer to complete flag
//			
//
//	Return:	TRUE if success (whether capture complete or not)
//			FALSE if NOT success
//
//------------------------------------------------------------------------------
Boolean prvStartCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState, Boolean* pComplete )
{
	Boolean fSuccess = TRUE;

	// Clear any existing interrupt status
	if ( prvHasDataCapInt() )
	{
		halWriteReg8( REG0300_IRQSTATUS0, 0x10 );
		halWriteReg8( REG0300_IRQSTATUS0, 0x00 );
	}

	// Capture JPEG data from camera to video memory
	// Capture Frame but no repeat capture enable
	halWriteReg8( REG0108_CMCONTROL, 0x02 );
	//76543210b
	//|||||||+--CMCLK Output Hold/Reset
	//||||||+---Capture Frame
	//|||||+----Repeat Capture Enable
	//||||+-----Capture Stop
	//|||+------Camera Data Type
	//+++-------n/a

	// Wait till capture complete
	while ( !prvHasDataCapInt() && fSuccess )
	{
#ifndef MV850E
		fSuccess = jpegNoTimeOut( pJpegInfo );
#endif
	}

	*pComplete = prvIsCaptureComplete();
	return( fSuccess );
}

//------------------------------------------------------------------------------
//	prvStopCapture - stop current capture in 13711 hardware
//
//	Input:	n/a
//
//	Return:	previous CameraControl
//
//------------------------------------------------------------------------------
UInt8 prvStopCapture( void )
{
	UInt8 camControl = halReadReg8( REG0108_CMCONTROL );
	//76543210b
	//|||||||+--CMCLK Output Hold/Reset
	//||||||+---Capture Frame
	//|||||+----Repeat Capture Enable
	//||||+-----Capture Stop
	//|||+------Camera Data Type
	//+++-------n/a
	UInt8 newCamControl = (UInt8)((camControl & ~0x02) | 0x08);
	halWriteReg8( REG0108_CMCONTROL, newCamControl );

	// Set 13711 to JPEG mode
	halWriteReg8( REG0108_CMCONTROL, 0x08 );

	// Stop Camera Capture must be set to 0 before starting capture
	halWriteReg8( REG0108_CMCONTROL, 0x00 );

	return ( camControl );
}

//------------------------------------------------------------------------------
//	prvCopy2Jpeg - copies data to JPEG image
//
//	Input:	ppDest		pointer to pointer to destination buffer
//			pSource		pointer to source data
//			size		number of bytes of source data to copy
//			pEnd		end marker
//
//	Return:	true means success
//			false means NO success
//
//------------------------------------------------------------------------------
Boolean prvCopy2Jpeg( UInt8** ppDest, const void* pSource, UInt32 size, UInt8* pEnd )
{
	UInt8*	pDest		= *ppDest;
	Boolean fSuccess	= (Boolean)(pDest + size <= pEnd);
	if ( !fSuccess )
	{
		jpegSetLastError( JPEG_ERR_ENCSIZELIMIT );
	}
	else
	{
		memcpy( pDest, pSource, size );
		pDest	+= size;
		*ppDest = pDest;
	}
	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvHasDataCapInt - determines if Camera Data Capture Interrupt occurred
//
//	Input:	n/a
//
//	Return:	TRUE means camera capture complete
//			FALSE means camera capture NOT complete
//
//------------------------------------------------------------------------------
Boolean prvHasDataCapInt( void )
{
	UInt8	status		= halReadReg8( REG0300_IRQSTATUS0 );
	Boolean fIsCapture	= (status & 0x10) ? TRUE : FALSE;
	return ( fIsCapture );
}

//------------------------------------------------------------------------------
//	prvAccessAgilent - Tests to see if Agilent is responding
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//
//	Return:	TRUE means success
//			FALSE means NOT successful
//
//------------------------------------------------------------------------------
Boolean prvAccessAgilent( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
	UInt32	loop = 0;
	Boolean fSuccess;

	do
	{
		fSuccess = AgilentIsYUV( &pCamEncState->m_fYUV );
		if ( !fSuccess )
		{
			halDelayUS( 200 * 1000 );
			++loop;
#ifdef MV850E
			pCamEncState->m_fNoTimeOut = (Boolean)(loop <= 100);
#else
			pCamEncState->m_fNoTimeOut = jpegNoTimeOut( pJpegInfo );
#endif
			if ( pJpegInfo->m_fVerbose )
			{
				//printf( "Agilent Delay Loop:%u\n", loop );
			}
		}
	}
	while ( (!fSuccess) && pCamEncState->m_fNoTimeOut );

	if ( !fSuccess )
	{
		jpegSetLastError( JPEG_ERR_AGILENT_ERROR );
	}
	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvIsCaptureComplete - Determines if complete JPEG capture has occured
//
//	Input:	n/a
//
//	Return:	TRUE if capture copmlete
//			FALSE if capture truncated
//
//------------------------------------------------------------------------------
Boolean prvIsCaptureComplete( void )
{
	UInt8 cameraStatus = halReadReg8( REG010A_CMSTATUS );
	//76543210b
	//|||||||+--Camera Interface Status
	//||||||+---JPEG Capture Data Status (RO)
	//++++++----n/a
	Boolean fComplete = (cameraStatus & 0x02) ? TRUE : FALSE;
	return ( fComplete );
}

//------------------------------------------------------------------------------
//	prvSetJpegMode - sets camera into JPEG mode
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//
//	Return:	TRUE means JPEG mode set successfully
//			FALSE means JPEG mode NOT set successfully
//
//------------------------------------------------------------------------------
Boolean prvSetJpegMode( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
	Boolean fSuccess = prvAccessAgilent( pJpegInfo, pCamEncState );
	if ( fSuccess )
	{
		fSuccess = AgilentJpegTimeOut( pJpegInfo->m_TimeOutTgt, pJpegInfo->m_TimeOutMS );
		if ( !fSuccess )
		{
			jpegSetLastError( JPEG_ERR_AGILENT_ERROR );
		}
		else
		{
			fSuccess = prvAccessAgilent( pJpegInfo, pCamEncState );
			if ( fSuccess )
			{
				fSuccess = AgilentResSet( (UInt32)pJpegInfo->m_Width, (UInt32)pJpegInfo->m_Height );
			    if ( !fSuccess )
				{
					jpegSetLastError( JPEG_ERR_DIM_RESTRICT );
				}
			}
		}

		if ( pJpegInfo->m_fVerbose )
		{
			AgilentDump( FALSE );
			AgilentDump( TRUE );
		}
	}

	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	prvInitNextCapture - program structures for next image
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//
//	Output:	TRUE if valid parameters
//			FALSE if NOT valid paramaters
//
//-----------------------------------------------------------------------------
Boolean prvInitNextCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
	Boolean fSuccess = TRUE;

	prvAdvanceStartAddr( pJpegInfo, &pCamEncState->m_StartAddr );
	
	if ( pJpegInfo->m_FrameCount > 1 )
	{
		fSuccess = (pJpegInfo->m_EncSize <= pJpegInfo->m_BufSize) ? TRUE : FALSE;
		if ( !fSuccess )
		{	
			jpegSetLastError( JPEG_ERR_ENCSIZELIMIT );
		}
		else
		{
			mfInitializeMultiFrame( pCamEncState->m_pMulFrm, pJpegInfo->m_EncSize );
			pJpegInfo->m_pBuf						+= pCamEncState->m_pMulFrm->PaddedSize;
			pJpegInfo->m_BufSize					-= pCamEncState->m_pMulFrm->PaddedSize;
			pCamEncState->m_pMulFrm->FrameNumber	= pCamEncState->m_Count;
			pCamEncState->m_TotalSize				+= (sizeof( JPEGMULTIFRAME ) + pCamEncState->m_pMulFrm->PaddedSize);
		}
	}
	else
	{
		pJpegInfo->m_BufSize		-= pJpegInfo->m_EncSize;
		pJpegInfo->m_pBuf			+= pJpegInfo->m_EncSize;
		pCamEncState->m_TotalSize	+= pJpegInfo->m_EncSize;
	}

	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	prvSelectQTable - Select quality table for Agilent camera
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//
//	Output:	TRUE if successful
//			FALSE if NOT successful
//
//-----------------------------------------------------------------------------
Boolean prvSelectQTable( JPEGINFO* pJpegInfo )
{
	UInt16	mainCfg2; 
	//Boolean	fSuccess = AgilentProcessorRead( apr04_CFG_MAIN2, &mainCfg2 );
	//FEDCBA9876543210b
	//|||||||||||||||+--Pixel Correction
	//||||||||||||||+---Sizer Control
	//|||||||||||||+----Demosiac operation control
	//||||||||||||+-----JPEG Compression control
	//|||||||||||+------Auto Configure Image Sensor
	//||||||||||+-------ALQ Auto load quantization table RAM
	//|||||||||+--------AQA Auto Q Adjust
	//||||||||+---------Auto load gamma correction lookup table RAM
	//|||||||+----------Sign of Cr bit - MCU (minimum coded unit)
	//||||||+-----------Sign of Cb bit - MCU (minimum coded unit)
	//|||||+------------Sign of Y bit (minimum coded unit)
	//||||+-------------Compression Control Mode
	//|||+--------------Frame Header /trailer disable
	//||+---------------Unused
	//|+----------------Soft nReset for sensor
	//+-----------------Soft nReset

	if ( fSuccess )
	{
		if ( (pJpegInfo->m_QTable >= 0) && (pJpegInfo->m_QTable <= 63) )
		{
			// Using Fixed Q-Tables Page 57 of Agilent ADCM-1650 CIF Reslution CMOS Camera Module
			// Disable Auto Q Adjust
			mainCfg2 &= ~0x0040;
			fSuccess = AgilentProcessorWrite( apr04_CFG_MAIN2, mainCfg2 );
			if ( fSuccess )
			{
				// Select QTable
				fSuccess = AgilentProcessorWrite( apr62_QTABLE_SELECT, pJpegInfo->m_QTable );
				if ( fSuccess )
				{
					// Set ALQ to load JPEG quantization table RAM
					mainCfg2 |= 0x0020;
					fSuccess = AgilentProcessorWrite( apr04_CFG_MAIN2, mainCfg2 );
				}
			}
		}
		else
		{
			// Enable Auto Q Adjust
			mainCfg2 |= 0x0040;
			fSuccess = AgilentProcessorWrite( apr04_CFG_MAIN2, mainCfg2 );
		}
	}	

	if ( !fSuccess )
	{
		jpegSetLastError( JPEG_ERR_AGILENT_ERROR );
	}
	return ( fSuccess );
}

//---------------------------------------------------------------------------
//	prvInitAnyCapture - prepares data structures for next capture
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//
//	Return:	TRUE successful, FALSE if error.
//---------------------------------------------------------------------------
Boolean prvInitAnyCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
	Boolean fSuccess = prvValidateParam( pJpegInfo );
	if ( fSuccess )
	{
		UInt32 loop = 0;

		pCamEncState->m_fNoTimeOut	= TRUE;
		pCamEncState->m_fCamClk		= prvGetCamClk();

		// Stop Camera Capture
		pCamEncState->m_CamControl = prvStopCapture();

		// Start Camera Clock if not already started
		if ( !pCamEncState->m_fCamClk )
		{
			prvSetCamClk( TRUE );
			halDelayUS( 200 * 1000 );
		}

		fSuccess = prvAccessAgilent( pJpegInfo, pCamEncState );
		if ( fSuccess )
		{
			pCamEncState->m_AutoFrameXfer = (Boolean)halReadReg8( REG003C_LCDXFER );

			// Turn off auto frame transfer if enabled
			if ( pCamEncState->m_AutoFrameXfer & 1 )
			{
				halWriteReg8( REG003C_LCDXFER, 0 );
			}
		}

	}
	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	prvRestoreReg - Restore list of previously saved registers
//
//	Input:	pSaveReg	pointer to array of register values to restore
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvRestoreReg( UInt8* pSaveReg )
{
	UInt8 i;

	for ( i = 0; i < CAMENC_REG_SIZE; ++i, ++pSaveReg )
	{
		halWriteReg8( gCamEncodeRegIndex[ i ], *pSaveReg );
	}
}

//-----------------------------------------------------------------------------
//	prvRestoreState - restore saved world state
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//			fSuccess		status of encoding
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvRestoreState( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState, Boolean fSuccess )
{
	// Restore pointers to starting state
	pJpegInfo->m_pBuf		= pCamEncState->m_pStart;
	pJpegInfo->m_BufSize	= pCamEncState->m_BufSize;
	pJpegInfo->m_EncSize	= pCamEncState->m_TotalSize;

	if ( fSuccess  )
	{
		// Restore previous address
		halWriteReg8( REG0102_CMSIGNAL, pCamEncState->m_CamSignal );
		//76543210b
		//|||||||+--Valid Input Clock Edge
		//||||||+---CMVREF Active Select 1=Active Hi 0=Active Lo
		//|||||+----CMHREF Active Select 1=Active Hi 0=Active Lo
		//|||++-----YUV Data Format Select
		//||+-------Clock Mode Select
		//|+--------Reserved
		//+---------n/a

		// Force Capture Frame to start capture (this bit is write only so having
		// saved the original state does not help
		halWriteReg8( REG0108_CMCONTROL, (UInt8)((pCamEncState->m_CamControl & ~0x08) | 0x02));
		//76543210b
		//|||||||+--CMCLK Output Hold Reset
		//||||||+---Capture Frame
		//|||||+----Repeat Capture Frame
		//||||+-----Camera Capture Stop
		//|||+------Camera Data Type 1=YUV 0=JPEG
		//+++-------n/a

		prvSetStartAddrImg( pCamEncState->m_StartAddr.m_Prev );
	}

	if ( pCamEncState->m_fUnWind )
	{
		prvUnWind( pJpegInfo, pCamEncState );
    }

	prvRestoreReg( pCamEncState->m_aSaveReg );
}

//-----------------------------------------------------------------------------
//	prvSaveReg - Saves hardware to array of register values
//
//	Input:	pSaveReg	pointer to array of register values to save
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvSaveReg( UInt8* pSaveReg )
{
	UInt8 i;

	for ( i = 0; i < CAMENC_REG_SIZE; ++i, ++pSaveReg )
	{
		*pSaveReg = halReadReg8( gCamEncodeRegIndex[ i ] );
	}
}

//------------------------------------------------------------------------------
//	prvAdvanceStartAddr - Advances start address to next location allowing for
//							image wrap
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pStartAddr		pointer to start address object
//
//	Return:	n/a
//
//------------------------------------------------------------------------------
void prvAdvanceStartAddr( JPEGINFO* pJpegInfo, prvStartAddr* pStartAddr )
{
	if ( pJpegInfo->m_fMultiLoc )
	{
		// Advance to next memory location
		pStartAddr->m_Cur += pJpegInfo->m_EncSize + 15;
		pStartAddr->m_Cur &= 0xFFFFFFF0;

		// Check for memory wrap
		if ( pStartAddr->m_Cur > 0x27000L )
		{
			// If wrap reset to "starting" address
			pStartAddr->m_Cur = pStartAddr->m_First;
		}
		prvSetStartAddrImg( pStartAddr->m_Cur );
	}
}

//------------------------------------------------------------------------------
//	prvSetStartAddrImg - writes start address hardware
//
//	Input:	image write start address
//
//	Return:	n/a
//
//------------------------------------------------------------------------------
void prvSetStartAddrImg( UInt32 address )
{
	prvHalWriteReg24( REG0224_YUVWRSTART0, address );
}

//------------------------------------------------------------------------------
//	prvInitStartAddr - initializes start address object & hardware
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pStartAddr		pointer to start address object
//
//	Return:	n/a
//
//------------------------------------------------------------------------------
void prvInitStartAddr( JPEGINFO* pJpegInfo, prvStartAddr* pStartAddr )
{
	UInt32 height	= GetDisplayHeight( cl_MAIN_WINDOW );
	UInt32 stride	= GetStride( cl_MAIN_WINDOW );
	UInt32 start	= GetStartAddress( cl_MAIN_WINDOW );
	UInt32 end		= start + (stride * height);

	pStartAddr->m_Cur = pStartAddr->m_Prev = prvHalReadReg24( REG0224_YUVWRSTART0 );

	if ( pJpegInfo->m_fMultiLoc )
	{
		pStartAddr->m_First = pStartAddr->m_Cur;
	}
	else
	{
		pStartAddr->m_Cur = end;
	}

	prvSetStartAddrImg( pStartAddr->m_Cur );
	pJpegInfo->m_CapAddr = pStartAddr->m_Cur;
}

//-----------------------------------------------------------------------------
//	prvUnWind - Restore previous hardware states
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvUnWind( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
    if ( pCamEncState->m_fYUV )
	{
		AgilentYUVCIF();
	}

	if ( !pCamEncState->m_fCamClk )
	{
		prvSetCamClk( FALSE );
	}

    // Restore previous auto frame transfer if enabled
	if ( pCamEncState->m_AutoFrameXfer & 0x01 )
	{	
		halWriteReg8( REG003C_LCDXFER, 1 );
    }
}

//-----------------------------------------------------------------------------
//	prvValidateParam - validate JPEGINFO object
//
//	Input:	pJpegInfo	pointer to JPEGINFO object
//
//	Output:	TRUE if valid parameters
//			FALSE if NOT valid paramaters
//
//-----------------------------------------------------------------------------
Boolean prvValidateParam( JPEGINFO* pJpegInfo )
{
	Boolean fSuccess = jpegValidateBuf( pJpegInfo );
	if ( pJpegInfo->m_fFixBug )
	{
		fSuccess = (pJpegInfo->m_XStart & 0x03) ? FALSE : TRUE;
		if ( fSuccess )
		{
			fSuccess = (pJpegInfo->m_YStart & 0x03) ? FALSE : TRUE;
			if ( fSuccess )
			{
				fSuccess = (pJpegInfo->m_Height & 0x03) ? FALSE : TRUE;
				if ( fSuccess )
				{
					fSuccess = (pJpegInfo->m_Width & 0x03) ? FALSE : TRUE;
				}
			}
		}
		if ( !fSuccess )
		{
			jpegSetLastError( JPEG_ERR_DIM_RESTRICT );
		}
	}

	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvWaitForCapture - captures image despite truncation
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//			pCamEncState	pointer to prvCamEncodeState object
//
//	Return:	TRUE means valid image
//			FALSE means NOT valid image
//
//------------------------------------------------------------------------------
Boolean prvWaitForCapture( JPEGINFO* pJpegInfo, prvCamEncodeState* pCamEncState )
{
	Boolean fSuccess;
	Boolean fCompleteImage;
	UInt32	retry = 0;

	do
	{
		fSuccess = prvStartCapture( pJpegInfo, pCamEncState, &fCompleteImage );
		if ( fSuccess )
		{
			if ( !fCompleteImage )
			{
				++retry;
				if ( pJpegInfo->m_fVerbose )
				{
					UInt8 qTable = halReadDisplay8( pCamEncState->m_StartAddr.m_Cur + 3 );
					//printf( "Image %d truncated for QTable %u.  Retry %u\n", pCamEncState->m_Count, qTable, retry );
				}
				fCompleteImage = (pJpegInfo->m_fUseTrunc && pJpegInfo->m_fWrAlways) ? TRUE : FALSE;
				if ( fCompleteImage )
				{
					if ( pJpegInfo->m_fVerbose )
					{
						//printf( "Using truncated image %d\n", pCamEncState->m_Count );
					}
				}

				fSuccess = (retry < pJpegInfo->m_TruncAt) ? TRUE : FALSE;
				if ( !fSuccess )
				{
					jpegSetLastError( JPEG_ERR_TRUNCATED_JPEG_DATA );
				}
			}
		}
	}
	while ( (!fCompleteImage) && fSuccess );

	return ( fSuccess );
}
