//===========================================================================
//	JPEGCOMM.C
//	This file contains JPEG functons used by the JPEG library but not exposed
//	to the outside world.
//---------------------------------------------------------------------------
//  Copyright (c) 2003 Epson Research and Development, Inc.
//  All Rights Reserved.
//===========================================================================

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "multiframe.h"
#include "chiplib.h"
#include "Hal.h"
#include "JPEG.h"
#include "JPEGComm.h"

//------------------------------------------------------------------------------
// Local Definitions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Local Function Prototypes
//------------------------------------------------------------------------------
static void prvLoadTable( UInt16 index, UInt8* pTable, int size );

//------------------------------------------------------------------------------
// Local Data
//------------------------------------------------------------------------------
static jpegError gJpegComLastError;			// Jdecode last error code

#ifdef JPEGUSEFILESYS
//-----------------------------------------------------------------------------
//	jpegCreateFileName - create appropriate filename
//
//	Input:	pJpegInfo	pointer to JPEGINFO object
//			pszName		pointer to user's string name
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void jpegCreateFileName( JPEGINFO* pJpegInfo, const char* pszName )
{
	char* pszExt = pJpegInfo->m_fMulFrm ? "mf" : "jpg";

	// Write JPEG image to JFIF file
	piCreateFileName( pJpegInfo->m_pszFile, pszName, pszExt ); 
}
#endif

//---------------------------------------------------------------------------
//  jpegGetLastError() - Get last error code.
//---------------------------------------------------------------------------
jpegError jpegGetLastError( void )
{
	return gJpegComLastError;
}

//------------------------------------------------------------------------------
//	jpegHasEndOfImage - Determines if given image has JPEG end-of-image marker
//
//	Input:	pJpegInfo	pointer to JPEGINFO structure
//
//	Return:	TRUE means image has end-of-image marker
//			FALSE means that cannot locate end-of-image marker
//
//------------------------------------------------------------------------------
Boolean jpegHasEndOfImage( JPEGINFO* pJpegInfo )
{
	UInt32 i;

	for ( i=2; i<=5; i++ )
		if ( prvIsEndOfImage(pJpegInfo,i*sizeof(UInt8)) )
			return TRUE;

	jpegSetLastError( JPEG_ERR_NO_ENDOFIMAGE );
	return FALSE;
}

//------------------------------------------------------------------------------
//	jpegHasImageMarkers - Determines if given image has JPEG image markers
//
//	Input:	pJpegInfo	pointer to JPEGINFO structure
//
//	Return:	TRUE means image has image markers
//			FALSE means that cannot locate image markers
//
//------------------------------------------------------------------------------
Boolean jpegHasImageMarkers( JPEGINFO* pJpegInfo )
{
	Boolean fSuccess = jpegHasStartOfImage( pJpegInfo );
	if ( fSuccess )
	{
		fSuccess = jpegHasEndOfImage( pJpegInfo );
	}
	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	jpegHasStartOfImage - determines if given offset into JPEG image has 
//							start-of-image marker
//
//	Input:	pJpegInfo	pointer to JPEGINFO structure
//
//	Return:	TRUE means image has end-of-image marker at given offset
//			FALSE means that image does not have end-of-image marker at given offset
//
//------------------------------------------------------------------------------
Boolean jpegHasStartOfImage( JPEGINFO* pJpegInfo )
{
	// Verify that end-of-image marker is present
	UInt8*	pEOI		= pJpegInfo->m_pBuf;
	UInt16	marker		= *(pEOI + 1) | (*(pEOI + 0) << 8);
	Boolean fSuccess	= (marker == 0xFFD8) ? TRUE : FALSE;
	if ( !fSuccess )
	{
		jpegSetLastError( JPEG_ERR_NO_STARTOFIMAGE );
	}
	return ( fSuccess );
}

//---------------------------------------------------------------------------
// jpegInit() - initialize JPEG encode library
//
//	Input:	pJpegInfo	pointer to JPEGINFO structure
//
// Function: Prepare for encode.
// Return:	TRUE successful, FALSE if error.
//---------------------------------------------------------------------------
Boolean	jpegInit( JPEGINFO* pJpegInfo )
{
	InitChipLib();

	// Initialize JEncodeInfo structure with default values
	prvGetResizeRect( &pJpegInfo->m_XStart, &pJpegInfo->m_YStart, &pJpegInfo->m_XEnd, &pJpegInfo->m_YEnd );
	pJpegInfo->m_RWidth			= 0;
	pJpegInfo->m_RHeight		= 0;
	pJpegInfo->m_Width			= pJpegInfo->m_XEnd - pJpegInfo->m_XStart + 1;
	pJpegInfo->m_Height			= pJpegInfo->m_YEnd - pJpegInfo->m_YStart + 1;

	pJpegInfo->m_BufSize		= 0;
	//pJpegInfo->m_CapAddr		= 0;
	pJpegInfo->m_Delay			= 30000;
	//pJpegInfo->m_fDumpOnly		= FALSE;
	pJpegInfo->m_EncSize		= 0;
	//pJpegInfo->m_fFixBug		= FALSE;
	pJpegInfo->m_pszFile		= NULL; 
	pJpegInfo->m_fFastCpu		= FALSE;
	pJpegInfo->m_FrameCount		= 1;
	pJpegInfo->m_HScaleRate		= prvGetResizeHScaleRate();
	pJpegInfo->m_fIndScale		= prvGetResizeIndScaleRate();
	pJpegInfo->m_fInfLoop		= FALSE;
	//pJpegInfo->m_fInit			= TRUE;
	//pJpegInfo->m_fInterpret		= FALSE;
	//pJpegInfo->m_LoopCount		= 1;
	//pJpegInfo->m_MinDelay		= 1000;
	pJpegInfo->m_fMulFrm		= FALSE;
	//pJpegInfo->m_fMultiLoc		= FALSE;
	pJpegInfo->m_fPause			= FALSE;
	pJpegInfo->m_pBuf			= NULL;
	//pJpegInfo->m_QTable			= -1; // Which means use auto selection
	pJpegInfo->m_ScaleMode		= prvGetResizeScaleMode();
	pJpegInfo->m_fSOE			= FALSE;
	pJpegInfo->m_TimeOutMS		= 0; // Default is no time out
	//pJpegInfo->m_TruncAt		= 20;
	//pJpegInfo->m_fVerbose		= FALSE;
	//pJpegInfo->m_fVideoOnly		= FALSE;
	//pJpegInfo->m_fViewImage		= FALSE;
	pJpegInfo->m_VScaleRate		= prvGetResizeVScaleRate();
	pJpegInfo->m_fWrAlways		= FALSE;
	pJpegInfo->m_YUVFmt			= YUVFmt422;
	pJpegInfo->m_fAFAP			= FALSE;
	pJpegInfo->m_IsYUVfile		= FALSE;
	pJpegInfo->m_MemEnc			= FALSE;
	//pJpegInfo->m_MemSize		= 0;
	jpegSetLastError( JPEG_ERR_NONE );
	return ( TRUE );
}

//-----------------------------------------------------------------------------
//	jpegInitVideo - performs minimal steps to enable viewing of camera
//
//	Input:	pJpegInfo		pointer to JPEGINFO object
//
//	Output:	TRUE means camera enabled
//			FALSE means camera not enabled
//
//-----------------------------------------------------------------------------
Boolean jpegInitVideo( JPEGINFO* pJpegInfo )
{
	// Enable camera clock
	halWriteReg8( REG0108_CMCONTROL, 0x01 );
	halWriteReg8( REG0108_CMCONTROL, 0x00 );
	
	halWriteReg8( REG0104_CMMODE0, (UInt8)(halReadReg8(REG0104_CMMODE0)|0x01) );

	// Delay 750 milliseconds to stablize camera output.
	halDelayUS( 750 * 1000 );

	// Start capture
	halWriteReg8( REG0108_CMCONTROL, 0x16 );

	// Enable View Encode (simultaneous display)
	halWriteReg8( REG0202_RESIZE, (UInt8)(halReadReg8(REG0202_RESIZE)|0x08));

	// Enable and reset resizer [norman delete]
	//prvSetResize( TRUE, pJpegInfo );
	// Enable and Reset Resizer [norman add]
	prvSetResize( TRUE, pJpegInfo );
	prvResetResize( TRUE, pJpegInfo );
	prvResetResize( FALSE, pJpegInfo );
	
	// LCD Interface Auto Transfer Enable
	halWriteReg8( REG003C_LCDXFER, 0x01 );

// The following test is invalid, because REG[010Ah-b0] keeps changing from 1 to 0 over time. Jiliang knows why, but I have no clue. [Roc 2003-AUG-4]
//	value	= halReadReg8( REG010A_CMSTATUS );
//	fCamera = (Boolean)(value & 0x01) ? TRUE : FALSE;
//	return ( fCamera );
	return TRUE;
}

//---------------------------------------------------------------------------
//  jpegSetLastError() - Set last error code.
//---------------------------------------------------------------------------
void jpegSetLastError( jpegError error )
{
	gJpegComLastError = error;
}

//-----------------------------------------------------------------------------
//	jpegValidateBuf - validates input parameters
//
//	Input:	pJpegInfo	pointer to JPEGINFO object
//
//	Output:	TRUE means success
//			FALSE means NOT success
//
//-----------------------------------------------------------------------------
Boolean jpegValidateBuf( JPEGINFO* pJpegInfo )
{
	Boolean fSuccess = (pJpegInfo->m_pBuf != NULL) ? TRUE : FALSE;
	if ( !fSuccess )
	{
		jpegSetLastError( JPEG_ERR_NULL_BUFFER );
	}
	else
	{
		fSuccess = ( pJpegInfo->m_BufSize != 0) ? TRUE : FALSE;
		if ( !fSuccess )
		{
			jpegSetLastError( JPEG_ERR_ZERO_BUFFER_SIZE );
		}
	}
	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	jpegWaitTillLCDNotBusy - waits till LCD is not busy or time out
//
//	Input:	pJpegInfo	pointer to JPEG decode info object
//
//	Output:	TRUE means LCD not busy (and no time out)
//			FALSE means time out
//
//-----------------------------------------------------------------------------
Boolean jpegWaitLCDNotBusy( JPEGINFO* pJpegInfo )
{
	Boolean fSuccess = TRUE;
	Boolean fBusy;

	do
	{
		fBusy = (halReadReg8( REG0038_LCDSTATUS ) & 0x01 ) ? TRUE : FALSE;
#ifndef MV850E
		if ( fBusy )
		{
			//fSuccess = jpegNoTimeOut( pJpegInfo );
		}
#endif
	}
	while ( fBusy && fSuccess );
	return ( fSuccess );
}

#ifdef JPEGUSEFILESYS
//-----------------------------------------------------------------------------
//	jpegWriteImage - writes memory image as single or multiple files
//
//	Input:	pszName			pointer to name prefix for file
//			pJpegInfo	pointer to JPEG Encode Info object
//
//	Output:	write output file result
//
//-----------------------------------------------------------------------------
Boolean jpegWriteImage( const char* pszName, JPEGINFO* pJpegInfo )
{
	Boolean fSuccess = TRUE;

	if ( pJpegInfo->m_fMulFrm || (pJpegInfo->m_FrameCount == 1) )
	{
		jpegCreateFileName( pJpegInfo, pszName );
		fSuccess = jpegWriteFile( pJpegInfo );
	}
	else
	{
		// Save original settings
		UInt32 count;
		UInt8* pBuf		= pJpegInfo->m_pBuf;
		UInt32 encSize	= pJpegInfo->m_EncSize;

		for ( count = 1; (count <= pJpegInfo->m_FrameCount) && fSuccess; ++count )
		{
			char szFile[ 256 ];
			JPEGMULTIFRAME* pJpegMultiFrame = (JPEGMULTIFRAME*)pJpegInfo->m_pBuf;
			pJpegInfo->m_pBuf				+= sizeof( JPEGMULTIFRAME );
			pJpegInfo->m_EncSize = pJpegMultiFrame->ResultSize;
			//sprintf( szFile, "%s%04d", pszName, count ); 
			jpegCreateFileName( pJpegInfo, szFile );
			fSuccess			= jpegWriteFile( pJpegInfo );
			pJpegInfo->m_pBuf	+= pJpegMultiFrame->PaddedSize;
		}

		// Restore original settings
		pJpegInfo->m_pBuf		= pBuf;
		pJpegInfo->m_EncSize	= encSize;
	}

	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	jpegWriteFile - writes JPEG encoded data to given file
//			if R180 flag is set, we have to process and re-write the JPEG file.				
//
//	Input:	pJpegInfo	pointer to JPEGINFO structure
//
//	Output:	TRUE if data successfully written to file
//			FALSE if data NOT successfully written to file
//
//-----------------------------------------------------------------------------
Boolean jpegWriteFile( JPEGINFO* pJpegInfo )
{
	FILE*	pStream;
	Boolean fSuccess = mfCreateFile( pJpegInfo->m_pszFile, &pStream );
	if ( fSuccess )
	{
		//printf( "Writing %s\n", pJpegInfo->m_pszFile );
		fSuccess = mfWriteFile( pJpegInfo->m_pszFile, pStream, pJpegInfo->m_pBuf, pJpegInfo->m_EncSize );
		fclose( pStream );
	}
	return ( fSuccess );
}
#endif

//-----------------------------------------------------------------------------
//	jpegGetYUVFmt - get YUV Format Select
//
//	Input:	n/a
//
//	Output:	yuvFmt
//
//-----------------------------------------------------------------------------
jpegYUVFmt jpegGetYUVFmt( void )
{
	UInt8 operationModeSetting = halReadReg8( REG1000_OPMODE );
	//76543210
	//||||||++--YUV Format Select
	//|||||+----JPEG Operation Select
	//||||+-----Marker Insert Enable
	//|||+------DNL Insert Enable
	//+++-------n/a
	jpegYUVFmt yuvFmt = (jpegYUVFmt)(operationModeSetting & 0x03);
	return ( yuvFmt );
}

//-----------------------------------------------------------------------------
//	jpegYUVValue2Fmt - process function to interpret digits as enumerated constants
//
//	Input:	value	to process
//
//	Output:	processed value
//
//-----------------------------------------------------------------------------
int jpegYUVValue2Fmt( int value )
{
	int yuvFmt;

	switch( value )
	{
	case 444: 
		yuvFmt = YUVFmt444;
		break;

	case 422: 
	default:
		yuvFmt = YUVFmt422;
		break;

	case 420:
		yuvFmt = YUVFmt420;
		break;

	case 411: 
		yuvFmt = YUVFmt411;
		break;
	}

	return ( yuvFmt );
}

//-----------------------------------------------------------------------------
//	prvGetCamClk - determines status of camera clock output
//
//	Input:	n/a
//
//	Output:	TRUE means clock output is enabled
//			FALSE means clock output is disabled after 1 frame period finished.
//
//-----------------------------------------------------------------------------
Boolean prvGetCamClk( void )
{
	UInt8 camModeSetting = halReadReg8( REG0104_CMMODE0 );
	//76543210b
	//|||||||+--CMCLK Output Enable
	//||||||+---Camera Active Pulldown Enable
	//|||||+----ITU-R BT656 Enable
	//||||+-----ITU-R BT656 Interrupt Enable
	//++++------n/a
	Boolean fCamClk = (camModeSetting & 1) ? TRUE : FALSE;
	return ( fCamClk );
}

//-----------------------------------------------------------------------------
//	prvSetCamClk - sets status of camera clock output
//
//	Input:	fclkOut		TRUE means Clk Out End
//						FALSE means clock output is disabled after 1 frame period
//							finished
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvSetCamClk( Boolean fClkOut )
{
	UInt8 camModeSetting = halReadReg8( REG0104_CMMODE0 );
	//76543210b
	//|||||||+--CMCLK Output Enable
	//||||||+---Camera Active Pulldown Enable
	//|||||+----ITU-R BT656 Enable
	//||||+-----ITU-R BT656 Interrupt Enable
	//++++------n/a

	Boolean fPrevClk = (camModeSetting & 1) ? TRUE : FALSE;
	camModeSetting	&= (~1);
	camModeSetting	|= fClkOut ? 0x01 : 0x00;
	halWriteReg8( REG0104_CMMODE0, camModeSetting );
	if ( fClkOut != fPrevClk )
	{
		halDelayUS( 70 * 1000 );
	}
}

//-----------------------------------------------------------------------------
//	prvExitSequence - clean up after all JPEG encoding is complete
//
//	Input:	success		current state of process
//			fDecode		flag to determine if decode shutdown required
//
//	Output:	true means success
//			false means NOT success
//
//-----------------------------------------------------------------------------
Boolean prvExitSequence( Boolean fSuccess, Boolean fDecode )
{
	// Check the JPEG operation Status
	Boolean fComplete = prvGetOpWork() ? FALSE : TRUE;

	if ( fSuccess )
	{
		fSuccess = fComplete;
		if ( !fSuccess )
		{
			jpegSetLastError( JPEG_ERR_INCOMPLETE );
		}
		else
		{
			if ( fDecode )
			{
				switch ( (halReadReg8( REG101E_RSTOPSTATUS ) >> 3) & 0x0F )
				{
				default: 
					break;

				case 0x0B: 
					jpegSetLastError( JPEG_ERR_RESTART_INTERVAL );
					fSuccess = FALSE;
					break;

				case 0x0C:
					jpegSetLastError( JPEG_ERR_IMAGE_SIZE );
					fSuccess = FALSE;
					break;
				}
			}
		}
	}

	// Disable all interrupts
	prvSetIrqControl( IrqDisable );

	// Clear all status bits
	prvSetJpegStatus( IrqDisable );

	// Stop capturing
	halReadReg8( REG1004_JPEGOP );	// Must be read before writing to REG[0810h].
	halWriteReg8( REG0810_JPEGSTARTSTOP, 0x00 );

	// Software reset the JPEG codec
	prvResetJpegCodec( );

	// Perform a JPEG Software Reset
	prvSetJpeg( TRUE, TRUE );

	// Disable the JPEG codec
	prvSetJpeg( FALSE, FALSE );

	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	prvHalReadReg16 - retrieves 2 8-bt registers as a 16-bit value
//
//	Input:	lsbIndex		LSB index to read
//			msbIndex		MSB index to read
//
//	Output:	16-bit value read
//
//-----------------------------------------------------------------------------
UInt16 prvHalReadReg16( UInt16 lsbIndex, UInt16 msbIndex )
{
	UInt8 lsb		= halReadReg8( lsbIndex );
	UInt8 msb		= halReadReg8( msbIndex );
	UInt16 value	= (UInt16)((msb << 8) + lsb);
	return ( value );
}

//-----------------------------------------------------------------------------
//	prvHalReadReg24 - retrieves 3 consecutive 8-bit registers as 24-bit value
//
//	Input:	index		starting index to read
//
//	Output:	32-bit value read
//
//-----------------------------------------------------------------------------
UInt32 prvHalReadReg24( UInt16 index )
{
	UInt8 lsb		= halReadReg8( index + 0 );
	UInt8 ob		= halReadReg8( index + 1 );
	UInt8 msb		= halReadReg8( index + 2 );
	UInt32 value	= (UInt32)((msb << 16) + (ob << 8) + lsb);
	return ( value );
}

//-----------------------------------------------------------------------------
//	prvHalWriteReg16 - writes 2 8-bt registers by separating 16-bit value
//
//	Input:	lsbIndex		LSB index to write
//			msbIndex		MSB index to write
//			value		16-bit value written
//
//	Output:	16-bit value read
//
//-----------------------------------------------------------------------------
void prvHalWriteReg16( UInt16 lsbIndex, UInt16 msbIndex, UInt16 value )
{
	UInt8 lsb = (UInt8)(value & 0x00FF);
	UInt8 msb = (UInt8)((value >> 8) & 0x00FF);
	halWriteReg8( lsbIndex, lsb );
	halWriteReg8( msbIndex, msb );
}

//-----------------------------------------------------------------------------
//	prvHalWriteReg24 - writes 3 consecutive 8-bit registers by separating a
//						32-bit value
//
//	Input:	index		starting index to write
//			value		32-bit value
//
//	Output:	32-bit value read
//
//-----------------------------------------------------------------------------
void prvHalWriteReg24( UInt16 index, UInt32 value )
{
	UInt8 lsb	= (UInt8)(value & 0x00FF);
	UInt8 ob	= (UInt8)((value >> 8) & 0x00FF);
	UInt8 msb	= (UInt8)((value >> 16) & 0x00FF);
	halWriteReg8( index + 0, lsb );
	halWriteReg8( index + 1, ob );
	halWriteReg8( index + 2, msb );
}

//------------------------------------------------------------------------------
//	prvGetIrqControl
//
//	Input:	n/a
//
//	Return:	prvJpegFlag
//
//------------------------------------------------------------------------------
prvJpegFlag prvGetIrqControl( void )
{
	return ( prvHalReadReg16( REG080A_JPEGIRQ0, REG080C_JPEGIRQ1 ) );
	//FEDBCA9876543210b
	//|||||||||||||||+--Reserved JPEG Line Buffer Interrupt Enable
	//||||||||||||||+---JPEG Codec Interrupt Enable
	//|||||||||||||+----Line Buffer Overflow Interrupt Enable
	//||||||||||||+-----n/a
	//|||||||||||+------Decode Marker Read Interrupt Enable
	//||||||||||+-------JPEG Decode Complete Interrupt Enable
	//||||||||++--------n/a
	//|||||||+----------JPEG Encode Overflow Interrupt Enable
	//|||||++-----------n/a
	//||||+-------------Encode Size Limit Over Interrupt Enable
	//++++--------------n/a
}

//------------------------------------------------------------------------------
//	prvSetIrqControl
//
//	Input:	flag		flag to set
//
//	Return:	n/a
//
//------------------------------------------------------------------------------
void prvSetIrqControl( prvJpegFlag flag )
{
	prvHalWriteReg16( REG080A_JPEGIRQ0, REG080C_JPEGIRQ1, (UInt16)flag );
	//FEDBCA9876543210b
	//|||||||||||||||+--Reserved JPEG Line Buffer Interrupt Enable
	//||||||||||||||+---JPEG Codec Interrupt Enable
	//|||||||||||||+----Line Buffer Overflow Interrupt Enable
	//||||||||||||+-----n/a
	//|||||||||||+------Decode Marker Read Interrupt Enable
	//||||||||||+-------JPEG Decode Complete Interrupt Enable
	//||||||||++--------n/a
	//|||||||+----------JPEG Encode Overflow Interrupt Enable
	//|||||++-----------n/a
	//||||+-------------Encode Size Limit Over Interrupt Enable
	//++++--------------n/a
}

//-----------------------------------------------------------------------------
//	prvLoadHuffmanTables - loads Huffman table registers with default values
//
//	Input:	n/a
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvLoadHuffmanTables( void )
{
	static UInt8 aDCHuffmanTable0[] = 
	{
		0x00,0x01,0x05,0x01,0x01,0x01,0x01,0x01,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b
	};
	static UInt8 aACHuffmanTable0[] = 
	{
		0x00,0x02,0x01,0x03,0x03,0x02,0x04,0x03,0x05,0x05,0x04,0x04,0x00,0x00,0x01,0x7d,
		0x01,0x02,0x03,0x00,0x04,0x11,0x05,0x12,0x21,0x31,0x41,0x06,0x13,0x51,0x61,0x07,
		0x22,0x71,0x14,0x32,0x81,0x91,0xa1,0x08,0x23,0x42,0xb1,0xc1,0x15,0x52,0xd1,0xf0,
		0x24,0x33,0x62,0x72,0x82,0x09,0x0a,0x16,0x17,0x18,0x19,0x1a,0x25,0x26,0x27,0x28,
		0x29,0x2a,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x43,0x44,0x45,0x46,0x47,0x48,0x49,
		0x4a,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x63,0x64,0x65,0x66,0x67,0x68,0x69,
		0x6a,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x83,0x84,0x85,0x86,0x87,0x88,0x89,
		0x8a,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0xa2,0xa3,0xa4,0xa5,0xa6,0xa7,
		0xa8,0xa9,0xaa,0xb2,0xb3,0xb4,0xb5,0xb6,0xb7,0xb8,0xb9,0xba,0xc2,0xc3,0xc4,0xc5,
		0xc6,0xc7,0xc8,0xc9,0xca,0xd2,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,0xd9,0xda,0xe1,0xe2,
		0xe3,0xe4,0xe5,0xe6,0xe7,0xe8,0xe9,0xea,0xf1,0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,0xf8,
		0xf9,0xfa
	};
	static UInt8 aDCHuffmanTable1[] = 
	{
		0x00,0x03,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x00,0x00,0x00,0x00,0x00,
		0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b
	};
	static UInt8 aACHuffmanTable1[] = 
	{
		0x00,0x02,0x01,0x02,0x04,0x04,0x03,0x04,0x07,0x05,0x04,0x04,0x00,0x01,0x02,0x77,
		0x00,0x01,0x02,0x03,0x11,0x04,0x05,0x21,0x31,0x06,0x12,0x41,0x51,0x07,0x61,0x71,
		0x13,0x22,0x32,0x81,0x08,0x14,0x42,0x91,0xa1,0xb1,0xc1,0x09,0x23,0x33,0x52,0xf0,
		0x15,0x62,0x72,0xd1,0x0a,0x16,0x24,0x34,0xe1,0x25,0xf1,0x17,0x18,0x19,0x1a,0x26,
		0x27,0x28,0x29,0x2a,0x35,0x36,0x37,0x38,0x39,0x3a,0x43,0x44,0x45,0x46,0x47,0x48,
		0x49,0x4a,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x63,0x64,0x65,0x66,0x67,0x68,
		0x69,0x6a,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x82,0x83,0x84,0x85,0x86,0x87,
		0x88,0x89,0x8a,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0xa2,0xa3,0xa4,0xa5,
		0xa6,0xa7,0xa8,0xa9,0xaa,0xb2,0xb3,0xb4,0xb5,0xb6,0xb7,0xb8,0xb9,0xba,0xc2,0xc3,
		0xc4,0xc5,0xc6,0xc7,0xc8,0xc9,0xca,0xd2,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,0xd9,0xda,
		0xe2,0xe3,0xe4,0xe5,0xe6,0xe7,0xe8,0xe9,0xea,0xf2,0xf3,0xf4,0xf5,0xf6,0xf7,0xf8,
		0xf9,0xfa
	};

	prvLoadTable( REG1400_JDCHUFF0REG0, aDCHuffmanTable0, sizeof( aDCHuffmanTable0 ) / sizeof( UInt8 ) );
	prvLoadTable( REG1440_JACHUFF0REG0, aACHuffmanTable0, sizeof( aACHuffmanTable0 ) / sizeof( UInt8 ) );
	prvLoadTable( REG1600_JDCHUFF1REG0, aDCHuffmanTable1, sizeof( aDCHuffmanTable1 ) / sizeof( UInt8 ) );
	prvLoadTable( REG1640_JACHUFF1REG0, aACHuffmanTable1, sizeof( aACHuffmanTable1 ) / sizeof( UInt8 ) );
}

//------------------------------------------------------------------------------
//	prvIsEndOfImage - determines if given offset into JPEG image has end-of-image
//							marker
//
//	Input:	pJpegInfo	pointer to JPEGINFO structure
//			offset		offset to check for end-of-image
//
//	Return:	TRUE means image has end-of-image marker at given offset
//			FALSE means that image does not have end-of-image marker at given offset
//
//------------------------------------------------------------------------------
Boolean prvIsEndOfImage( JPEGINFO* pJpegInfo, UInt32 offset )
{
	// Verify that end-of-image marker is present
	UInt8* pEOI			= pJpegInfo->m_pBuf + pJpegInfo->m_EncSize - offset;
	UInt16 marker		= *(pEOI + 1) | (*(pEOI + 0) << 8);
	Boolean fSuccess	= (marker == 0xFFD9) ? TRUE : FALSE;
	return ( fSuccess );
}

//-----------------------------------------------------------------------------
//	prvLoadQuantizationTables - loads quantization tables with default values
//
//	Input:	n/a
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvLoadQuantizationTables( void )
{
	static UInt8 aQuantizationTable0[] = 
	{
		0x0c,0x08,0x07,0x0c,0x12,0x1e,0x26,0x2d,0x09,0x09,0x0a,0x0e,0x13,0x2b,0x2d,0x29,
		0x0a,0x09,0x0c,0x12,0x1e,0x2a,0x33,0x2a,0x0a,0x0c,0x10,0x15,0x26,0x41,0x3c,0x2e,
		0x0d,0x10,0x1b,0x2a,0x33,0x51,0x4d,0x39,0x12,0x1a,0x29,0x30,0x3c,0x4e,0x54,0x45,
		0x24,0x30,0x3a,0x41,0x4d,0x5a,0x5a,0x4b,0x36,0x45,0x47,0x49,0x54,0x4b,0x4d,0x4a
	};
	static UInt8 aQuantizationTable1[] = 
	{
		0x0c,0x0d,0x12,0x23,0x4a,0x4a,0x4a,0x4a,0x0d,0x0f,0x13,0x31,0x4a,0x4a,0x4a,0x4a,
		0x12,0x13,0x2a,0x4a,0x4a,0x4a,0x4a,0x4a,0x23,0x31,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,
		0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,
		0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a,0x4a
	};

	prvLoadTable( REG1200_JQUANT0, aQuantizationTable0, sizeof( aQuantizationTable0 ) / sizeof( UInt8 ) );
	prvLoadTable( REG1280_JQUANT1, aQuantizationTable1, sizeof( aQuantizationTable1 ) / sizeof( UInt8 ) );
}

//-----------------------------------------------------------------------------
//	prvLoadTable - loads a given table with a list of values
//
//	Input:	index		starting index for table load (assumed every 2nd registers)
//			pTable		pointer to table values to program in registers
//			size		number of entries in given table
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvLoadTable( UInt16 index, UInt8* pTable, int size )
{
	for ( ; size; index += sizeof( UInt16), --size, ++pTable )
	{
		halWriteReg8( index, *pTable );
	}
}

//-----------------------------------------------------------------------------
//	prvGetJpeg - get status of JPEG module
//
//	Input:	n/a
//
//	Output:	on	true means clock supplies to Jpeg module
//				false means clock does not supply to Jpeg module
//
//-----------------------------------------------------------------------------
Boolean prvGetJpeg( void )
{
	UInt8 control = halReadReg8( REG0800_JPEGCONTROL );
	//76543210b
	//|||||||+--Jpeg Module On/Off
	//|++++++---n/a
	//+---------SW Reset (WO)
	Boolean fEnable = (control & 1) ? TRUE : FALSE;
	return ( fEnable );
}

//-----------------------------------------------------------------------------
//	prvSetJpeg - sets state of JPEG module
//
//	Input:	fEnable		true means clock supplies to Jpeg module
//						false means clock does not supply to Jpeg module
//			fReset
//
//	Output:	n/a
//
//-----------------------------------------------------------------------------
void prvSetJpeg( Boolean fEnable, Boolean fReset )
{
	UInt8 control = halReadReg8( REG0800_JPEGCONTROL );
	//76543210b
	//|||||||+--Jpeg Module On/Off
	//|++++++---n/a
	//+---------SW Reset (WO)
	control &= (~1);
	control |= fEnable ? 1 : 0;
	control |= fReset ? 0x80 : 0;
	halWriteReg8( REG0800_JPEGCONTROL, control );
}

//-----------------------------------------------------------------------------
//	prvGetOpWork - get JPEG Operation Status
//
//	Input:	n/a
//
//	Out:	FALSE means idle
//			TRUE means work
//
//-----------------------------------------------------------------------------
Boolean prvGetOpWork( void )
{
	Boolean fWork = halReadReg8( REG1004_JPEGOP ) ? TRUE : FALSE;
	return ( fWork );
}

//------------------------------------------------------------------------------
//	prvRestoreWorld - Restore previously saved values
//
//	Input:	pJpegWorld		pointer to prvWorld object
//
//	Return:	TRUE means success
//			FALSE means NOT success
//
//------------------------------------------------------------------------------
Boolean prvRestoreWorld( JPEGINFO* pJpegInfo, prvWorld* pJpegWorld )
{
	Boolean fSuccess;

	// Restore Previous State
	halWriteReg8( REG1000_OPMODE, pJpegWorld->m_OpMode );
	prvSetJpeg( pJpegWorld->m_fJpeg, FALSE );

	fSuccess = prvSetResizeRect
	( 
		pJpegWorld->m_XStart, 
		pJpegWorld->m_YStart, 
		pJpegWorld->m_XEnd, 
		pJpegWorld->m_YEnd 
	);
	if ( fSuccess )
	{
		fSuccess = prvSetResizeScaleMode( pJpegWorld->m_ScaleMode, pJpegInfo );
		if ( fSuccess )
		{
			fSuccess = prvSetResizeHScaleRate( pJpegWorld->m_HScaleRate, pJpegInfo );
			if ( fSuccess )
			{
				fSuccess = prvSetResizeVScaleRate( pJpegWorld->m_VScaleRate, pJpegInfo );
				if ( fSuccess )
				{
					fSuccess = prvSetResize( pJpegWorld->m_fResize, pJpegInfo );
					if ( fSuccess )
					{
						if ( pJpegWorld->m_fCamClk != prvGetCamClk() )
						{
							prvSetCamClk( pJpegWorld->m_fCamClk );
						}
					}
				}
			}
		}
	}
	return ( fSuccess );
}

//------------------------------------------------------------------------------
//	prvSaveWorld - Save values
//
//	Input:	pJpegWorld		pointer to prvWorld object
//
//	Return:	n/a
//
//------------------------------------------------------------------------------
void prvSaveWorld( prvWorld* pJpegWorld )
{
	// Save registers that don't "belong" to JPEG
	pJpegWorld->m_fCamClk = prvGetCamClk();

	prvGetResizeRect
	( 
		&pJpegWorld->m_XStart, 
		&pJpegWorld->m_YStart,
		&pJpegWorld->m_XEnd,
		&pJpegWorld->m_YEnd
	);
	pJpegWorld->m_ScaleMode		= prvGetResizeScaleMode();
	pJpegWorld->m_HScaleRate	= prvGetResizeHScaleRate();
	pJpegWorld->m_VScaleRate	= prvGetResizeVScaleRate();
	pJpegWorld->m_fResize		= prvGetResize();
	pJpegWorld->m_fJpeg			= prvGetJpeg();
	pJpegWorld->m_OpMode		= halReadReg8( REG1000_OPMODE );
}

prvJpegFlag prvGetJpegStatus( void )
{
	return ( prvHalReadReg16( REG0802_JPEGSTATUS0, REG0804_JPEGSTATUS1 ) );
}

//------------------------------------------------------------------------------
//	prvSetJpegStatus
//
//	Input:	flag		flag to set
//
//	Return:	n/a
//
//------------------------------------------------------------------------------
void prvSetJpegStatus( prvJpegFlag flag )
{
	prvHalWriteReg16( REG0802_JPEGSTATUS0, REG0804_JPEGSTATUS1, (UInt16)flag );
	//FEDBCA9876543210b
	//|||||||||||||||+--Reserved JPEG Line Buffer Interrupt Flag
	//||||||||||||||+---JPEG Code Interrupt Flag
	//|||||||||||||+----Line Buffer Overflow Flag
	//||||||||||||+-----Reserved Size Mismatch Flag
	//|||||||||||+------Decode Marker Read Flag
	//||||||||||+-------JPEG Decode Complete Flag
	//|||||||||+--------n/a
	//||||||||+---------Reserved
	//|||||||+----------JPEG Encode Overflow Flag
	//|||||++-----------n/a
	//||||+-------------Encode Size Limit Violation Flag
	//||++--------------n/a
	//|+----------------JPEG Codec File Out Status
	//+-----------------Reserved
}

UInt32 jpegGetEncSizeResult( void )
{
	return ( prvHalReadReg24( REG081C_ENCODERESULT0 ) );
}

void prvStartJpeg( void )
{
	halWriteReg8( REG1002_COMMAND, 0x01 );
	//76543210b
	//|||||||+--JPEG Operation Enable
	//|++++++---n/a
	//+---------SW Reset
}

void prvResetJpegCodec( void )
{
	halWriteReg8( REG1002_COMMAND, 0x80 );
	//76543210b
	//|||||||+--JPEG Operation Enable
	//|++++++---n/a
	//+---------SW Reset
}
