/******************************************************************************
 * Power Task (pwr)
 * Design and coding by Svend Kristian Lindholm, skl@ti.com
 *
 * PWR Message Processing
 *
 * $Id: lcc_handle_message.c,v 1.1.1.1 2004/06/19 06:00:29 root Exp $
 *
 ******************************************************************************/
#include <string.h>
#include "lcc.h"
#include "lcc_api.h"
#include "lcc_trace.h"
#include "lcc_task.h"
#include "lcc_tm_i.h"
#include "lcc_cfg_i.h"
#include "lcc_modulate.h"

#include "ffs.h"

#include "etm_env.h"

#include "abb.h"

// TISH031218: Added by Jason
#include "rvf_api.h"
#include "rvm_use_id_list.h"


/******************************************************************************
 * Function prototypes
 ******************************************************************************/

void ttr(unsigned trmask, char *format, ...);
void str(unsigned mask, char *string);

void *pwr_malloc(int size);

T_RVM_RETURN pwr_task_init(void);

UINT32 pwr_timer_elapsed(UINT32 time_begin, UINT32 current_timer);
void build_name(const char *ch_pre, char *cfg_id , UINT8 index, const char * ch_post,  char * name);

void pwr_modulate_init(void);
void pwr_modulate_on(void);
int pwr_capacity(uint16 Vbat);

void pwr_check_timers();
void pwr_stop_timers();

T_RVM_RETURN pwr_start_timer(UINT32 *timer_begin);
T_RVM_RETURN pwr_stop_timer(UINT32 *timer_begin);

T_RVM_RETURN pwr_check_files(void);
T_RVM_RETURN pwr_read_files(void);
T_RVM_RETURN pwr_read_chg_files(void);
T_RVM_RETURN pwr_read_cal_files(void);

void pwr_send_msg(uint32 msg_id ,T_RVF_ADDR_ID src_addr_id, T_RVF_ADDR_ID  dest_addr_id);
void mmi_send_msg(struct mmi_info_ind_s *event);

void start_q401_charge(void);
void start_q402_charge(void);
void start_pre_charge(void);
void stop_q401_charge(void);
void stop_q402_charge(void);
void stop_pre_charge(void);
void charger_unplug_house_keeping(void);
void cv_charging_house_keeping(void);
void end_charging_house_keeping(void);
int check_chg_unplug(void);


#if (TEST_PWR_MMI_INTERFACE == 1)
	#include "lcc_mmi_api.c"
#endif

extern T_PWR_CTRL_BLOCK *pwr_ctrl;
extern T_PWR_CFG_BLOCK  *pwr_cfg;

/* Robert.Chen add, 2004-05-28 */
UINT8	cv_stage = 0;
UINT8	cv_overflow = 0;

/* zhq, 2004/06/16, add this for in CAL & INI state, we can not read temp<n>.cfg,
so we just use this to get a rough value of bat capacity. */
 uint16   cap_CAL_INI[9+1] = { 
    /* Voltage VS Capacity vector */
    4200,		//           100%,
    4050,		//            90%,
    3950,		//            80%,
    3850,		//            70%,
    3750,		//            60%,
    3650,		//            50%,
    3550,		//            40%,
    3500,		//            30%,
    3400,		//            20%,
    3350		//            10%
    };      /* 20 */ // Capacity as a function of voltage measured at t=20 degrees - C(t=20)
                          // Capacity vector : 100%, 90%, 80%, 70% ,60% ,50%, 40%, 30%, 20%, 10%

/* zhq, 2004/06/16, add this for send event to MMI at least 6 times we enter PRE, to prevent that
send start charging event to MMI too frequently. */
static UINT8	eventsend = 0;

/******************************************************************************
 * Functions 
 ******************************************************************************/

void build_name(const char *ch_pre, char *cfg_id, UINT8 index, const char * ch_post,  char * name) 
{
	// TISH031218: Modified by Jason
	//char tmp[1];
	char tmp[2];

	strcpy(name, ch_pre);
	tmp[0] = *cfg_id;
	tmp[1] = 0;	// null-termination of strings!
	strcpy(&(name[index]), tmp);
	strcat(name, ch_post);
	ttw(ttr(TTrInit, "build_name: '%s'" NL, name));
}


T_RVM_RETURN pwr_task_init()
{
// Perform battery initialization of the pwr task by
// reading battery id and FFS configuration files
	T_FFS_SIZE       error;

	ttw(ttr(TTrInit, "pwr_task_init(%d)" NL, 0));

#if 0 //TISH040218
	// Check configuration and calibration files
	if ((error = pwr_check_files()) < 0)
		return error;

	// Read configuration files
	if ((error = pwr_read_files()) < 0)
		return error;
#endif
	pwr_read_files(); //TISH040218

	ttw(ttr(TTrInit, "pwr_task_init(%d)" NL, 0xFF));
	return RV_OK;

}

int pwr_chg_start(int Tbat) {
	int capacity;
	// Tbat is the temperature measured in Celsius
	// Check start conditions for charging - it is assumed that a charger is plugged
	// This check is used in state SUP

	// Charging start will _not_ be performed if ...
	// 1) Battery temperature is outside limits (measurement)
	// 2) Charger identity is unknown

	ttw(ttr(TTrCharge, "pwr_chg_start(%d)" NL, 0));

	// 1)
	// Battery temperature is outside limits (measurement)
	// FIXME: Check also low temperatures
	//    if ((Tbat < pwr_cfg->temp.tbat_min) || (Tbat > pwr_cfg->temp.tbat_max)) {

/* zhq, 2004/06/16, add this macro to prevent using thermal resistor to detect
temperature for we do not have its support on hardware. */
#if (LCC_TEMP_DETECT_EN == 1)	
	if ((Tbat > pwr_cfg->temp.tbat_max))
	{
		ttw(ttr(TTrCharge, "pwr_chg_start(%d) Tbat=(%d)" NL, 0xFF, Tbat));
		return FALSE;
	}
#endif

	// 2) 
	// Charger identity unknown
/* zhq, 2004/06/16, add this macro to prevent using ID resistor to detect battery ID
for we do not have its support on hardware. */
#if (LCC_BATID_DETECT_EN == 1)	
	if ((pwr_ctrl->flag_chg_unknown == 1) && (pwr_ctrl->flag_mmi_registered == 1))
	{
		ttw(ttr(TTrCharge, "pwr_chg_start(%d) Chg unknown=(%d)" NL, 0xFF, Tbat));
		pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_UNKNOWN_IND;
		mmi_send_msg(pwr_ctrl->mmi_ptr);
		// Send only once
		pwr_ctrl->flag_bat_unknown = 0;
		return FALSE;
	}
#endif

	ttw(ttr(TTrCharge, "pwr_chg_start(%d)" NL, 0xFF));
	return TRUE;
}

int pwr_chg_stop(int Tbat, int Vbat) {
	// Check stop conditions for charging
	// I.e.
	// - temperature (RF+Battery) thresholds
	// - battery voltage thresholds
	// - regulation loop gain (duty cycle)
	//
	// This check is used regularly in state CCI, LCI, CCV and LCV

	// Charging will be stopped if ...
	// 1) Battery temperature is outside limits
	// 2) Battery voltage is equal to - or above - charging stop battery voltage

	// Nickel checks:
	// 3) The RF temperature versus battery temperature rise is above threshold (Possible?)
	//
	// Lithium checks:
	// 4) The regulation parameter k is below configured threshold limit
	// Returns TRUE if the charging algorithm should be stopped
	// Returns FALSE if the charging algorithm should be continued


	ttw(ttr(/*TTrCharge*/0xff000000, "pwr_chg_stop(%d)" NL, 0));

	// 1)
	// Battery temperature is outside limits
	// FIXME: Check also low temperatures
	
/* zhq, 2004/06/16, add this macro to prevent using thermal resistor to detect
temperature for we do not have its support on hardware. */
#if (LCC_TEMP_DETECT_EN == 1)	
	if (Tbat > pwr_cfg->temp.tbat_max)
	{
		ttw(ttr(0xff000000, "pwr_chg_stop(%d) Tbat=(%d)" NL, 0xFF, Tbat));
		if (pwr_ctrl->flag_mmi_registered == 1)
		{
			pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_FAILED_IND;
			pwr_ctrl->mmi_ptr->cause         = BATTERY_TEMPERATURE_HIGH;
			mmi_send_msg(pwr_ctrl->mmi_ptr);
		}
		return TRUE;
	}
#endif

	// 2)
	// Battery voltage is equal to - or above - charging stop battery voltage
	
	if(pwr_cfg->bat.type != LITHIUM)	/* Robert.Chen added this condition, 2004-05-29 */
	{
	if ((Vbat > pwr_cfg->bat.chg_stop_thr))
	{
		ttw(ttr(0xff000000, "pwr_chg_stop(%d) Vbat=(%d)" NL, 0xFF, Vbat));
		ttw(ttr(TTrCharge, "pwr_cfg->bat.chg_stop_thr=%d" NL, 0xFF, pwr_cfg->bat.chg_stop_thr));
		return TRUE;
	}
	}
#if 0
	#ifndef PWR_FFS_CFG
	pwr_cfg->common.rise_thr = PWR_RISE_THR ;
	#endif

	// 3)
	// The RF temperature versus battery temperature rise is above threshold (Possible?)
	// Did the battery temperature rise more than the RF temperature rise - stop for Nickel charging!
	// Only for Nickel batteries
	// FIXME: How big time difference should there be between the deltas? Configurable?
	if ((pwr_cfg->data.dTbat - pwr_cfg->data.dT_rf > pwr_cfg->common.rise_thr) == TRUE)
	{
		ttw(ttr(TTrCharge, "pwr_chg_stop(%d) d(Tbat-Trf)=(%d)" NL, 0xFF, pwr_cfg->data.dTbat - pwr_cfg->data.dT_rf));
		return TRUE;
	}
#endif

	// 4) FIXME: Remove states in this part of the function
	// Regulation parameter k is below the configured limit
	#if 0	/* Robert.Chen modified, 2004-05-29 */
	if ((pwr_cfg->bat.type == LITHIUM) &&
		((pwr_ctrl->state == LCV) || (pwr_ctrl->state == CCV)) &&
		(pwr_cfg->data.k < pwr_cfg->bat.chg_ctrl_thr))
	#else
	if ((pwr_cfg->bat.type == LITHIUM) &&
		(Vbat > pwr_cfg->bat.chg_stop_thr) && (pwr_ctrl->state == CCV) &&
		(pwr_cfg->data.k <= pwr_cfg->bat.chg_ctrl_thr))
	#endif	
	{
		ttw(ttr(0xff000000, "k=%d < %d" NL, pwr_cfg->data.k, pwr_cfg->bat.chg_ctrl_thr));
		return TRUE;
	}

	ttw(ttr(TTrCharge, "pwr_chg_stop(%d)" NL, 0xFF));
	return FALSE;
}

int pwr_chg_ci_cv_transition(int Vbat) {
	// Check CI -> CV condition for charging
	// Only for Lithium batteries
	// Returns TRUE if CI-> CV transition should be made
	// Returns FALSE if CI-> CV transition should NOT be made

	ttw(ttr(TTrCharge, "pwr_chg_ci_cv_transition(%d)" NL, 0));


	// 1) Check CV charging start threshold
	/* Robert.Chen modified, 2004-06-03, to speed up charging */
	if ((Vbat > pwr_cfg->bat.chg_start_thr))
	//if ((Vbat > pwr_cfg->bat.chg_stop_thr))
	/*end*/
	
	{
		ttw(ttr(TTrCharge, "pwr_chg_ci_cv_transition(%d) Vbat=(%d)" NL, 0xFF, Vbat));
		return TRUE;
	}

	ttw(ttr(TTrCharge, "pwr_chg_ci_cv_transition(%d)" NL, 0xFF));
	return FALSE;
}

int pwr_capacity(uint16 Vbat) {
	// Calculate the capacity, C(T,V), of the battery based on
	// - Battery Voltage Measurement
	// - Battery Temperature Measurement
	// - Configuration Parameters
	// Returns a value in the interval 0..100 [%]
	uint8 i;
	uint8 cap;

	ttw(ttr(TTrEventLow,"pwr_capacity (Vbat=%d)" NL, Vbat));
	cap = 100; // 100%
	for (i=0;i<=9;i++)
	{
		if (Vbat < pwr_cfg->temp.cap[i])
		{
			cap -= 10; // Count down 10% for each index
			ttw(ttr(TTrEventLow,"Trying cap=%d" NL, cap));
		}
		else
		{
			// FIXME: Interpolate ... instead of truncate
			ttw(ttr(TTrEventLow,"capacity=%d" NL, cap));
			return cap;
		}
	}
	ttw(ttr(TTrEventLow,"pwr_capacity (%d)" NL, 0xFF));
	return 0; // Couldn't lookup capacity
}

int pwr_capacity_CAL_INI(uint16 Vbat) {
	// Calculate the capacity, C(T,V), of the battery based on
	// - Battery Voltage Measurement
	// - Battery Temperature Measurement
	// - Configuration Parameters
	// Returns a value in the interval 0..100 [%]
	uint8 i;
	uint8 cap;

	ttw(ttr(TTrEventLow,"pwr_capacity (Vbat=%d)" NL, Vbat));
	cap = 100; // 100%
	for (i=0;i<=9;i++)
	{
		if (Vbat < cap_CAL_INI[i])
		{
			cap -= 10; // Count down 10% for each index
			ttw(ttr(TTrEventLow,"Trying cap=%d" NL, cap));
		}
		else
		{
			// FIXME: Interpolate ... instead of truncate
			ttw(ttr(TTrEventLow,"capacity=%d" NL, cap));
			return cap;
		}
	}
	ttw(ttr(TTrEventLow,"pwr_capacity (%d)" NL, 0xFF));
	return 0; // Couldn't lookup capacity
}


int8 pwr_temp_lookup(uint16 ADC, uint8 i_meas) 
{
// temperature = f(ADC, I_meas)
// f(x) is table-lized using ETM command 'pww 4 ...'

	int i;
	int temp;

#define MAX_TEMP_SIZE 9

	i = 0;
	temp = -20;
	// FIXME: Always using the second current - should also use the first...
	while (ADC < pwr_cfg->temp.v2t_2[i] && i < MAX_TEMP_SIZE)
	{
		ttw(ttr(TTrEventLow,"ADC=%d < pwr_cfg->temp.v2t_2[%d]=0x%x" NL, ADC,i, pwr_cfg->temp.v2t_2[i]));
		ttw(ttr(TTrEventLow,"temp=%d" NL, temp));
		i++;
		temp += 10;
	}
	if (i == MAX_TEMP_SIZE)
	{
		// No temperature found - return NULL value (0xFF)
		ttr(TTrWarning, "Temperature lookup failed %d" NL, ADC);
		return 0xFF;
	}
	if (i != 0)
	{
		// Interpolate
		if (pwr_cfg->temp.v2t_2[i-1] == 0xFFFF)
			temp += 10*(ADC - pwr_cfg->temp.v2t_2[i])/(pwr_cfg->temp.v2t_2[i] - 1024);
		else
			temp += 10*(ADC - pwr_cfg->temp.v2t_2[i])/(pwr_cfg->temp.v2t_2[i] - pwr_cfg->temp.v2t_2[i-1]);
		ttw(ttr(TTrEventLow,"Interpolated temp=%d" NL, temp));
	}
	return((int8)temp);
}


enum
{
	Vbat = 0,
	Vchg = 1,
	Ichg = 2,
	Vbt2 = 3,
	Type = 4,
	Tbat = 5,
	T_rf = 6,
	Tc_x = 7,
	Tc_y = 8,
	State= 9
} adc_index_e;

// The driving element of the LCC task - the ADC measurements
T_RV_RET process_spi_adc_indication (T_PWR_REQ *request)
{
	struct pwr_adc_ind_s *ind;
	int error, i, index;
	int8 temp;
	UINT32 timer;
	UINT16 status; // Result of charger plug - unplug - polled
	UINT32 tmp_vbat_mV;

	ind = (struct pwr_adc_ind_s*)request;

	ttw(ttr(TTrEventLow,"State=%d" NL, ind->data[State]));
	
	// Basic trace - not for CAL
	switch (pwr_ctrl->state)
	{
	case PRE :
		ttw(ttr(TTrEventLow,"Vbat=%d" NL, ADC_to_mV(ind->data[Vbat])));
		ttw(ttr(TTrEventLow,"Vchg=%d" NL, ADC_to_mV(ind->data[Vchg])));
		ttw(ttr(TTrEventLow,"Ichg=%d" NL, ind->data[Ichg]));
		break;
	case INI :
	case SUP :
	case CCI :
	case LCI :
	case CCV :
	case LCV :
		ttw(ttr(TTrEventLow,"Vbat=%d" NL, ADC_to_mV(ind->data[Vbat])));
		ttw(ttr(TTrEventLow,"Vbat_adc=%d" NL, ind->data[Vbat]));
#if (USE_Q402_CHG_CIRCUIT == 1)
		// Vchg Scaling factor for Q402 circuit
		pwr_cfg->data.Vchg = 5*ADC_to_mV(ind->data[Vchg])/4;
		ttw(ttr(TTrEventLow,"Vchg=%d" NL, pwr_cfg->data.Vchg));
#else
		pwr_cfg->data.Vchg = ADC_to_mV(ind->data[Vchg]);
		ttw(ttr(TTrEventLow,"Vchg=%d" NL, pwr_cfg->data.Vchg));
#endif
		ttw(ttr(TTrEventLow,"Ichg=%d" NL, ind->data[Ichg]));
		ttw(ttr(TTrEventLow,"Type=%d" NL, ind->data[Type]));
		ttw(ttr(TTrEventLow,"Tbat=%d" NL, ind->data[Tbat]));
		ttw(ttr(TTrEventLow,"T_rf=%d" NL, ind->data[T_rf]));
		ttw(ttr(TTrEventLow,"Vbt2=%d" NL, ADC_to_mV(ind->data[Vbt2])));
		break;
	}

	// Updates variables in 'service' mode - not for CAL, PRE & INI
	switch (pwr_ctrl->state)
	{
	case SUP :
	case CCI :
	case LCI :
	case CCV :
	case LCV :
		index = pwr_ctrl->index % CONSECUTIVE_CHG_UNPLUGS;
		ttw(ttr(TTrEventLow,"Using index: (%d)" NL, index));
		pwr_ctrl->index++;
		// Get current nucleus time
		timer  = rvf_get_tick_count();

		// T0 timer expired?
		pwr_ctrl->time_elapsed_T0  = pwr_timer_elapsed(pwr_ctrl->time_begin_T0, timer);
		ttw(ttr(TTrTimerLow,"T0: %d ms elapsed " NL, pwr_ctrl->time_elapsed_T0));

		// Compare T0 with configured sampling rate
		if (pwr_ctrl->time_elapsed_T0 > pwr_cfg->common.sampling)
		{
			ttw(ttr(TTrTimer, "T0 Reset: elapsed (%d)" NL, pwr_ctrl->time_elapsed_T0));
			pwr_ctrl->time_begin_T0 = timer;

		}

		pwr_cfg->data.Vbat = ind->data[Vbat]; 
		pwr_cfg->data.Vbat_avg = (ind->data[Vbat] + pwr_cfg->common.alfa1 * pwr_cfg->data.Vbat_avg)/(pwr_cfg->common.alfa1 + 1);
		//pwr_cfg->data.Vbat_avg_mV = ADC_to_mV(pwr_cfg->data.Vbat_avg);

		tmp_vbat_mV = (ADC_to_mV(ind->data[Vbat]) + pwr_cfg->data.Vbat_avg_mV * pwr_cfg->common.alfa1) / (pwr_cfg->common.alfa1 + 1);
		pwr_cfg->data.Vbat_avg_mV = (UINT16)tmp_vbat_mV;
		
		/* Robert.Chen added, 2004-05-29 */
		if(pwr_cfg->data.Vbat_avg_mV >= pwr_cfg->bat.chg_stop_thr && pwr_ctrl->state == CCV)
		{
			cv_overflow = 1;
		}
		/* add end */
		
		ttw(ttr(TTrEventLow,"Vbat_avg=%d" NL, pwr_cfg->data.Vbat_avg_mV));
		MI_dbgPrintf("zhq: PowerState=%d, Vbat=%d, Vbat_avg_mV=%d,Vbat=%d,Vbat_avg=%d", pwr_ctrl->state, ADC_to_mV(ind->data[Vbat]), pwr_cfg->data.Vbat_avg_mV,ind->data[Vbat],pwr_cfg->data.Vbat_avg);
		
		pwr_cfg->data.Tbat = ind->data[Tbat];
		pwr_cfg->data.Tbat_avg = (ind->data[Tbat] + pwr_cfg->common.alfa2 * pwr_cfg->data.Tbat_avg)/(pwr_cfg->common.alfa2 + 1);
		ttw(ttr(TTrEventLow,"Tbat_avg=%d" NL, pwr_cfg->data.Tbat_avg));
		temp = pwr_temp_lookup(pwr_cfg->data.Tbat_avg, 0x01);
		ttw(ttr(TTrEventLow,"temp=%d" NL, temp));

		pwr_cfg->data.T_rf_avg = ind->data[T_rf];
		pwr_cfg->data.T_rf_avg = (ind->data[T_rf] + pwr_cfg->common.alfa3 * pwr_cfg->data.T_rf_avg)/(pwr_cfg->common.alfa3 + 1);
		ttw(ttr(TTrEventLow,"T_rf_avg=%d" NL, pwr_cfg->data.T_rf_avg));

		pwr_cfg->data.Ichg = ind->data[Ichg];
		break;
	}

	// 'Main' switch
	ttw(ttr(TTrEventLow,"state=%d" NL, pwr_ctrl->state));
	switch (pwr_ctrl->state)
	{
	case CAL :
		MI_dbgPrintf("zhq: we entered CAL state.");

		// Read calibration files
		// Decide if SW precharge should be started (state PRE) or if a normal configuration scheme should be inititated (state INI)
		error = pwr_read_cal_files();		
		if (error >= 0)
		{
			MI_dbgPrintf("Vbat = %d", ADC_to_mV(ind->data[Vbat]));
			// Save measurements - also initial values for averaging
			pwr_cfg->data.Vbat_avg = pwr_cfg->data.Vbat = ind->data[Vbat];
			pwr_cfg->data.Vbat_avg_mV = ADC_to_mV(pwr_cfg->data.Vbat_avg);
			ttw(ttr(TTrEventLow,"Vbat=%d" NL, ADC_to_mV(ind->data[Vbat])));
			ttw(ttr(TTrEventLow,"Vchg=%d" NL, ADC_to_mV(ind->data[Vchg])));

			// Charger was already inserted?
			status = ind->data[State];

			if (status & CHGPRES)
			{
				// Charger plugged caused a wakeup
				pwr_ctrl->flag_chg_plugged = 1;
				pwr_ctrl->flag_chg_prev_plugged = 1;
				ttw(ttr(TTrInitLow,"Polled - charger plugged  (%d)" NL, status));
			}

			// Check Pre-charge - immediately change state to PRE
			if ((pwr_cfg->data.Vbat_avg_mV < VBAT_PRECHG_START) && (pwr_ctrl->flag_chg_plugged == 1))
			{
				ttw(ttr(TTrInitLow,"precharge (%d)" NL, ADC_to_mV(ind->data[Vbat])));
				// Charger interrupts (plug/unplug) are already masked off - in case it was a linear charger
				pwr_ctrl->state = PRE;

			#if (LCC_ALWAYS_MMI_INFO == 1)
			#if 0
				if(pwr_ctrl->flag_mmi_registered == 1)
				{
					pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG;
					mmi_send_msg(pwr_ctrl->mmi_ptr);
				}
			#endif	
			#endif
			}
			else
			{
				pwr_ctrl->state = INI;
			}
		}
	/* zhq, 2004/06/16, adding this for in INI and CAL state, no bat info is send to MMI, though mmi info timer
	has been started in pwr_register(). pay attention that in INI and CAL, we should use a temporary capacity array.*/
	#if (LCC_ALWAYS_MMI_INFO == 1)	
		pwr_check_timers();
	#endif
		break;
	case PRE :
		MI_dbgPrintf("zhq: we entered PRE state.");
		// Start fast charge immediately after 3.2V boot
		// Enter INI state - in order to read FFS configuration files - when Vbat reaches 3.6V
/* zhq, 2004/06/16, add below code for in PRE state, MMI should be
nofitied that charging has beging. */
		if (pwr_ctrl->flag_mmi_registered == 1)
		{
			if(eventsend == 0)
			{
				pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_START_IND;
				mmi_send_msg(pwr_ctrl->mmi_ptr);
				eventsend++;

			}
			else if (eventsend == 9)
			{
				eventsend = 0;
			}
			else
			{
				eventsend++;
			}
			
		}
		
		// Hardcoding moving average to PRECHG_AVG_WINDOW_SIZE since we haven't read any configuration
		pwr_cfg->data.Vbat = ind->data[Vbat];
		pwr_cfg->data.Vbat_avg = (ind->data[Vbat] + PRECHG_AVG_WINDOW_SIZE  * pwr_cfg->data.Vbat_avg)/(PRECHG_AVG_WINDOW_SIZE + 1); 
		pwr_cfg->data.Vbat_avg_mV = ADC_to_mV(pwr_cfg->data.Vbat_avg);
		if (pwr_cfg->data.Vbat_avg_mV > VBAT_PRECHG_STOP)
		{
			ttw(ttr(TTrInitLow,"state PRE (%d > VBAT_PRECHG_STOP)" NL, pwr_cfg->data.Vbat_avg_mV));
			pwr_ctrl->state = INI;
			pwr_ctrl->flag_prechg_started = 0;
			stop_pre_charge();
		}
		else
		{
			ttw(ttr(TTrInitLow,"state PRE (%d < VBAT_PRECHG_STOP)" NL, pwr_cfg->data.Vbat_avg_mV));
			// Start fast charging NOW
			if (pwr_ctrl->flag_prechg_started == 0)
			{
				pwr_ctrl->flag_prechg_started = 1;
				start_pre_charge();
			}
		}

	#if (LCC_ALWAYS_MMI_INFO == 1)	/* Robert.Chen added, 2004-05-28, getting battery info at 'PRE' state */
		pwr_check_timers();
	#endif
		break;
	case INI :
		MI_dbgPrintf("zhq: we entered INI state.");
		// Charger was already inserted?
		status = ind->data[State];
		pwr_ctrl->chg_unplug_vec[index] = (status & CHGPRES);
		if (status & CHGPRES)
		{
			// Charger plugged caused a wakeup
			pwr_ctrl->flag_chg_plugged = 1;
			pwr_ctrl->flag_chg_prev_plugged = 1;
			ttw(ttr(TTrInitLow,"Polled - charger plugged  (%d)" NL, status));
		}

		if (pwr_ctrl->flag_ini_virgo == 0)
		{
			// Perform some very basic initialization

			/* Precharge (C/20) is switched OFF since it was finished in state PRE */
			ABB_Write_Register_on_page(PAGE0, BCICTL2, 0x0002);

			// Enable battery type identification
			// Switch on battery type identification (TYPEN)
			// NOTE: Battery type identification is ALWAYS done with 10uA (TYPEN)
			ttw(ttr(TTrInitLow,"Enable type ident (%d)" NL, 0x00));
		#if (LCC_BATID_DETECT_EN == 1)		/* Robert.Chen added,  */
			ABB_Write_Register_on_page(PAGE0, BCICTL1, THERMAL_SENSOR_30uA | TYPEN);
		#endif	
			pwr_ctrl->flag_ini_virgo = 1;

		}

		// Don't read battery files before we know which battery id to use
		// We will ignore the first BATTERY_TYPE_SLIP measurements of battery id
		if ((pwr_ctrl->count_bat_type == BATTERY_TYPE_SLIP))
		{

			error = pwr_task_init();

			// Change state
			pwr_ctrl->state = SUP;

			// Readout /pwr/common.cfg
			ttw(ttr(TTrInitLow,"common.cfg: pins: %d" NL, pwr_cfg->common.pins));
			ttw(ttr(TTrInitLow,"chg_dedic %d" NL, pwr_cfg->common.chg_dedic));
			ttw(ttr(TTrInitLow,"sampling %d" NL, pwr_cfg->common.sampling));
			ttw(ttr(TTrInitLow,"mod_cycle %d" NL, pwr_cfg->common.mod_cycle));
			ttw(ttr(TTrInitLow,"alfa1 %d" NL, pwr_cfg->common.alfa1));
			ttw(ttr(TTrInitLow,"alfa2 %d" NL, pwr_cfg->common.alfa2));
			ttw(ttr(TTrInitLow,"alfa3 %d" NL, pwr_cfg->common.alfa3));
			ttw(ttr(TTrInitLow,"rise_thr %d" NL, pwr_cfg->common.rise_thr));

			// Readout /pwr/bat/bat.cfg
			ttw(ttr(TTrInitLow,"bat.cfg: type: %d" NL, pwr_cfg->bat.type));
			ttw(ttr(TTrInitLow,"rf_temp: %d" NL, pwr_cfg->bat.rf_temp));
			ttw(ttr(TTrInitLow,"id_low: %d" NL, pwr_cfg->bat.id_low));
			ttw(ttr(TTrInitLow,"id_high: %d" NL, pwr_cfg->bat.id_high));
			ttw(ttr(TTrInitLow,"cbat: %d" NL, pwr_cfg->bat.cbat));
			ttw(ttr(TTrInitLow,"ratio: %d" NL, pwr_cfg->bat.ratio));
			ttw(ttr(TTrInitLow,"T1: %d" NL, pwr_cfg->bat.T1));
			ttw(ttr(TTrInitLow,"T2: %d" NL, pwr_cfg->bat.T2));
			ttw(ttr(TTrInitLow,"T3: %d" NL, pwr_cfg->bat.T3));
			ttw(ttr(TTrInitLow,"chg_start_thr: %d" NL, pwr_cfg->bat.chg_start_thr));
			ttw(ttr(TTrInitLow,"chg_stop_thr: %d" NL, pwr_cfg->bat.chg_stop_thr));
			ttw(ttr(TTrInitLow,"chg_ctrl_thr: %d" NL, pwr_cfg->bat.chg_ctrl_thr));
			ttw(ttr(TTrInitLow,"chg_again_thr: %d" NL, pwr_cfg->bat.chg_again_thr));

			// Readout /pwr/bat/temp<N>.cfg
			ttw(ttr(TTrInitLow,"sizeof(temp): %d" NL, sizeof(pwr_cfg->temp)));
			ttw(ttr(TTrInitLow,"sizeof(T_PWR_BAT_TEMP_CFG_BLOCK): %d" NL, sizeof(T_PWR_BAT_TEMP_CFG_BLOCK)));
			ttw(ttr(TTrInitLow,"temp.cfg: tbat_min: %d" NL, pwr_cfg->temp.tbat_min));
			ttw(ttr(TTrInitLow,"tbat_max: %d" NL, pwr_cfg->temp.tbat_max));
			ttw(ttr(TTrInitLow,"i_meas1: %d" NL, pwr_cfg->temp.i_meas1));
			for (i=0;i<=8;i++)
			{
				ttw(ttr(TTrInitLow,"v2t_1[]: %d" NL, pwr_cfg->temp.v2t_1[i]));
			}
			ttw(ttr(TTrInitLow,"i_meas2: %d" NL, pwr_cfg->temp.i_meas2));
			for (i=0;i<=8;i++)
			{
				ttw(ttr(TTrInitLow,"v2t_2[]: %d" NL, pwr_cfg->temp.v2t_2[i]));
			}
			for (i=0;i<=9;i++)
			{
				ttw(ttr(TTrInitLow,"cap[]: %d" NL, pwr_cfg->temp.cap[i]));
			}
			ttw(ttr(TTrInitLow,"a0: %d" NL, pwr_cfg->temp.a0));
			ttw(ttr(TTrInitLow,"a1: %d" NL, pwr_cfg->temp.a1));
			ttw(ttr(TTrInitLow,"a2: %d" NL, pwr_cfg->temp.a2));

			// Readout /mmi/pwr/bsie.cfg
			ttw(ttr(TTrInitLow,"mmi repetition: %d" NL, pwr_cfg->mmi.repetition));

			// Initialize battery temperature - must not be made in the virgo part since Tbat=0
			pwr_cfg->data.Tbat_avg = ind->data[Tbat];

			// Setup thermal sensor - don't re-enable TYPEN
/* zhq, 2004/06/16, add this macro to prevent using thermal resistor to detect
temperature for we do not have its support on hardware. */
#if (LCC_TEMP_DETECT_EN == 1)	
			if (pwr_cfg->temp.i_meas2 == 50)
			{
				ttw(ttr(TTrInitLow,"ABB set i_current: %d" NL, pwr_cfg->temp.i_meas2));
				ABB_Write_Register_on_page(PAGE0, BCICTL1, THERMAL_SENSOR_50uA);
			}
			else if (pwr_cfg->temp.i_meas2 == 30)
			{
				ttw(ttr(TTrInitLow,"ABB set i_current: %d" NL, pwr_cfg->temp.i_meas2));
				ABB_Write_Register_on_page(PAGE0, BCICTL1, THERMAL_SENSOR_30uA);
			}
			else
			{
				// Default 30uA
				ttw(ttr(TTrInitLow,"ABB set i_current default: %d" NL, pwr_cfg->temp.i_meas2));
				ABB_Write_Register_on_page(PAGE0, BCICTL1, THERMAL_SENSOR_30uA);
			}
#endif


		}
		else
		{
			// Compare battery id with values found in FFS
			// First value(s) are NOT to be trusted
			pwr_ctrl->count_bat_type++;
			pwr_cfg->data.bat_id = ind->data[Type];

		}
	/* zhq, 2004/06/16, adding this for in INI and CAL state, no bat info is send to MMI, though mmi info timer
	has been started in pwr_register(). pay attention that in INI and CAL, we should use a temporary capacity array.*/
	#if (LCC_ALWAYS_MMI_INFO == 1)	
		pwr_check_timers();
	#endif
		break;
	case SUP :
		MI_dbgPrintf("zhq: we entered SUP state.");
		// Charger inserted?? Poll!!
		// Reason: 100Hz interrupts will occur if an unregulated charger is inserted & charger interrupts are enabled
		//         This will cause an immediate crash!
		status = ind->data[State];
		pwr_ctrl->chg_unplug_vec[index] = (status & CHGPRES);
		if (status & CHGPRES)
		{
			// Charger is plugged - continue
			ttw(ttr(TTrEventLow,"Polled - chg plugged (%d)" NL, status));
			pwr_ctrl->flag_chg_plugged = 1;
			pwr_ctrl->capacity = pwr_capacity(pwr_cfg->data.Vbat_avg_mV);
		}
		else
		{
			pwr_ctrl->flag_chg_plugged = 0;
		}

		if (pwr_ctrl->flag_mmi_registered == 1)
		{
			// If the charger is plugged and no charging is initiated this message will be repeated until charging has started
			// Charger plugged
			if ((pwr_ctrl->flag_chg_plugged == 1) &&  (pwr_ctrl->flag_chg_prev_plugged == 0))
			{
				pwr_ctrl->flag_chg_prev_plugged = 1;
				pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_PLUG_IND;
				mmi_send_msg(pwr_ctrl->mmi_ptr);
			}
			// Charger unplugged
			if ((pwr_ctrl->flag_chg_plugged == 0) && (pwr_ctrl->flag_chg_prev_plugged == 1))
			{
				pwr_ctrl->flag_chg_prev_plugged = 0;
				pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_UNPLUG_IND;
				mmi_send_msg(pwr_ctrl->mmi_ptr);
			}

			if (pwr_ctrl->flag_bat_unknown == 1)
			{
				// Battery unknown - inform the MMI
				pwr_ctrl->mmi_ptr->header.msg_id = MMI_BAT_UNKNOWN_IND;
				mmi_send_msg(pwr_ctrl->mmi_ptr);
				// Send only once
				pwr_ctrl->flag_bat_unknown = 0;
			}

		}

		pwr_check_timers();

		// If charger is plugged AND charger configuration is not read AND capacity is below configured limit - charger id Vchg(Load=0) must be measured
		if ((pwr_ctrl->flag_chg_plugged == 1) && 
			(pwr_ctrl->flag_chg_cfg_read == 0) && 
			(pwr_ctrl->capacity < pwr_cfg->bat.chg_again_thr)
		   )
		{

			// Measurement loop - Take the 2nd Vchg(0) measurements
			if (pwr_ctrl->count_chg_type == CHARGER_TYPE_SLIP)
			{
				if (pwr_cfg->data.Vchg > CHARGER_TYPE_OUT_OF_BOUND_LOW)
				{
					pwr_cfg->data.chg_id = pwr_cfg->data.Vchg;
				}
				else
				{
					// Out of bound!!
					ttr(TTrWarning, "Charger id out of bound! %d" NL, pwr_cfg->data.Vchg);
					pwr_cfg->data.chg_id = CHARGER_TYPE_TYPICAL;
				}
				ttw(ttr(TTrEventLow,"Using chg id=%d" NL, pwr_cfg->data.chg_id));
				error = pwr_read_chg_files();
				pwr_ctrl->flag_chg_cfg_read = 1;

				// Initialize the circular buffer with charger plugs
				for (i=0;i<= CONSECUTIVE_CHG_UNPLUGS-1 ;i++)
					pwr_ctrl->chg_unplug_vec[i] = CHGPRES;

				if (pwr_chg_start(pwr_temp_lookup(pwr_cfg->data.Tbat_avg, 0x01)) == TRUE)
				{
					if (pwr_cfg->chg.type == CI)
					{
						pwr_ctrl->state = CCI;
						// Enable charger plug/unplug interrupts for a CI charger						
						ABB_Write_Register_on_page(PAGE0, ITMASK, ALL_IT_UMSK);
					}
					else
					{
						if (pwr_cfg->chg.type == UNREGULATED)
							pwr_ctrl->state = LCI;
						else
						{
							ttr(TTrWarning, "Unknown charger type! %d" NL, pwr_cfg->chg.type);
							pwr_ctrl->state = LCI;
						}
					}
				#if (USE_Q401_CHG_CIRCUIT == 1)
					start_q401_charge();
				#endif

				#if (USE_Q402_CHG_CIRCUIT == 1)
					start_q402_charge();
				#endif

					// Stop the mmi info repetition timer
				#if (LCC_ALWAYS_MMI_INFO == 0)	/* Robert.Chen add, 2004-05-28 */
					pwr_stop_timer(&pwr_ctrl->time_begin_mmi_rep);
				#endif
				}
			}
			else
			{
				pwr_ctrl->count_chg_type++;
				ttw(ttr(TTrEventLow,"count_chg_type=%d" NL, pwr_ctrl->count_chg_type));
			}
		}
		break;
	case CCI :
		MI_dbgPrintf("zhq: we entered CCI state.");
		pwr_check_timers();
		// We are in the middle of a CI charging process - check stop criterias
		//
		if (pwr_chg_stop(pwr_temp_lookup(pwr_cfg->data.Tbat_avg, 0x01), pwr_cfg->data.Vbat_avg_mV) == FALSE)
		{
			// Continue to charge - change state?
			/* Robert.Chen modified, 2004-05-29 */
			if ((pwr_cfg->bat.type == LITHIUM) && (pwr_chg_ci_cv_transition(pwr_cfg->data.Vbat_avg_mV)== TRUE))
			//if ((pwr_cfg->bat.type == LITHIUM) && (pwr_chg_ci_cv_transition(ADC_to_mV(pwr_cfg->data.Vbat))== TRUE))
			{
				// Change state to CCV
				// Calculate k value
				// Start T2 timer
				// Start DC timer
				// Start T4 timer based on k value
				// Modulation ON
				pwr_ctrl->state = CCV;
				cv_stage = 0;
				cv_charging_house_keeping();
			}
		}
		else
		{
			// Change state
			pwr_ctrl->state = SUP;
			end_charging_house_keeping();

		}
		break;
	case LCI :
		// Poll ABB for charger unplug - Linear - unregulated chargers ONLY
		status = ind->data[State];
		pwr_ctrl->chg_unplug_vec[index] = (status & CHGPRES);
		if (status & CHGPRES)
		{
			// Charger is still plugged - continue
			ttw(ttr(TTrEventLow,"Polled - chg plugged (%d)" NL, status));
		}
		else
		{
			if (check_chg_unplug() == 1)
			{
				// Charger is not plugged anymore - stop!!
				ttw(ttr(TTrEventLow,"Verdict - chg not plugged (%d)" NL, status));
				// Change state
				pwr_ctrl->state = SUP;
				charger_unplug_house_keeping(); 
				pwr_free(request);
				return;
			}
			else
			{
				// False alarm - don't set flags
				ttw(ttr(TTrEventLow,"Polled - chg not plugged - FALSE alarm? (%d)" NL, status));
#if (USE_Q401_CHG_CIRCUIT == 1)
				if ((ind->data[Ichg] == 0) && (pwr_cfg->data.Vchg > 0))
				{
					// Charging current has disappeared due to a fast unplug??? Bug in IOTA ABB?
					pwr_ctrl->state = SUP;
					charger_unplug_house_keeping(); 
					pwr_free(request);
					return;
				}
#endif
#if (USE_Q402_CHG_CIRCUIT == 1)
				// FIXME: Really no false alarm code for Q402??
#endif

			}
		}
		pwr_check_timers();

		// We are in the middle of a CI charging process - check stop criterias
		if (pwr_chg_stop(pwr_temp_lookup(pwr_cfg->data.Tbat_avg, 0x01), pwr_cfg->data.Vbat_avg_mV) == FALSE)
		{
			// Continue to charge - change state?
			if ((pwr_cfg->bat.type == LITHIUM) && (pwr_chg_ci_cv_transition(pwr_cfg->data.Vbat_avg_mV)== TRUE))
			{
				// Change state to LCV
				// Calculate k value
				// Start T2 timer
				// Start DC timer
				// Start T4 timer based on k value
				// Modulation ON
				pwr_ctrl->state = LCV;
				cv_charging_house_keeping();
			}
		}
		else
		{

			// Change state
			pwr_ctrl->state = SUP;
			end_charging_house_keeping();

		}
		break;
	case CCV :
		MI_dbgPrintf("zhq: we entered CCV state.");
		pwr_check_timers();
		// We are in the middle of a CV charging process - check stop criterias
		if (pwr_chg_stop(pwr_temp_lookup(pwr_cfg->data.Tbat_avg, 0x01), pwr_cfg->data.Vbat_avg_mV) == FALSE)
		{
			// EMPTY - waiting for T4 and DC timeouts OR T2 timeout
		}
		else
		{
			// Change state
			pwr_ctrl->state = SUP;
			end_charging_house_keeping();
		}
		break;
	case LCV :
		// Poll ABB for charger unplug - Linear - unregulated chargers ONLY
		status = ind->data[State];
		pwr_ctrl->chg_unplug_vec[index] = (status & CHGPRES);
		if (status & CHGPRES)
		{
			// Charger is still plugged - continue
			ttw(ttr(TTrEventLow,"Polled - chg plugged (%d)" NL, status));
		}
		else
		{
#if (USE_Q401_CHG_CIRCUIT == 1)
			// Charger is not plugged anymore - stop!!
			ttw(ttr(TTrEventLow,"Verdict - chg not plugged (%d)" NL, status));
			// Change state
			pwr_ctrl->state = SUP;
			charger_unplug_house_keeping(); 
			pwr_free(request);
			return;
#endif
#if (USE_Q402_CHG_CIRCUIT == 1)
			if (check_chg_unplug() == 1)
			{
				// Charger is not plugged anymore - stop!!
				ttw(ttr(TTrEventLow,"Verdict - chg not plugged (%d)" NL, status));
				// Change state
				pwr_ctrl->state = SUP;
				charger_unplug_house_keeping(); 
				pwr_free(request);
				return;
			}
#endif
		}
		pwr_check_timers();

		// We are in the middle of a CV charging process - check stop criterias
		if (pwr_chg_stop(pwr_temp_lookup(pwr_cfg->data.Tbat_avg, 0x01), pwr_cfg->data.Vbat_avg_mV) == FALSE)
		{
			// EMPTY - waiting for T4 and DC timeouts OR T2 timeout
		}
		else
		{
			// Change state
			pwr_ctrl->state = SUP;
			end_charging_house_keeping();
		}
		break;
	default :
		{
			// Exception Handling - Unknown State
			ttr(TTrFatal, "process_spi_adc_indication: Unknown State: %d" NL, pwr_ctrl->state);
		}
	}
	pwr_free(request);
	return RV_OK;
}


T_RV_RET process_pwr_tm_read_request         (T_PWR_REQ *request)
{
	// pwtr - see RD818

	int index, fid;
	T_FFS_SIZE error;
	T_FFS_STAT stat;
	struct etm_tm_ind_s *msg;
	extern T_ETM_ENV_CTRL_BLK *etm_env_ctrl_blk;
	char name[20];
#if (TEST_PWR_MMI_INTERFACE == 1)
	T_RV_RETURN_PATH return_path;
#endif

	ttw(ttr(TTrTmpwr,"pwtr (%d)" NL, 0));
	msg = (struct etm_tm_ind_s *)request;

	fid   = msg->data[0];
	index = msg->data[1];

	ttw(ttr(TTrTmpwr, "data[0]=%d" NL, msg->data[0]));
	ttw(ttr(TTrTmpwr, "data[1]=%d" NL, msg->data[1]));
	ttw(ttr(TTrTmpwr, "data[2]=%d" NL, msg->data[2]));
	ttw(ttr(TTrTmpwr, "data[3]=%d" NL, msg->data[3]));

	name[0] = 0; // FIXME: Really needed?
	msg->status = 0; // Default status is OK

	// All reads start from index 2 since the fid + index is kept
	switch (index)
	{
	case PWR_CFG_ID :
		// We can read different configurations swiching between them using pwtw
		msg->size = PWR_CFG_ID_SIZE;
		msg->data[2] = pwr_ctrl->cfg_id;
		msg->data[3] = pwr_ctrl->chg_cfg_id;
		ttw(ttr(TTrTmpwr,"Using cfg_id %d" NL, msg->data[2]));
		ttw(ttr(TTrTmpwr,"Using chg_cfg_id %d" NL, msg->data[3]));

		// If either the bat.cfg or the chg.cfg lacks object not found is returned

		// If (battery) file can't be stat'ed then the configuration doesn't exist
		build_name("/pwr/bat/bat", &pwr_cfg->data.cfg_id, 12, ".cfg", name);
		error = ffs_stat(name, &stat);

		// If charger file can't be stat'ed then the configuration doesn't exist
		build_name("/pwr/chg/chg", &pwr_cfg->data.chg_cfg_id, 12, ".cfg", name);
		error = ffs_stat(name, &stat);
		break;
	case PWR_COMMON :
		// Read the /pwr/common.cfg file
		// NOTE: sizeof(pwr_cfg->common) = 16 <>  (14 byte alignment)
		msg->size = PWR_COMMON_CFG_SIZE;
		error = ffs_fread("/pwr/common.cfg", &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr,"Read %d bytes" NL, error));
		break;
	case PWR_CHG :
		// Read the /pwr/chg/chg<cfgid>.cfg file
		msg->size = PWR_CHG_CFG_SIZE;
		build_name("/pwr/chg/chg", &pwr_cfg->data.chg_cfg_id, 12, ".cfg", name);
		error = ffs_fread(name, &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr, "Read %d bytes " NL, error));
		break;
	case PWR_BAT :
		// Read the /pwr/bat/bat<cfgid>.cfg file
		msg->size = PWR_BAT_CFG_SIZE;
		build_name("/pwr/bat/bat", &pwr_cfg->data.cfg_id, 12, ".cfg", name);
		error = ffs_fread(name, &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr,"Read %d bytes" NL, error));
		break;
	case PWR_TEMP :
		// Read the /pwr/bat/temp<cfgid>.cfg file
		msg->size = PWR_TEMP_CFG_SIZE;
		build_name("/pwr/bat/temp", &pwr_cfg->data.cfg_id, 13, ".cfg", name);
		error = ffs_fread(name, &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr,"Read %d bytes" NL, error));
		break;
	case PWR_MMI:
		// Read the /mmi/pwr/bsie.cfg file
		msg->size = PWR_MMI_CFG_SIZE;
		error = ffs_fread("/mmi/pwr/bsie.cfg", &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr,"Read %d bytes" NL, error));
		break;
	case PWR_I2V_CAL :
		// Read the /pwr/i2v.cal file
		msg->size = PWR_I2V_CAL_SIZE;
		error = ffs_fread("/pwr/i2v.cal", &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr,"Read %d bytes" NL, error));
		break;
	case PWR_VBAT_CAL :
		// Read the /pwr/vbat.cal file
		msg->size = PWR_VBAT_CAL_SIZE;
		error = ffs_fread("/pwr/vbat.cal", &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr,"Read %d bytes" NL, error));
		break;
	case PWR_MMI_TEST :
#if (TEST_PWR_MMI_INTERFACE == 1)
// Trigger MMI registration 
		return_path.callback_func = mmi_test_cb_function;
		ttw(ttr(TTrInit,"before: &mmi_test_cb_function=(0x%x)" NL, return_path.callback_func));
		return_path.addr_id = NULL;	// FIXME??
		ttw(ttr(TTrTmpwr,"MMI testing callback: %d" NL, 0x0));
		msg->size = 1;
		msg->data[2] = 0xBA;
		pwr_register(&return_path, &mmi_data);
		ttw(ttr(TTrInit,"after: &mmi_test_cb_function=(0x%x)" NL, pwr_ctrl->rpath.callback_func));
		ttw(ttr(TTrTmpwr,"MMI testing callback: %d" NL, 0xFF));
#endif
		break;
	case PWR_DYNAMIC :
		// Dump 'dynamic' configuration data to trace
		msg->size = PWR_DYNAMIC_SIZE;
		msg->data[2] = 0xBA;
		ttr(TTrAll,"*PWR Module Version = 0x%x" NL, PWRVERSION );
		ttr(TTrAll,"*bat_id = %d" NL,               pwr_cfg->data.bat_id );
		ttr(TTrAll,"*state = %d" NL,                pwr_ctrl->state);
		ttr(TTrAll,"*chg_id = %d" NL,               pwr_cfg->data.chg_id );
		ttr(TTrAll,"*Vbat = %d" NL,                 pwr_cfg->data.Vbat);
		ttr(TTrAll,"*Vbat_avg = %d" NL,             pwr_cfg->data.Vbat_avg);
		ttr(TTrAll,"*Vbat_avg_mV = %d" NL,          pwr_cfg->data.Vbat_avg_mV);
		ttr(TTrAll,"*Tbat = %d" NL,                 pwr_cfg->data.Tbat);
		ttr(TTrAll,"*Tbat_avg = %d" NL,             pwr_cfg->data.Tbat_avg);
		ttr(TTrAll,"*Vchg = %d" NL,                 pwr_cfg->data.Vchg);
		ttr(TTrAll,"*Ichg = %d" NL,                 pwr_cfg->data.Ichg);
		ttr(TTrAll,"*Cbat = %d" NL,                 pwr_cfg->data.Cbat);
		ttr(TTrAll,"*cfg_id = %d" NL,               pwr_cfg->data.cfg_id);
		ttr(TTrAll,"*chg_cfg_id = %d" NL,           pwr_cfg->data.chg_cfg_id);
		ttr(TTrAll,"*bforce = %d" NL,               pwr_cfg->data.bforce);
		ttr(TTrAll,"*cforce = %d" NL,               pwr_cfg->data.cforce);
		ttr(TTrAll,"*k = %d" NL,                    pwr_cfg->data.k);
		ttr(TTrAll,"*T4 = %d" NL,                   pwr_cfg->data.T4);
		ttr(TTrAll,"*T1 elapsed = %d" NL,           pwr_ctrl->time_elapsed_T1);
		ttr(TTrAll,"*T2 elapsed = %d" NL,           pwr_ctrl->time_elapsed_T2);
		ttr(TTrAll,"*T3 elapsed = %d" NL,           pwr_ctrl->time_elapsed_T3);
		ttr(TTrAll,"*MMI timer elapsed = %d" NL,    pwr_ctrl->time_elapsed_mmi_rep);
		error = PWR_OK;
		break;
	case PWR_TRACE_MASK:
		// Read the trace mask of the PWR module 
		msg->size = PWR_TMASK_SIZE;
		memcpy(&msg->data[2], &pwr_ctrl->tmask, msg->size);
		ttw(ttr(TTrTmpwr,"tmask: 0x%x" NL, pwr_ctrl->tmask));
		break;
	default :
		{
			// Unknown index
			error = PWR_INDEX;
			ttr(TTrWarning, "Unknown index! %d" NL, index);
		}
	}

	ttw(ttr(TTrTmpwr, "data[0]=%d" NL, msg->data[0]));
	ttw(ttr(TTrTmpwr, "data[1]=%d" NL, msg->data[1]));
	ttw(ttr(TTrTmpwr, "data[2]=%d" NL, msg->data[2]));
	ttw(ttr(TTrTmpwr, "data[3]=%d" NL, msg->data[3]));
	ttw(ttr(TTrTmpwr, "data[4]=%d" NL, msg->data[4]));
	ttw(ttr(TTrTmpwr, "data[5]=%d" NL, msg->data[5]));

	if (error < 0)
		msg->status = -error;

	// Pack - re-use the buffer allocated in ETM
	// De-allocate in ETM
	msg->header.msg_id        = PWR_TM_READ_IND;
	msg->header.src_addr_id   = pwr_ctrl->addr_id;
	msg->header.dest_addr_id  = etm_env_ctrl_blk->addr_id;
	msg->header.callback_func = NULL;

	// Send indication with data to ETM task
	if (error = rvf_send_msg(etm_env_ctrl_blk->addr_id, msg) != RV_OK)
	{
		ttr(TTrFatal, "PWR FATAL: Send failed! %d" NL, error);
		return PWR_RV_FATAL;
	}
	ttw(ttr(TTrTmpwr,"pwtr (%d)" NL, 0xFF));
	return RV_OK;

}


T_RV_RET process_pwr_tm_write_request        (T_PWR_REQ *request)
{
	int fid, index;
	// pwtw - see RD818
	T_FFS_SIZE error;
	struct etm_tm_ind_s *msg;
	extern T_ETM_ENV_CTRL_BLK *etm_env_ctrl_blk;
	char name[20];

	ttw(ttr(TTrTmpwr,"pwtw (%d)" NL, 0));
	msg = (struct etm_tm_ind_s *)request;

	fid   = msg->data[0];
	index = msg->data[1];
	name[0] = 0; // FIXME: Really needed?
	msg->status = 0; // Default status is OK

	ttw(ttr(TTrTmpwr, "data[0]=%d" NL, msg->data[0]));
	ttw(ttr(TTrTmpwr, "data[1]=%d" NL, msg->data[1]));
	ttw(ttr(TTrTmpwr, "data[2]=%d" NL, msg->data[2]));
	ttw(ttr(TTrTmpwr, "data[3]=%d" NL, msg->data[3]));

	switch (index)
	{
	case PWR_CFG_ID :
		// Write the configuration id to be used
		pwr_ctrl->cfg_id = msg->data[2];
		pwr_ctrl->chg_cfg_id = msg->data[3];
		pwr_cfg->data.cfg_id = pwr_ctrl->cfg_id + '0';
		pwr_cfg->data.chg_cfg_id = pwr_ctrl->chg_cfg_id + '0';
		ttw(ttr(TTrTmpwr,"Switched to bat id=%d" NL, pwr_ctrl->cfg_id));
		ttw(ttr(TTrTmpwr,"Switched to chg id=%d" NL, pwr_ctrl->chg_cfg_id));
		break;
	case PWR_COMMON :
		// Write the /pwr/common.cfg file
		// Blocking version of ffs_fwrite is used since this is a test mode command
		error = ffs_fwrite("/pwr/common.cfg", &msg->data[2], PWR_COMMON_CFG_SIZE);
		ttw(ttr(TTrTmpwr,"Wrote %d bytes" NL, error));
		break;
	case PWR_CHG :
		// Write the /pwr/chg/chg<cfgid>.cfg file
		msg->size = PWR_CHG_CFG_SIZE;
		build_name("/pwr/chg/chg", &pwr_cfg->data.chg_cfg_id, 12, ".cfg", name);
		error = ffs_fwrite(name, &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr, "Wrote %d bytes = %d" NL, error));
		break;
	case PWR_BAT :
		// Write the /pwr/bat/bat<cfgid>.cfg file
		msg->size = PWR_BAT_CFG_SIZE;
		build_name("/pwr/bat/bat", &pwr_cfg->data.cfg_id, 12, ".cfg", name);
		error = ffs_fwrite(name, &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr, "Wrote %d bytes = %d" NL, error));
		break;
	case PWR_TEMP :
		// Write the /pwr/bat/temp<cfgid>.cfg file
		msg->size = PWR_TEMP_CFG_SIZE;
		build_name("/pwr/bat/temp", &pwr_cfg->data.cfg_id, 13, ".cfg", name);
		error = ffs_fwrite(name, &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr, "Wrote %d bytes = %d" NL, error));
		break;
	case PWR_MMI:
		// Write the /mmi/pwr/bsie.cfg file
		msg->size = PWR_MMI_CFG_SIZE;
		error = ffs_fwrite("/mmi/pwr/bsie.cfg", &msg->data[2], msg->size);
		ttw(ttr(TTrTmpwr, "Wrote %d bytes = %d" NL, error));
		break;
	case PWR_I2V_CAL :
		// Write the /pwr/i2v.cal file
		error = ffs_fwrite("/pwr/i2v.cal", &msg->data[2], PWR_I2V_CAL_SIZE);
		ttw(ttr(TTrTmpwr,"Wrote %d bytes" NL, PWR_I2V_CAL_SIZE));
		break;
	case PWR_VBAT_CAL:
		// Write the /pwr/vbat.cal file
		error = ffs_fwrite("/pwr/vbat.cal", &msg->data[2], PWR_VBAT_CAL_SIZE);
		ttw(ttr(TTrTmpwr,"Wrote %d bytes" NL, PWR_VBAT_CAL_SIZE));
		break;
	case PWR_MMI_TEST :
		break;
	case PWR_TRACE_MASK:
		// Write the trace mask of the PWR module 
		memcpy(&pwr_ctrl->tmask, &msg->data[2], sizeof(pwr_ctrl->tmask));
		ttw(ttr(TTrTmpwr,"Wrote tmask %d" NL, pwr_ctrl->tmask));
		pwr_ttr_init(pwr_ctrl->tmask);
		break;
	default :
		{
			// Unknown index
			error = PWR_INDEX;
			ttr(TTrWarning, "Unknown index! %d" NL, index);
		}
	}

	msg->size = 1; // Size of write message reply is always 1

	if (error < 0)
		msg->status = -error;

	// Pack - re-use the buffer allocated in ETM
	// De-allocate in ETM
	msg->header.msg_id        = PWR_TM_WRITE_IND;
	msg->header.src_addr_id   = pwr_ctrl->addr_id;
	msg->header.dest_addr_id  = etm_env_ctrl_blk->addr_id;
	msg->header.callback_func = NULL;

	// Send indication with status & data to ETM task
	if (error = rvf_send_msg(etm_env_ctrl_blk->addr_id, msg) != RV_OK)
	{
		ttr(TTrFatal, "PWR FATAL: Send failed! %d" NL, error);
		return PWR_RV_FATAL;
	}
	ttw(ttr(TTrTmpwr,"pwtw (%d)" NL, 0xFF));
	return RV_OK;
}


T_RV_RET process_abb_chg_unplugged_ind       (T_PWR_REQ *request)
{

	charger_unplug_house_keeping(); 
	// Disable charger interrupts - they where enabled when the CI charger was connected
#if (LCC_ALWAYS_MMI_INFO == 0)	/* Robert.Chen added, 2004-05-28 */
	ABB_Write_Register_on_page(PAGE0, ITMASK, CHARGER_IT_MSK);
#endif

	// A charger can be unplugged in any state
	switch (pwr_ctrl->state)
	{
	case SUP :
	case CCI :
	case CCV :
		// Change state - T3 is running
		pwr_ctrl->state = SUP;
		break;
#if (LCC_ALWAYS_MMI_INFO == 1)	/* Robert.Chen added, 2004-05-28 */
	case PRE :
		pwr_ctrl->state = INI;
		break;
#endif
		
	default :
		{
			// Exception Handling - Unknown State
			ttr(TTrFatal, "process_abb_chg_unplugged_ind: Unknown State: %d" NL, pwr_ctrl->state);
		}
	}
	pwr_free(request);
	return RV_OK;
}

void pwr_send_msg(uint32 msg_id, T_RVF_ADDR_ID src_addr_id, T_RVF_ADDR_ID  dest_addr_id) {
	struct pwr_req_s *msg;

	if ((msg = pwr_malloc(sizeof(struct pwr_req_s))) == NULL)
	{
		return;
	}
	msg->header.msg_id        = msg_id;
	msg->header.src_addr_id   = src_addr_id;
	msg->header.dest_addr_id  = dest_addr_id;
	msg->header.callback_func = NULL;
	if (rvf_send_msg(dest_addr_id, msg) != RV_OK)
	{
		ttr(TTrFatal, "PWR FATAL: Send failed! %d" NL, 0xFF);
	}
}

void mmi_send_msg(struct mmi_info_ind_s *event) 
{
	ttw(ttr(TTrInit,"mmi_send_msg(%d)" NL, 0));
	if (pwr_ctrl->rpath.callback_func)
	{
		ttw(ttr(TTrInit,"Using callback (0x%x)" NL, pwr_ctrl->rpath.callback_func));
		(pwr_ctrl->rpath.callback_func) (event);
	}
	else
	{
		if (pwr_ctrl->rpath.addr_id)
		{
			// TESTME
			rvf_send_msg(pwr_ctrl->rpath.addr_id, event);
		}
		else
		{
			ttr(TTrFatal,"PWR FATAL: mmi_send_msg(%d) No return path" NL, 0xFF);
			return;
		}
	}
	ttw(ttr(TTrInit,"mmi_send_msg(%d)" NL, 0xFF));
}

void start_pre_charge(void) {
	unsigned short	reg; 
	ttw(ttr(TTrInitLow,"start_pre_charge" NL, 0x00));
#if (USE_Q401_CHG_CIRCUIT == 1)
	/* Program the DAC with the constant current value */
	// Value is hardcoded!! No configuration exist for state PRE
	#if 0
	ABB_Write_Register_on_page(PAGE0,BCICTL2,0); 
	ABB_Write_Register_on_page(PAGE0,BCICTL2,0x4); 
	ABB_Write_Register_on_page(PAGE0,BCICTL2,0x25);
	reg = ABB_Read_Register_on_page(PAGE0, BCICTL2);
	ttr(0xffffffff, "BCICTL2=0x%2X" NL, reg);
	return;
	#endif
	/* Robert.Chen add */
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0x0002);
	
	ABB_Write_Register_on_page(PAGE0, CHGREG, ICHG_PRECHG);

	/* Enable the charger  */
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0x0003);
#endif
#if (USE_Q402_CHG_CIRCUIT == 1)
	pwr_modulate_init();
#endif
	ttw(ttr(TTrInitLow,"start_pre_charge" NL, 0xFF));

}

void start_q401_charge(void) {

	ttw(ttr(TTrInitLow,"start_q401_charge" NL, 0x00));
	if (pwr_ctrl->flag_mmi_registered == 1)
	{
		pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_START_IND;
		mmi_send_msg(pwr_ctrl->mmi_ptr);
	}

	/* Robert.Chen add */
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0x0002);
	
	/* Program the DAC with the constant current value taken from /pwr/chg/chg<N>.cfg */
	ABB_Write_Register_on_page(PAGE0, CHGREG, pwr_cfg->chg.ichg_max);

	/* Enable the CI charger  */
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0x0003);

	/* If not already started - start the overall charger timer T1 */
	if (pwr_ctrl->time_begin_T1 == 0)
	{
		pwr_start_timer(&pwr_ctrl->time_begin_T1);
		ttw(ttr(TTrTimerLow,"T1 started(%d)" NL, 0));
	}
	ttw(ttr(TTrInitLow,"start_q401_charge" NL, 0xFF));

}

void start_q402_charge(void) {

	ttw(ttr(TTrInitLow,"start_q402_charge(%d)" NL, 0x00));
	if (pwr_ctrl->flag_mmi_registered == 1)
	{
		pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_START_IND;
		mmi_send_msg(pwr_ctrl->mmi_ptr);
	}

	// Needed if we want Ichg measurements - else they are disabled!!
	ABB_Write_Register_on_page(PAGE0, CHGREG, pwr_cfg->chg.ichg_max);
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0x0003);

	pwr_modulate_init();

	/* If not already started - start the overall charger timer T1 */
	if (pwr_ctrl->time_begin_T1 == 0)
	{
		pwr_start_timer(&pwr_ctrl->time_begin_T1);
		ttw(ttr(TTrTimerLow,"T1 started(%d)" NL, 0));
	}
	ttw(ttr(TTrInitLow,"start_q402_charge(%d)" NL, 0xFF));

}

void stop_q401_charge(void) {

	ttw(ttr(TTrInitLow,"stop_q401_charge(%d)" NL, 0x00));
	if (pwr_ctrl->flag_mmi_registered == 1)
	{
		pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_STOP_IND;
		mmi_send_msg(pwr_ctrl->mmi_ptr);
	}
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0);
	ttw(ttr(TTrInitLow,"stop_q401_charge(%d)" NL, 0xFF));

}

void stop_q402_charge(void) {

	ttw(ttr(TTrInitLow,"stop_q402_charge(%d)" NL, 0x00));
	if (pwr_ctrl->flag_mmi_registered == 1)
	{
		pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_STOP_IND;
		mmi_send_msg(pwr_ctrl->mmi_ptr);
	}
	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0);

	pwr_modulate_off();
	ttw(ttr(TTrInitLow,"stop_q402_charge(%d)" NL, 0xFF));

}

void stop_pre_charge(void) {

	ABB_Write_Register_on_page(PAGE0, BCICTL2, 0);

}

int check_chg_unplug() {
	int i;

#if (USE_Q401_CHG_CIRCUIT == 1)
	// Check that the latest values of charger status confirm that we have a charger unplug
	// Search through the vector - charger is only assumed unplugged when ALL CONSECUTIVE_CHG_UNPLUGS elements are 0
	// Returns 1 (TRUE) if the charger is assumed unplugged
	// Returns 0 (FALSE) if the charger is assumed plugged
	// Assume that the charger unplug was true if the battery voltage is sufficiently high
	if (pwr_cfg->data.Vbat_avg_mV < VBAT_PRECHG_STOP)
	{
		for (i=0;i<= CONSECUTIVE_CHG_UNPLUGS-1 ;i++)
		{
			ttw(ttr(TTrTimerLow,"chg_unplug_vec(%d)" NL, pwr_ctrl->chg_unplug_vec[i]));
			if (pwr_ctrl->chg_unplug_vec[i] != 0)
				return 0;
		}
	}
	return 1;
#endif
#if (USE_Q402_CHG_CIRCUIT == 1)
	// Assume that the charger unplug was true if BOTH the charger voltage & charger current 
	// are below configurable Vchg & Ichg threshold values
	if ((pwr_cfg->data.Vchg < VCHG_Q402_THR) && (pwr_cfg->data.Ichg < ICHG_Q402_THR))
	{
		ttw(ttr(TTrInitLow,"check_chg_unplug: Vchg=(%d)" NL, pwr_cfg->data.Vchg));
		ttw(ttr(TTrInitLow,"check_chg_unplug: Ichg=(%d)" NL, pwr_cfg->data.Ichg));
		return 1;
	}
	return 0;
#endif
}

void charger_unplug_house_keeping(void) {
	// Flag the charger unplug
	pwr_ctrl->flag_chg_plugged = 0;
	pwr_ctrl->flag_chg_prev_plugged = 0;

	// Read charger configuration next time charger is inserted
	pwr_ctrl->flag_chg_cfg_read = 0;

	// We must have more charger id measurements if the charger is re-connected
	pwr_ctrl->count_chg_type = 0;

	if (pwr_ctrl->flag_mmi_registered == 1)
	{
		// Send charger unplug to the MMI
		pwr_ctrl->mmi_ptr->header.msg_id = MMI_CHG_UNPLUG_IND;
		mmi_send_msg(pwr_ctrl->mmi_ptr);

		// Re-start the mmi info repetition timer
	#if (LCC_ALWAYS_MMI_INFO == 0)	/* Robert.Chen add, 2004-05-28 */
		pwr_start_timer(&pwr_ctrl->time_begin_mmi_rep);
	#endif
	}

	// if T1 or T2 is running (which mean we have NOT had a normal charging stop)
	if ((pwr_ctrl->time_begin_T1 != 0) || (pwr_ctrl->time_begin_T2 != 0))
	{
		// Start T3 timer if not already started
		if (pwr_ctrl->time_begin_T3 == 0)
		{
			pwr_start_timer(&pwr_ctrl->time_begin_T3);
			ttw(ttr(TTrTimerLow,"T3 started(%d)" NL, 0));
		}
		// Switch of charging circuit

#if (USE_Q401_CHG_CIRCUIT == 1)
		stop_q401_charge();
#endif
#if (USE_Q402_CHG_CIRCUIT == 1)
		stop_q402_charge();
#endif
	}
}

void cv_charging_house_keeping(void) {
	// Calculate k value
	// Start DC timer
	// Start T4 timer based on k value
	// Modulation ON
	ttw(ttr(TTrTimerLow,"Vbat_avg, chg_start_thr, chg_stop_thr (%d), (%d), (%d)" NL, pwr_cfg->data.Vbat_avg_mV, pwr_cfg->bat.chg_start_thr, pwr_cfg->bat.chg_stop_thr));
#if 0	/* Robert.Chen replaced, 2004-05-28 */
	if ((pwr_cfg->data.Vbat_avg_mV - pwr_cfg->bat.chg_start_thr) <= 0)
	{
		// Assign k to max value if Vbat_avg value is below start value
		pwr_cfg->data.k = PWR_MAX_K;
	}
	else
	{
		pwr_cfg->data.k = (255 * (pwr_cfg->bat.chg_stop_thr - pwr_cfg->data.Vbat_avg_mV) / (pwr_cfg->bat.chg_stop_thr - pwr_cfg->bat.chg_start_thr));
		if (pwr_cfg->data.k <= PWR_MIN_K)
			pwr_cfg->data.k = PWR_MIN_K;
	}
#else
	if(pwr_cfg->data.Vbat_avg_mV >= pwr_cfg->bat.chg_stop_thr || cv_overflow)
	//if(ADC_to_mV(pwr_cfg->data.Vbat) >= pwr_cfg->bat.chg_stop_thr)
	{

		cv_overflow = 0;
		if(pwr_cfg->data.k >= pwr_cfg->bat.chg_ctrl_thr)
		{
		#if 0
			pwr_cfg->data.k = PWR_MAX_K / (1 + cv_stage); 
			cv_stage ++;
		#else
			cv_stage++;
			pwr_cfg->data.k = PWR_MAX_K - 100 * cv_stage;
			if(pwr_cfg->data.k <= pwr_cfg->bat.chg_ctrl_thr)
			{
				pwr_cfg->data.k = pwr_cfg->bat.chg_ctrl_thr - 1;
				cv_stage--;
			}	
		#endif			
		}
	}	
	else if(pwr_cfg->data.Vbat_avg_mV <= (pwr_cfg->bat.chg_start_thr - 100))
	{
		if(cv_stage > 0)
			cv_stage--;
	#if 0	 
		pwr_cfg->data.k = PWR_MAX_K / (1 + cv_stage);
	#else
		pwr_cfg->data.k = PWR_MAX_K - 100 * cv_stage;
	#endif
	}
	else if(cv_stage == 0)	/* first cv stage */
	{
		pwr_cfg->data.k = PWR_MAX_K;
	}
	
#endif	

	/* Robert.Chen added, 2004-05-29, fix TI bug, ensure start the CV timer only once during CV charging */
	if(pwr_ctrl->time_begin_T2 == 0)
	/* add end */
		pwr_start_timer(&pwr_ctrl->time_begin_T2);
		
	ttw(ttr(TTrTimerLow,"T2 started(%d)" NL, 0));

	//ttw(ttr(TTrTimerLow, "k=%d" NL, pwr_cfg->data.k));
	ttw(ttr(0xff000000, "k=%d" NL, pwr_cfg->data.k));	//Robert.Chen modified
	pwr_cfg->data.T4 = pwr_cfg->data.k * pwr_cfg->common.mod_cycle/255;
	ttw(ttr(TTrTimerLow,"T4 timer (%d)" NL, pwr_cfg->data.T4));

	ttw(ttr(TTrTimerLow,"DC started(%d)" NL, 0));
	pwr_start_timer(&pwr_ctrl->time_begin_mod_cycle);

	ttw(ttr(TTrTimerLow,"T4 started(%d)" NL, 0));
	pwr_start_timer(&pwr_ctrl->time_begin_T4);

	pwr_modulate_on();
}

void end_charging_house_keeping(void) {
	// Stop & reset timers
	pwr_stop_timers();

	// Stop charging
#if (USE_Q401_CHG_CIRCUIT == 1)
	stop_q401_charge();
#endif
#if (USE_Q402_CHG_CIRCUIT == 1)
	stop_q402_charge();
#endif

	// Start the mmi info repetition timer
#if (LCC_ALWAYS_MMI_INFO == 0)	/* Robert.Chen add, 2004-05-28 */
	if (pwr_ctrl->flag_mmi_registered == 1)
		pwr_start_timer(&pwr_ctrl->time_begin_mmi_rep);
#endif

	// Read charger configuration next time charger is inserted
	pwr_ctrl->flag_chg_cfg_read = 0;
}

#if (TEST_PWR_MMI_INTERFACE == 1)
// Callback function used for testing MMI reporting
static void mmi_test_cb_function(void *ptr) {
	T_PWR_MMI_INFO_IND_EVENT *event;


	event = (T_PWR_MMI_INFO_IND_EVENT *)ptr;

	ttw(ttr(TTrInit, "mmi_test_cb_function (%d)" NL, 0));
	ttw(ttr(TTrInit, "MMI event: (%d)" NL, event->header.msg_id));
	ttw(ttr(TTrInit, "mmi_test_cb_function (%d)" NL, 0xFF));
}
#endif

